Type: Package
Package: immunogenetr
Title: A Comprehensive Toolkit for Clinical HLA Informatics
Version: 1.3.0
Authors@R: c(
    person("Nicholas", "Brown", , "nicholas.brown@pennmedicine.upenn.edu", role = c("cre", "aut"),
           comment = c(ORCID = "0000-0002-0046-2315")),
    person("Busra", "Coskun", , "bcoskun@sas.upenn.edu", role = "aut",
           comment = c(ORCID = "0009-0008-6828-3453"))
  )
Maintainer: Nicholas Brown <nicholas.brown@pennmedicine.upenn.edu>
Description: A comprehensive toolkit for clinical Human Leukocyte Antigen (HLA) informatics, built on 'tidyverse' <https://tidyverse.tidyverse.org/> principles and making use of genotype list string (GL string, Mack et al. (2023) <doi:10.1111/tan.15126>) for storing and computing HLA genotype data.  Specific functionalities include: coercion of HLA data in tabular format to and from GL string; calculation of matching and mismatching in all directions, with multiple output formats; automatic formatting of HLA data for searching within a GL string; truncation of molecular HLA data to a specific number of fields; and reading HLA genotypes in HML files and extracting the GL string. This library is intended for research use. Any application making use of this package in a clinical setting will need to be independently validated according to local regulations.
License: GPL (>= 3)
URL: https://immunogenetr.org,
        https://github.com/k96nb01/immunogenetr_package,
        https://doi.org/10.1016/j.humimm.2025.111619
BugReports: https://github.com/k96nb01/immunogenetr_package/issues
Depends: R (>= 4.1.0)
Imports: cli (>= 3.0.0), dplyr (>= 1.1.1), purrr (>= 1.0.2), rlang (>=
        1.1.2), stringi (>= 1.7.0), stringr (>= 1.5.0), tibble (>=
        3.1.3), tidyr (>= 1.3.0), tidyselect (>= 1.2.0), xml2 (>=
        1.3.6)
Suggests: knitr, rmarkdown, testthat (>= 3.0.0)
VignetteBuilder: knitr
Config/testthat/edition: 3
Encoding: UTF-8
Language: en-US
LazyData: true
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-26 04:51:42 UTC; drnic
Author: Nicholas Brown [cre, aut] (ORCID:
    <https://orcid.org/0000-0002-0046-2315>),
  Busra Coskun [aut] (ORCID: <https://orcid.org/0009-0008-6828-3453>)
Repository: CRAN
Date/Publication: 2026-05-26 08:50:10 UTC
Built: R 4.6.0; ; 2026-05-26 10:29:44 UTC; unix

# Generated by roxygen2: do not edit by hand

export(GLstring_expand_longer)
export(GLstring_gene_copies_combine)
export(GLstring_genes)
export(GLstring_genes_expanded)
export(GLstring_genotype_ambiguity)
export(GLstring_regex)
export(GLstring_to_ambiguity_table)
export(HLA_column_repair)
export(HLA_columns_to_GLstring)
export(HLA_match_number)
export(HLA_match_summary_HCT)
export(HLA_mismatch_alleles)
export(HLA_mismatch_base)
export(HLA_mismatch_logical)
export(HLA_mismatch_number)
export(HLA_mismatched_alleles)
export(HLA_prefix_add)
export(HLA_prefix_remove)
export(HLA_truncate)
export(HLA_validate)
export(ambiguity_table_to_GLstring)
export(read_HML)
importFrom(cli,cli_abort)
importFrom(dplyr,"%>%")
importFrom(dplyr,across)
importFrom(dplyr,bind_rows)
importFrom(dplyr,contains)
importFrom(dplyr,distinct)
importFrom(dplyr,filter)
importFrom(dplyr,mutate)
importFrom(dplyr,na_if)
importFrom(dplyr,rename_with)
importFrom(dplyr,select)
importFrom(dplyr,starts_with)
importFrom(dplyr,summarise)
importFrom(dplyr,summarize)
importFrom(purrr,map)
importFrom(rlang,caller_env)
importFrom(stringi,stri_endswith_fixed)
importFrom(stringi,stri_extract_first_regex)
importFrom(stringi,stri_split_fixed)
importFrom(stringi,stri_startswith_fixed)
importFrom(stringr,str_c)
importFrom(stringr,str_detect)
importFrom(stringr,str_escape)
importFrom(stringr,str_extract)
importFrom(stringr,str_flatten)
importFrom(stringr,str_replace)
importFrom(stringr,str_replace_all)
importFrom(tibble,as_tibble)
importFrom(tibble,tibble)
importFrom(tidyr,pivot_wider)
importFrom(tidyr,unite)
importFrom(tidyselect,all_of)
importFrom(xml2,read_xml)
importFrom(xml2,xml_attr)
importFrom(xml2,xml_find_all)
importFrom(xml2,xml_ns)
importFrom(xml2,xml_text)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = FALSE,
  comment = "#>"
)

# Render data frames and tibbles as HTML tables in chunk output, instead of
# the default monospace console-style print. The kable HTML is wrapped in a
# pandoc raw-HTML block (```{=html} ... ```) so pandoc passes the content
# through verbatim. Without this, pandoc's markdown_in_html_blocks extension
# re-parses the cell contents as markdown, turning "^...^" in GL strings
# into <sup>...</sup>.
knit_print.data.frame <- function(x, ...) {
  html <- as.character(knitr::kable(x, format = "html"))
  knitr::asis_output(paste0("\n```{=html}\n", html, "\n```\n"))
}
registerS3method("knit_print", "data.frame", knit_print.data.frame, envir = asNamespace("knitr"))
registerS3method("knit_print", "tbl_df",     knit_print.data.frame, envir = asNamespace("knitr"))
registerS3method("knit_print", "tbl",        knit_print.data.frame, envir = asNamespace("knitr"))

## ----message=FALSE------------------------------------------------------------
library(immunogenetr)
library(dplyr)

## -----------------------------------------------------------------------------
# HLA_typing_1 contains typing for 10 individuals across all classical HLA loci.
head(HLA_typing_1, 3)

## -----------------------------------------------------------------------------
HLA_typing_GL <- HLA_typing_1 %>%
  # Convert all typing columns (A1 through DPB1_2) into a GL string.
  mutate(
    GL_string = HLA_columns_to_GLstring(., HLA_typing_columns = A1:DPB1_2),
    .after = patient
  ) %>%
  # Keep only patient ID and the new GL string column.
  select(patient, GL_string)

# View the GL strings.
(HLA_typing_GL)

## -----------------------------------------------------------------------------
# Take the first patient's GL string and split it into locus columns.
# Note: GLstring_genes and GLstring_genes_expanded use pivot_longer on all
# columns, so only pass the GL string column (no other data types).
single_patient <- HLA_typing_GL[1, "GL_string", drop = FALSE]
GLstring_genes(single_patient, "GL_string")

## -----------------------------------------------------------------------------
GLstring_genes_expanded(single_patient, "GL_string")

## -----------------------------------------------------------------------------
# Patient 7 is the recipient, patient 9 is the donor.
recip_gl <- HLA_typing_GL %>% filter(patient == 7) %>% pull(GL_string)
donor_gl <- HLA_typing_GL %>% filter(patient == 9) %>% pull(GL_string)

## -----------------------------------------------------------------------------
# Check if there is an HLA-A mismatch in the graft-vs-host direction.
HLA_mismatch_logical(recip_gl, donor_gl, "HLA-A", direction = "GvH")

# Check host-vs-graft direction.
HLA_mismatch_logical(recip_gl, donor_gl, "HLA-A", direction = "HvG")

## -----------------------------------------------------------------------------
# Count bidirectional mismatches across several loci at once.
HLA_mismatch_number(
  recip_gl, donor_gl,
  c("HLA-A", "HLA-B", "HLA-C", "HLA-DRB1"),
  direction = "bidirectional"
)

## -----------------------------------------------------------------------------
# Identify the specific mismatched alleles in the HvG direction.
HLA_mismatched_alleles(recip_gl, donor_gl, "HLA-A", direction = "HvG")

## -----------------------------------------------------------------------------
# Count the number of matches (complement of mismatches).
HLA_match_number(
  recip_gl, donor_gl,
  c("HLA-A", "HLA-B", "HLA-C", "HLA-DRB1"),
  direction = "bidirectional"
)

## -----------------------------------------------------------------------------
# X-of-8 matching (A, B, C, DRB1 bidirectional).
HLA_match_summary_HCT(recip_gl, donor_gl,
  direction = "bidirectional",
  match_grade = "Xof8"
)

# X-of-10 matching (adds DQB1).
HLA_match_summary_HCT(recip_gl, donor_gl,
  direction = "bidirectional",
  match_grade = "Xof10"
)

## -----------------------------------------------------------------------------
# Patient 3 is the recipient; compare against all 10 donors.
recipient <- HLA_typing_GL %>%
  filter(patient == 3) %>%
  select(GL_string) %>%
  rename(GL_string_recip = GL_string)

donors <- HLA_typing_GL %>%
  rename(GL_string_donor = GL_string, donor = patient) %>%
  # Cross-join to pair recipient with each donor.
  cross_join(recipient) %>%
  # Calculate 8/8 match grade for each pair.
  mutate(
    match_8of8 = HLA_match_summary_HCT(
      GL_string_recip, GL_string_donor,
      direction = "bidirectional",
      match_grade = "Xof8"
    ),
    .after = donor
  ) %>%
  # Sort best matches first.
  arrange(desc(match_8of8))

donors %>% select(donor, match_8of8)

## -----------------------------------------------------------------------------
# Truncate a four-field allele to two fields.
HLA_truncate("HLA-A*02:01:01:01", fields = 2)

# Works on full GL strings too.
HLA_truncate("HLA-A*02:01:01:01+HLA-A*03:01:01:02^HLA-B*07:02:01:01+HLA-B*44:02:01:01",
  fields = 2
)

## -----------------------------------------------------------------------------
# Remove all prefixes to get just the allele fields.
HLA_prefix_remove("HLA-A*02:01")

# Keep the locus designation but remove "HLA-".
HLA_prefix_remove("HLA-A*02:01", keep_locus = TRUE)

# Add the full prefix back.
HLA_prefix_add("02:01", "HLA-A*")

# "HLA-" is added by default.
HLA_prefix_add("A*02:01")

## -----------------------------------------------------------------------------
gl <- "HLA-A*02:01:01+HLA-A*68:01^HLA-B*07:01+HLA-B*15:01"

# A two-field search correctly matches the three-field allele.
pattern <- GLstring_regex("HLA-A*02:01")
stringr::str_detect(gl, pattern)

# But won't falsely match a longer allele number.
stringr::str_detect("HLA-A*02:149:01", GLstring_regex("HLA-A*02:14"))

## -----------------------------------------------------------------------------
# GLstring_genes returns tidyverse-friendly names by default.
repaired <- GLstring_genes(single_patient, "GL_string")
names(repaired)

# Convert back to WHO format with asterisks.
who_names <- HLA_column_repair(repaired, format = "WHO", asterisk = TRUE)
names(who_names)

## -----------------------------------------------------------------------------
# immunogenetr ships with two example HML files.
hml_path <- system.file("extdata", "HML_1.hml", package = "immunogenetr")
hml_result <- read_HML(hml_path)
hml_result

#' @title GLstring_regex
#'
#' @description This function will format an HLA allele (e.g. "HLA-A*02:01") to
#' a regex pattern for searching within a GL string. Note that in order for this
#' function to work properly, the full HLA allele name, including prefixes, is
#' required. Allele values of "A*02:01" will need to be updated to "HLA-A*02:01",
#' and "A2" will need to be updated to "HLA-A2". The `HLA_prefix_add` function
#' is useful in these situations.
#'
#' @param data A string containing an HLA allele.
#'
#' @return A string with the HLA allele formatted as a regex pattern.
#'
#' @examples
#'
#' # To understand how the function works we can see how it alters the allele "HLA-A*02:01":
#'
#' GLstring_regex("HLA-A*02:01")
#'
#' # The result is the same allele with extra formatting to escape special characters found
#' # in a GL string, as well as the ability to accurately search for an allele in a GL string.
#' # For example, we would not want the allele "HLA-A*02:14" to match to "HLA-A*02:149:01",
#' # which would happen if we simply escaped the special characters:
#'
#' library(stringr)
#' str_view("HLA-A*02:149:01", str_escape("HLA-A*02:14"), match = NA)
#'
#' # Using `GLstring_regex` prevents this:
#'
#' str_view("HLA-A*02:149:01", GLstring_regex("HLA-A*02:14"), match = NA)
#'
#' # Using a longer GL string with multiple alleles and loci:
#'
#' GL_string <- "HLA-A*02:01:01+HLA-A*68:01^HLA-B*07:01+HLA-B*15:01"
#'
#' # We can match any allele accurately:
#'
#' str_view(GL_string, GLstring_regex("HLA-A*68:01"), match = NA)
#'
#' # Note that alleles supplied with fewer fields than in the GL string will also match:
#'
#' str_view(GL_string, GLstring_regex("HLA-A*02:01"), match = NA)
#'
#' @export
#'
#' @importFrom cli cli_abort
#' @importFrom stringr str_detect
#' @importFrom stringr str_c
#' @importFrom stringr str_escape
#' @importFrom tibble tibble
#' @importFrom dplyr mutate

GLstring_regex <- function(data) {
  # Validate input
  check_gl_string(data, "data")

  # This regex will only work if the full HLA allele, including the "HLA-" prefix is present. This will throw an error if "HLA-" is not detected in the allele to be modified.
  if (any(!str_detect(data, "HLA-"))) {
    cli_abort("All alleles must have the full {.val HLA-} prefix for {.fn GLstring_regex} to work properly. Use {.fn HLA_prefix_add} to add the prefix first.")
  }
  # Format the HLA alleles for searching within a GL string.
  table <- tibble(alleles = data) %>%
    mutate(alleles_regex = str_c(str_escape(alleles), "(?=(\\?|\\^|\\||\\+|\\~|/|:|$))")) # The regex includes all the delimiters in a GL string, a colon, which separates fields, and the end of a string.
  return(table$alleles_regex)
}

library(testthat)
library(stringr)

test_that("GLstring_regex correctly formats HLA alleles to regex patterns", {
  # Basic two-field alleles should be properly escaped with lookahead.
  expect_equal(GLstring_regex("HLA-A*02:01"), "HLA-A\\*02:01(?=(\\?|\\^|\\||\\+|\\~|/|:|$))")
  expect_equal(GLstring_regex("HLA-B*07:02"), "HLA-B\\*07:02(?=(\\?|\\^|\\||\\+|\\~|/|:|$))")
  # Multi-character locus name.
  expect_equal(GLstring_regex("HLA-DRB1*03:01"), "HLA-DRB1\\*03:01(?=(\\?|\\^|\\||\\+|\\~|/|:|$))")
})

test_that("GLstring_regex errors without HLA- prefix", {
  # Alleles without "HLA-" prefix should produce an error.
  expect_error(GLstring_regex("A*02:01"))
  expect_error(GLstring_regex("B27"))
})

test_that("GLstring_regex handles vector input", {
  # Should accept and process a vector of alleles.
  alleles <- c("HLA-A*02:01", "HLA-B*07:02", "HLA-C*04:01")
  result <- GLstring_regex(alleles)

  # Should return a vector of the same length.
  expect_length(result, 3)
  # Each element should be a valid regex string.
  expect_true(all(grepl("\\(\\?=", result)))
})

test_that("GLstring_regex produces patterns that match correctly in GL strings", {
  # Set up a GL string with multiple alleles.
  gl <- "HLA-A*02:01:01+HLA-A*68:01^HLA-B*07:01+HLA-B*15:01"

  # Two-field allele should match the four-field allele (fewer fields match more).
  pattern_a <- GLstring_regex("HLA-A*02:01")
  expect_true(str_detect(gl, pattern_a))

  # Full three-field allele should also match.
  pattern_a3 <- GLstring_regex("HLA-A*02:01:01")
  expect_true(str_detect(gl, pattern_a3))

  # A non-present allele should not match.
  pattern_missing <- GLstring_regex("HLA-A*03:01")
  expect_false(str_detect(gl, pattern_missing))
})

test_that("GLstring_regex prevents partial field matching", {
  # The regex should NOT match "HLA-A*02:149:01" when searching for "HLA-A*02:14".
  gl_with_long_allele <- "HLA-A*02:149:01"
  pattern <- GLstring_regex("HLA-A*02:14")
  expect_false(str_detect(gl_with_long_allele, pattern))

  # But "HLA-A*02:14" should match itself at end of string.
  expect_true(str_detect("HLA-A*02:14", pattern))
})

test_that("GLstring_regex matches alleles at GL string boundaries", {
  # Allele at the end of a GL string (followed by $).
  gl <- "HLA-A*01:01+HLA-A*02:01"
  pattern <- GLstring_regex("HLA-A*02:01")
  expect_true(str_detect(gl, pattern))

  # Allele before a "^" locus separator.
  gl2 <- "HLA-A*02:01^HLA-B*07:02"
  expect_true(str_detect(gl2, pattern))

  # Allele before a "+" gene copy separator.
  gl3 <- "HLA-A*02:01+HLA-A*03:01"
  expect_true(str_detect(gl3, pattern))

  # Allele before a "/" allele ambiguity separator.
  gl4 <- "HLA-A*02:01/HLA-A*02:06"
  expect_true(str_detect(gl4, pattern))

  # Allele before a "|" genotype ambiguity separator.
  gl5 <- "HLA-A*02:01|HLA-A*02:01+HLA-A*03:01"
  expect_true(str_detect(gl5, pattern))
})

test_that("GLstring_regex handles high-resolution and rare loci", {
  # Four-field allele should match itself in a GL string context.
  result_4f <- GLstring_regex("HLA-A*02:01:01:01")
  expect_true(str_detect("HLA-A*02:01:01:01+HLA-A*03:01", result_4f))

  # DQA1 locus.
  result_dqa <- GLstring_regex("HLA-DQA1*05:01")
  expect_true(str_detect("HLA-DQA1*05:01+HLA-DQA1*01:02", result_dqa))

  # DPB1 locus.
  result_dpb <- GLstring_regex("HLA-DPB1*04:01")
  expect_true(str_detect("HLA-DPB1*04:01+HLA-DPB1*02:01", result_dpb))
})

test_that("GLstring_regex validates inputs", {
  # NA should still work through check_gl_string (allows NA).
  # But should error because "HLA-" is not detected in NA.
  expect_error(GLstring_regex(NA))
  # Empty string should error (no "HLA-" prefix).
  expect_error(GLstring_regex(""))
  # Mixed valid/invalid should error.
  expect_error(GLstring_regex(c("HLA-A*02:01", "A*03:01")))
})

#' @title HLA_mismatch_base
#'
#' @description A function to return a string of mismatches between recipient
#' and donor HLA genotypes represented as GL strings. The function finds
#' mismatches based on the direction of comparison specified in the inputs
#' and also handles homozygosity. As the name implies, this function is the base
#' for all other mismatch (and matching) functions. This function is not meant to be
#' called directly; it is better to use one of the derivative functions.
#'
#' @param GL_string_recip A GL string representing the recipient's HLA genotype.
#' @param GL_string_donor A GL string representing the donor's HLA genotype.
#' @param loci A character vector specifying the loci to be considered for
#' mismatch calculation. HLA-DRB3/4/5 (and their serologic equivalents DR51/52/53)
#' are considered once locus for this function, and should be called in this argument
#' as "HLA-DRB3/4/5" or "HLA-DR51/52/53", respectively.
#' @param direction A character string indicating the direction of mismatch.
#' Options are "HvG" (host vs. graft) or "GvH" (graft vs. host).
#' @param homozygous_count An integer specifying how to handle homozygosity.
#' Defaults to 2, where homozygous alleles are treated as duplicated for
#' mismatch calculations. Can be specified to be 1, in which case homozygous
#' alleles are treated as single occurrences without duplication.
#'
#' @return A character vector, where each element is a string summarizing the
#' mismatches for the specified loci. The strings are formatted as
#' comma-separated locus mismatch entries if multiple loci are supplied, or
#' simple GL strings if a single locus is supplied.

#'
#' @examples
#' file <- HLA_typing_1[, -1]
#' GL_string <- HLA_columns_to_GLstring(file, HLA_typing_columns = everything())
#'
#' GL_string_recip <- GL_string[1]
#' GL_string_donor <- GL_string[2]
#'
#' loci <- c("HLA-A", "HLA-DRB3/4/5", "HLA-DPB1")
#' mismatches <- HLA_mismatch_base(GL_string_recip, GL_string_donor, loci, direction = "HvG")
#' print(mismatches)
#'
#' @export
#'
#' @importFrom cli cli_abort
#' @importFrom stringr str_split
#' @importFrom stringr str_flatten
#' @importFrom stringr str_detect
#' @importFrom stringr str_replace
#' @importFrom stringr str_c
#' @importFrom purrr keep
#' @importFrom purrr discard
#' @importFrom purrr map
#' @importFrom purrr map_chr
#' @importFrom purrr map2_chr
#' @importFrom purrr modify_if
#'

HLA_mismatch_base <- function(GL_string_recip, GL_string_donor, loci, direction, homozygous_count = 2) {
  # Validate inputs
  check_gl_string(GL_string_recip, "GL_string_recip")
  check_gl_string(GL_string_donor, "GL_string_donor")
  check_loci(loci)
  check_homozygous_count(homozygous_count)

  direction <- match.arg(direction, c("HvG", "GvH"))
  # Ensure input vectors are of the same length - each input should be a single GL string.
  if (length(GL_string_recip) != length(GL_string_donor)) {
    cli_abort("{.arg GL_string_recip} and {.arg GL_string_donor} must be of equal length.")
  }

  # Check for ambiguity
  if (any(str_detect(GL_string_recip, "[|/]") | str_detect(GL_string_donor, "[|/]"))) {
    cli_abort("The matching/mismatching functions do not support ambiguous GL strings containing {.val |} or {.val /}. Process your GL strings to result in unambiguous genotypes before using these functions.")
  }

  # Maps the serologic naming of the DRB locus to molecular so that only one name is used
  unify_locus <- function(x) stringr::str_replace(x, "^HLA-DR51/52/53$", "HLA-DRB3/4/5")
  # Keep original locus names for display later and normalize loci
  original_loci <- loci
  loci <- unify_locus(loci)
  # Maps internal keys back to the user's original loci inputs for display
  display_name <- function(name) {
    idx <- match(name, loci)
    if (is.na(idx)) name else original_loci[[idx]]
  }

  # Function to preprocess GL strings: handle null alleles and homozygosity
  preprocess_GL_string <- function(GL_string, homozygous_count) {
    # Split GL string into alleles
    alleles <- str_split(GL_string, "\\+", simplify = TRUE)

    # Replace any null allele (ending with uppercase "N" expression suffix) with a
    # placeholder "XXN". The WHO nomenclature defines "N" as the null expression suffix,
    # appearing after the colon-separated allele fields (e.g. "HLA-A*01:01N").
    # The lookbehind matches the locus prefix up to the asterisk, allowing locus names
    # of any length (e.g. HLA-A, HLA-DRB1, HLA-DRB345).
    alleles <- str_replace(alleles, "(?<=HLA-[:alnum:]{1,10}\\*).+N$", "XXN")

    # Handle homozygosity
    if (length(alleles) == 1 && homozygous_count == 2) {
      alleles <- rep(alleles, times = homozygous_count)
    } else if (length(alleles) == 2 && alleles[1] == alleles[2] && homozygous_count == 1) {
      alleles <- alleles[1]
    }

    # Return processed alleles as a single GL string
    return(str_flatten(alleles, collapse = "+", na.rm = TRUE))
  }

  # Function to process each pair of recipient/donor alleles
  process_pair <- function(recip_str, donor_str) {
    # Split GL strings by "^" to separate loci
    recip_alleles_list <- unlist(strsplit(recip_str, "\\^"))
    donor_alleles_list <- unlist(strsplit(donor_str, "\\^"))

    # Function to process the alleles strings.
    process_alleles <- function(alleles_list, homozygous_count) {
      # Process DRB3/4/5 or DR51/52/53 alleles to add them to a single locus "HLA-DRB3/4/5"
      alleles_list_DRB345 <- alleles_list %>%
        keep(str_detect(., "(HLA-DRB[345])|(HLA-DR5[123])")) %>%
        paste(collapse = "+") %>%
        # If no DRB3/4/5 alleles are present, use a null placeholder, so that expressed alleles
        # at this locus will be output as mismatches.
        { if (. == "") "HLA-DRB3*XX:XXN" else . }


      # Remove DRB3/4/5 or DR51/52/53 alleles from the original list
      alleles_list_no_DRB345 <- alleles_list %>%
        keep(!str_detect(., "(HLA-DRB[345])|(HLA-DR5[123])"))

      # Combine lists and preprocess each allele string
      alleles_list_processed <- c(alleles_list_no_DRB345, alleles_list_DRB345) %>%
        keep(~ str_detect(.x, "[:graph:]")) %>%
        map_chr(preprocess_GL_string, homozygous_count)

      # Filter out any NA values
      alleles_list_processed <- alleles_list_processed[!is.na(alleles_list_processed)]

      return(alleles_list_processed)
    }

    # Process recipient and donor allele strings:
    recip_alleles_list_processed <- process_alleles(recip_alleles_list, homozygous_count)
    donor_alleles_list_processed <- process_alleles(donor_alleles_list, homozygous_count)

    # Helper to extract the locus name from a GL string allele entry
    extract_locus_name <- function(allele_str) {
      case_when(
        # Molecular nomenclature
        str_detect(allele_str, "\\*") & str_detect(allele_str, "HLA-DRB[345]") ~ "HLA-DRB3/4/5",
        str_detect(allele_str, "\\*") ~ strsplit(allele_str, "\\*") %>% map_chr(1),
        # Serologic nomenclature
        !str_detect(allele_str, "\\*") & str_detect(allele_str, "HLA-DR5[123]") ~ "HLA-DR51/52/53",
        !str_detect(allele_str, "\\*") ~ str_extract(allele_str, "^HLA-[A-Za-z]+")
      )
    }

    # Set names for each locus using the shared helper
    names(recip_alleles_list_processed) <- extract_locus_name(recip_alleles_list_processed)
    names(donor_alleles_list_processed) <- extract_locus_name(donor_alleles_list_processed)

    # Applies unify_locus function to map the serologic naming of the DRB locus to molecular
    names(recip_alleles_list_processed) <- unify_locus(names(recip_alleles_list_processed))
    names(donor_alleles_list_processed) <- unify_locus(names(donor_alleles_list_processed))

    # Find which supplied loci are missing from recipient and donor genotypes
    missing_loci_from_recipient <- setdiff(loci, names(recip_alleles_list_processed))
    missing_loci_from_donor <- setdiff(loci, names(donor_alleles_list_processed))
    # Exclude DRB3/4/5 from missing locus checks because it is optional
    missing_loci <- setdiff(union(missing_loci_from_recipient, missing_loci_from_donor), "HLA-DRB3/4/5")

    if (length(missing_loci) > 0) {
      cli_abort("The recipient and/or donor GL strings are missing these loci: {.val {missing_loci}}.")
    }

    # Mismatch results calculation
    mismatch_results <- map(loci, function(locus_name) {

      # Defensive guard: if either donor or recip lacks DRB3/4/5, return "=NA"
      # for that locus instead of error. Currently unreachable because
      # process_alleles() inserts a null placeholder for missing DRB3/4/5,
      # but kept as a safety net in case that logic changes.
      if (locus_name == "HLA-DRB3/4/5" &&
          (!(locus_name %in% names(recip_alleles_list_processed)) ||
           !(locus_name %in% names(donor_alleles_list_processed)))) {
        return(paste0(display_name(locus_name), "=NA"))
      }

      # Pull out the allele list for each locus and split once for reuse.
      recip_alleles_str <- recip_alleles_list_processed[locus_name]
      donor_alleles_str <- donor_alleles_list_processed[locus_name]
      recip_alleles <- unlist(strsplit(recip_alleles_str, "\\+"))
      donor_alleles <- unlist(strsplit(donor_alleles_str, "\\+"))

      # Assign sides based on direction:
      # - "match_from": the side we index [1],[2] for match checking
      # - "mismatch_from": the side whose valid alleles we check for mismatches
      # GvH: matches from donor side, mismatches from recip side
      # HvG: matches from recip side, mismatches from donor side
      if (direction == "GvH") {
        match_from <- donor_alleles
        mismatch_from <- recip_alleles
      } else {
        match_from <- recip_alleles
        mismatch_from <- donor_alleles
      }

      # Calculate matches (including nulls): match_from alleles found in mismatch_from
      matched_allele_1 <- intersect(match_from[1], mismatch_from)
      matched_allele_2 <- intersect(match_from[2], mismatch_from)
      matched_alleles <- discard(c(matched_allele_1, matched_allele_2), is.na)
      # Calculate mismatches (excluding nulls): mismatch_from alleles NOT in match_from
      mismatch_valid <- mismatch_from[!str_detect(mismatch_from, "[Nn]$")]
      mismatched_allele_1 <- setdiff(mismatch_valid[1], match_from)
      mismatched_allele_2 <- setdiff(mismatch_valid[2], match_from)
      mismatched_alleles <- discard(c(mismatched_allele_1, mismatched_allele_2), is.na)
      # Count number of matches and mismatches
      total_match_mismatch <- length(matched_alleles) + length(mismatched_alleles)
      # If total matches + mismatches < 2, and homozygous_count == 2, repeat mismatched alleles.
      if (total_match_mismatch < 2 && homozygous_count == 2) {
        mismatched_alleles <- rep(mismatched_alleles, times = homozygous_count)
      }

      # Create a string of mismatched alleles or 'NA' if no mismatches are found.
      # Use the user's original label through the display_name function
      allele_mismatches_str <-
        if (length(mismatched_alleles) > 0) {
          paste0(display_name(locus_name), "=", paste(mismatched_alleles, collapse = "+"))
        } else {
          paste0(display_name(locus_name), "=", "NA")
        }
    })

    result <- paste(mismatch_results, collapse = ", ")

    # If only a single locus was selected in the arguments, output without starting with the locus name followed by an equal sign.
    if (length(loci) == 1) {
      result %>%
        str_replace(str_c(original_loci, "="), "") %>%
        na_if("NA")
    } else {
      return(result)
    }
  }

  # Return final result by applying the GL strings to the process_pair function defined above.
  map2_chr(GL_string_recip, GL_string_donor, process_pair)
}

globalVariables(c(".", "process_alleles"))

library(testthat)
library(stringr)
library(dplyr)

test_that("HLA_mismatch_base correctly identifies mismatches", {
  GL_string_recip <- "HLA-A*02:01+HLA-A*03:01^HLA-B*07:02+HLA-B*08:01^HLA-DRB1*04:01+HLA-DRB1*07:01"
  GL_string_donor <- "HLA-A*01:01+HLA-A*03:01^HLA-B*07:02+HLA-B*44:02^HLA-DRB1*04:01+HLA-DRB1*13:01"
  loci <- c("HLA-A", "HLA-B", "HLA-DRB1")

  result_HvG <- HLA_mismatch_base(GL_string_recip, GL_string_donor, loci, direction = "HvG")
  expect_equal(result_HvG, "HLA-A=HLA-A*01:01, HLA-B=HLA-B*44:02, HLA-DRB1=HLA-DRB1*13:01")

  result_GvH <- HLA_mismatch_base(GL_string_recip, GL_string_donor, loci, direction = "GvH")
  expect_equal(result_GvH, "HLA-A=HLA-A*02:01, HLA-B=HLA-B*08:01, HLA-DRB1=HLA-DRB1*07:01")

  result_single_locus_HvG <- HLA_mismatch_base(GL_string_recip, GL_string_donor, "HLA-A", direction = "HvG")
  expect_equal(result_single_locus_HvG, "HLA-A*01:01")

  result_single_locus_GvH <- HLA_mismatch_base(GL_string_recip, GL_string_donor, "HLA-A", direction = "GvH")
  expect_equal(result_single_locus_GvH, "HLA-A*02:01")

  result_HvG_homo1 <- HLA_mismatch_base(GL_string_recip, GL_string_donor, loci, direction = "HvG", homozygous_count = 1)
  expect_equal(result_HvG_homo1, "HLA-A=HLA-A*01:01, HLA-B=HLA-B*44:02, HLA-DRB1=HLA-DRB1*13:01")

  expect_error(
    HLA_mismatch_base(c(GL_string_recip, GL_string_recip), GL_string_donor, loci, direction = "HvG"),
    "must be of equal length"
  )
})


test_that("DRB3/4/5 and DR51/52/53 absent on both sides", {
  recip <- "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01"
  donor <- "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01"

  loci_mol <- c("HLA-A", "HLA-DRB3/4/5")
  out_mol <- HLA_mismatch_base(recip, donor, loci_mol, direction = "HvG")
  expect_equal(out_mol, "HLA-A=NA, HLA-DRB3/4/5=NA")

  loci_ser <- c("HLA-A", "HLA-DR51/52/53")
  out_ser <- HLA_mismatch_base(recip, donor, loci_ser, direction = "HvG")
  expect_equal(out_ser, "HLA-A=NA, HLA-DR51/52/53=NA")
})


test_that("DRB3/4/5 is in one but not in the other", {
  recip <- "HLA-A*01:01+HLA-A*02:01^HLA-DRB3*01:01^HLA-DRB4*01:01"
  donor <- "HLA-A*01:01+HLA-A*02:01^HLA-DRB1*15:01+HLA-DRB1*07:01"

  out_mol <- HLA_mismatch_base(recip, donor, c("HLA-A", "HLA-DRB3/4/5"), direction = "HvG")
  expect_equal(out_mol, "HLA-A=NA, HLA-DRB3/4/5=NA")

  out_ser <- HLA_mismatch_base(recip, donor, c("HLA-A", "HLA-DR51/52/53"), direction = "HvG")
  expect_equal(out_ser, "HLA-A=NA, HLA-DR51/52/53=NA")
})


test_that("DRB51/52/53 is in one but not in the other", {
  recip <- "HLA-A*01:01+HLA-A*02:01^HLA-DR52^HLA-DR53"
  donor <- "HLA-A*01:01+HLA-A*02:01^HLA-DRB1*15:01+HLA-DRB1*07:01"

  out_ser <- HLA_mismatch_base(recip, donor, c("HLA-A", "HLA-DR51/52/53"), direction = "HvG")
  expect_equal(out_ser, "HLA-A=NA, HLA-DR51/52/53=NA")

  out_mol <- HLA_mismatch_base(recip, donor, c("HLA-A", "HLA-DRB3/4/5"), direction = "HvG")
  expect_equal(out_mol, "HLA-A=NA, HLA-DRB3/4/5=NA")
})


test_that("DRB3/4/5 or DR51/52/53 as the only tested allele", {
  recip <- "HLA-A*01:01+HLA-A*02:01^HLA-DR52^HLA-DR53"
  donor <- "HLA-A*01:01+HLA-A*02:01^HLA-DRB1*15:01+HLA-DRB1*07:01"

  out_mol <- HLA_mismatch_base(recip, donor, "HLA-DRB3/4/5", direction = "HvG")
  expect_true(is.na(out_mol))

  out_ser <- HLA_mismatch_base(recip, donor, "HLA-DR51/52/53", direction = "HvG")
  expect_true(is.na(out_ser))
})


test_that("Matching across DRB3/4/5 (molecular) and DR51/52/53 (serologic)", {
  recip <- "HLA-A*01:01+HLA-A*02:01^HLA-DRB3*01:01^HLA-DRB4*01:01"
  donor <- "HLA-A*01:01+HLA-A*02:01^HLA-DR52^HLA-DR53"

  out_mol <- HLA_mismatch_base(recip, donor, c("HLA-A", "HLA-DRB3/4/5"), direction = "HvG")
  expect_equal(out_mol, "HLA-A=NA, HLA-DRB3/4/5=HLA-DR52+HLA-DR53")

  out_ser <- HLA_mismatch_base(recip, donor, c("HLA-A", "HLA-DR51/52/53"), direction = "HvG")
  expect_equal(out_ser, "HLA-A=NA, HLA-DR51/52/53=HLA-DR52+HLA-DR53")
})

test_that("A genotype without any DRB3/4/5 alleles compared to a genotype with will have a mismatch", {
  recip <- "HLA-DRB1*01:01^HLA-DRB3*01:01^HLA-DRB4*01:01"
  donor <- "HLA-DRB1*01:01^HLA-B*15:01"

  mismatches <- HLA_mismatch_base(recip, donor, loci = "HLA-DRB3/4/5", direction = "GvH")
  expect_equal(mismatches, "HLA-DRB3*01:01+HLA-DRB4*01:01")

  no_mismatches <- HLA_mismatch_base(recip, donor, loci = "HLA-DRB3/4/5", direction = "HvG")
  expect_equal(no_mismatches, NA_character_)

  expect_error(
    HLA_mismatch_base(recip, donor, loci = "HLA-B", direction = "HvG"),
    "missing these loci"
  )

  recip <- "HLA-DR1^HLA-DR52^HLA-DRB53"
  donor <- "HLA-DR1^HLA-B*75"

  sero_MM <- HLA_mismatch_base(recip, donor, loci = "HLA-DR51/52/53", direction = "GvH")
  expect_equal(sero_MM, "HLA-DR52+HLA-DRB53")

  sero_NA <- HLA_mismatch_base(recip, donor, loci = "HLA-DR51/52/53", direction = "HvG") # NA
  expect_equal(sero_NA, NA_character_)
})


# --- Input validation tests ---

test_that("HLA_mismatch_base rejects NULL inputs", {
  gl <- "HLA-A*01:01+HLA-A*02:01"
  expect_error(HLA_mismatch_base(NULL, gl, "HLA-A", "HvG"), "GL_string_recip")
  expect_error(HLA_mismatch_base(gl, NULL, "HLA-A", "HvG"), "GL_string_donor")
  expect_error(HLA_mismatch_base(gl, gl, NULL, "HvG"), "loci")
})

test_that("HLA_mismatch_base rejects non-character GL strings", {
  gl <- "HLA-A*01:01+HLA-A*02:01"
  expect_error(HLA_mismatch_base(123, gl, "HLA-A", "HvG"), "must be a character")
  expect_error(HLA_mismatch_base(gl, 123, "HLA-A", "HvG"), "must be a character")
})

test_that("HLA_mismatch_base rejects invalid direction", {
  gl <- "HLA-A*01:01+HLA-A*02:01"
  expect_error(HLA_mismatch_base(gl, gl, "HLA-A", "invalid"))
})

test_that("HLA_mismatch_base rejects ambiguous GL strings", {
  gl_ambig <- "HLA-A*01:01/HLA-A*01:02+HLA-A*02:01"
  gl_normal <- "HLA-A*01:01+HLA-A*02:01"
  expect_error(HLA_mismatch_base(gl_ambig, gl_normal, "HLA-A", "HvG"), "ambiguous")
})

test_that("HLA_mismatch_base works with vectorized inputs", {
  recip <- c(
    "HLA-A*01:01+HLA-A*02:01",
    "HLA-A*03:01+HLA-A*24:02"
  )
  donor <- c(
    "HLA-A*01:01+HLA-A*03:01",
    "HLA-A*03:01+HLA-A*24:02"
  )
  # Should return a vector of length 2

  result <- HLA_mismatch_base(recip, donor, "HLA-A", "HvG")
  expect_length(result, 2)
  # First pair has a mismatch, second pair is a perfect match
  expect_equal(result[1], "HLA-A*03:01")
  expect_true(is.na(result[2]))
})

test_that("HLA_mismatch_base returns NA for perfect match at single locus", {
  gl <- "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01"
  result <- HLA_mismatch_base(gl, gl, "HLA-A", "HvG")
  expect_true(is.na(result))
})

#' Data on HLA alleles in the 2008 HLA dictionary
#'
#' @docType data
#'
#' @usage data(HLA_dictionary)
#'
#' @format
#' A tibble:
#' \describe{
#'   \item{IPD.Accession}{IPD-IMGT/HLA Accession number}
#'   \item{HLA.Allele}{HLA allele name}
#'   \item{WHO.Assigned.Type}{WHO-assigned serologic equivalent}
#'   ...
#' }
#'
#' @keywords datasets
#'
#' @source <https://doi.org/10.1111/j.1399-0039.2008.01183.x>
#'
"HLA_dictionary"

#' @title GLstring_gene_copies_combine
#'
#' @description A function for combining two columns of typing from the same
#' locus into a single column in the appropriate GL string format.
#'
#' @param .data A data frame
#' @param columns The names of the columns in the data frame that contain typing
#' information to be combined
#' @param sample_column The name of the column that identifies samples in the
#' data frame. Default is "sample".
#'
#' @return A data frame with the specified columns combined into a single column
#' for each locus, in GL string format.
#'
#' @examples
#' library(dplyr)
#' HLA_typing_1 %>%
#'   mutate(across(A1:B2, ~ HLA_prefix_add(.))) %>%
#'   GLstring_gene_copies_combine(c(A1:B2), sample_column = patient)
#'
#' @export
#'
#' @importFrom dplyr select
#' @importFrom dplyr mutate
#' @importFrom dplyr filter
#' @importFrom dplyr summarise
#' @importFrom dplyr rename_with
#' @importFrom dplyr %>%
#' @importFrom tidyr pivot_longer
#' @importFrom tidyr pivot_wider
#' @importFrom stringr str_extract
#' @importFrom stringr str_c
#' @importFrom stringr str_replace


GLstring_gene_copies_combine <- function(.data, columns, sample_column = "sample") {
  # Validate input
  check_data_frame(.data, ".data")

  # Identify the columns to modify
  cols2mod <- names(select(.data, {{ columns }}))

  .data %>%
    pivot_longer(all_of(cols2mod), names_to = "locus", values_to = "allele") %>%
    mutate(locus = str_extract(allele, "HLA-[:alnum:]+")) %>%
    filter(!is.na(locus)) %>%
    summarise(allele = str_c(allele, collapse = "+"), .by = c({{ sample_column }}, locus)) %>%
    pivot_wider(names_from = locus, values_from = allele) %>%
    rename_with(~ str_replace(., "HLA\\-", "HLA_"))
}

globalVariables(c("locus", "allele"))

library(testthat)
library(dplyr)
library(tidyr)
library(stringr)

test_that("GLstring_gene_copies_combine correctly combines HLA gene copies", {
  HLA_type <- data.frame(
    sample = c("sample1", "sample2"),
    HLA_A1 = c("HLA-A*01:01", "HLA-A*02:01"),
    HLA_A2 = c("HLA-A*01:02", "HLA-A*02:02"),
    stringsAsFactors = FALSE
  )
  result <- GLstring_gene_copies_combine(HLA_type, columns = c("HLA_A1", "HLA_A2"))

  expect_s3_class(result, "data.frame")
  expect_true("HLA_A" %in% colnames(result))
  expect_equal(nrow(result), 2)

  expect_match(result$HLA_A[1], "HLA-A\\*01:01\\+HLA-A\\*01:02")
  expect_match(result$HLA_A[2], "HLA-A\\*02:01\\+HLA-A\\*02:02")
})

test_that("GLstring_gene_copies_combine handles missing values correctly", {
  HLA_type <- data.frame(
    sample = c("sample1", "sample2"),
    HLA_A1 = c("HLA-A*01:01", NA),
    HLA_A2 = c("HLA-A*01:02", "HLA-A*02:02"),
    stringsAsFactors = FALSE
  )

  result <- GLstring_gene_copies_combine(HLA_type, columns = c("HLA_A1", "HLA_A2"))

  expect_s3_class(result, "data.frame")
  expect_true("HLA_A" %in% colnames(result))
  expect_equal(nrow(result), 2)

  expect_match(result$HLA_A[1], "HLA-A\\*01:01\\+HLA-A\\*01:02")
  expect_match(result$HLA_A[2], "HLA-A\\*02:02")
})

test_that("GLstring_gene_copies_combine works with multiple loci", {
  HLA_type <- data.frame(
    sample = c("sample1", "sample2"),
    HLA_A1 = c("HLA-A*01:01", "HLA-A*02:01"),
    HLA_A2 = c("HLA-A*01:02", "HLA-A*02:02"),
    HLA_B1 = c("HLA-B*07:01", "HLA-B*08:01"),
    HLA_B2 = c("HLA-B*07:02", "HLA-B*08:02"),
    stringsAsFactors = FALSE
  )

  result <- GLstring_gene_copies_combine(HLA_type, columns = c("HLA_A1", "HLA_A2", "HLA_B1", "HLA_B2"))

  expect_s3_class(result, "data.frame")
  expect_true(all(c("HLA_A", "HLA_B") %in% colnames(result)))
  expect_equal(nrow(result), 2)

  # Check correct GL string combination for multiple loci
  expect_match(result$HLA_A[1], "HLA-A\\*01:01\\+HLA-A\\*01:02")
  expect_match(result$HLA_A[2], "HLA-A\\*02:01\\+HLA-A\\*02:02")
  expect_match(result$HLA_B[1], "HLA-B\\*07:01\\+HLA-B\\*07:02")
  expect_match(result$HLA_B[2], "HLA-B\\*08:01\\+HLA-B\\*08:02")
})

#' @title ambiguity_table_to_GLstring
#'
#' @description A function that converts a data table of HLA allele ambiguities
#' (e.g. as created by `GLstring_expand_longer` or `GLstring_to_ambiguity_table`)
#' into a GL string format. The function processes the table by combining allele
#' ambiguities, haplotypes, gene copies, and loci into a structured GL string.
#'
#' @param data A data frame containing columns that represent possible gene
#' locations, loci, genotype ambiguities, genotypes, and haplotypes.
#'
#' @param remove_duplicates A logical value indicating if the function will check
#' for duplicate entries at each step and remove them before assembling the final
#' GL string. Useful if the ambiguity table has been altered, for example by
#' truncating allele designations. Default is FALSE.
#'
#' @return A GL string representing the combined gene locations, loci, genotype
#' ambiguities, genotypes, and haplotypes.
#'
#' @examples
#' # Example data frame input
#' data <- tibble::tribble(
#'   ~value, ~entry, ~possible_gene_location,
#'   ~locus, ~genotype_ambiguity, ~genotype, ~haplotype, ~allele,
#'   "HLA-A*01:01:01:01", 1, 1,
#'   1, 1, 1, 1, 1,
#'   "HLA-A*01:02", 1, 1,
#'   1, 1, 1, 1, 2,
#'   "HLA-A*01:03", 1, 1,
#'   1, 1, 1, 1, 3,
#'   "HLA-A*01:95", 1, 1,
#'   1, 1, 1, 1, 4,
#'   "HLA-A*24:02:01:01", 1, 1,
#'   1, 1, 2, 1, 1,
#'   "HLA-A*01:01:01:01", 1, 1,
#'   1, 2, 1, 1, 1,
#'   "HLA-A*01:03", 1, 1,
#'   1, 2, 1, 1, 2,
#'   "HLA-A*24:03:01:01", 1, 1,
#'   1, 2, 2, 1, 1,
#'   "HLA-B*07:01:01", 1, 1,
#'   2, 1, 1, 1, 1,
#'   "B*15:01:01", 1, 1,
#'   2, 1, 2, 1, 1,
#'   "B*15:02:01", 1, 1,
#'   2, 1, 2, 1, 2,
#'   "B*07:03", 1, 1,
#'   2, 2, 1, 1, 1,
#'   "B*15:99:01", 1, 1,
#'   2, 2, 2, 1, 1,
#'   "HLA-DRB1*03:01:02", 1, 1,
#'   3, 1, 1, 1, 1,
#'   "HLA-DRB5*01:01:01", 1, 1,
#'   3, 1, 1, 2, 1,
#'   "HLA-KIR2DL5A*0010101", 1, 1,
#'   3, 1, 2, 1, 1,
#'   "HLA-KIR2DL5A*0010201", 1, 1,
#'   3, 1, 3, 1, 1,
#'   "HLA-KIR2DL5B*0010201", 1, 2,
#'   1, 1, 1, 1, 1,
#'   "HLA-KIR2DL5B*0010301", 1, 2,
#'   1, 1, 2, 1, 1
#' )
#'
#' ambiguity_table_to_GLstring(data)
#'
#' @export
#'
#' @importFrom dplyr summarise
#' @importFrom dplyr pull
#' @importFrom dplyr %>%
#' @importFrom stringr str_flatten

ambiguity_table_to_GLstring <- function(data, remove_duplicates = FALSE) {
  # Validate inputs
  check_data_frame(data, "data")
  check_logical_flag(remove_duplicates, "remove_duplicates")

  # Helper to optionally deduplicate then collapse values at each GL string level.
  # Each level of the GL string hierarchy (allele, haplotype, genotype, etc.) follows
  # the same pattern: optionally remove duplicates, then summarise with a separator.
  collapse_level <- function(data, by_cols, collapse_char) {
    if (remove_duplicates) {
      data <- distinct(data, across(all_of(c("value", by_cols))), .keep_all = TRUE)
    }
    summarise(data, value = str_flatten(value, collapse = collapse_char), .by = all_of(by_cols))
  }

  # Collapse each level of the GL string hierarchy from most specific to least.
  # GL string separators: "/" (allele ambiguity), "~" (haplotype), "+" (gene copies),
  # "|" (genotype ambiguity), "^" (locus), "?" (possible gene location).
  data %>%
    collapse_level(c("entry", "possible_gene_location", "locus", "genotype_ambiguity", "genotype", "haplotype"), "/") %>%
    collapse_level(c("entry", "possible_gene_location", "locus", "genotype_ambiguity", "genotype"), "~") %>%
    collapse_level(c("entry", "possible_gene_location", "locus", "genotype_ambiguity"), "+") %>%
    collapse_level(c("entry", "possible_gene_location", "locus"), "|") %>%
    collapse_level(c("entry", "possible_gene_location"), "^") %>%
    collapse_level(c("entry"), "?") %>%
    pull(value)
}

globalVariables(c("."))

#' @title read_HML
#'
#' @description Reads the GL strings of HML files and returns a tibble with
#' the full genotype for each sample.
#'
#' @param HML_file The path to an HML file.
#'
#' @return A tibble with the sample name and the GL string.
#'
#' @examples
#' HML_1 <- system.file("extdata", "HML_1.hml", package = "immunogenetr")
#' HML_2 <- system.file("extdata", "hml_2.hml", package = "immunogenetr")
#'
#' read_HML(HML_1)
#' read_HML(HML_2)
#'
#' @export
#'
#' @importFrom cli cli_abort
#' @importFrom xml2 read_xml
#' @importFrom dplyr %>%
#' @importFrom xml2 xml_find_all
#' @importFrom xml2 xml_ns
#' @importFrom purrr map
#' @importFrom xml2 xml_attr
#' @importFrom xml2 xml_find_all
#' @importFrom xml2 xml_text
#' @importFrom tibble tibble
#' @importFrom dplyr bind_rows
#' @importFrom dplyr mutate
#' @importFrom stringr str_extract
#' @importFrom dplyr distinct
#' @importFrom tidyr pivot_wider
#' @importFrom tidyr unite

read_HML <- function(HML_file) {
  # Validate input
  check_gl_string(HML_file, "HML_file")
  if (!file.exists(HML_file)) {
    cli_abort("The file {.file {HML_file}} does not exist.")
  }

  # Load the HML file
  HML <- tryCatch(
    {
      read_xml(HML_file)
    },
    error = function(e) {
      cli_abort("Failed to read HML file {.file {HML_file}}. Check that file is in compliant HML format.")
    }
  )

  # Discover the XML namespace dynamically instead of hardcoding "d1:".
  # xml2 auto-registers default (unprefixed) namespaces as "d1", "d2", etc.,
  # but this can vary depending on the file structure.
  ns <- xml_ns(HML)

  # Build XPath queries using the discovered namespace prefix.
  # If a namespace is present, use its prefix; otherwise use unqualified names.
  if (length(ns) > 0) {
    # Use the first registered namespace prefix (typically "d1" for default namespace).
    ns_prefix <- names(ns)[1]
    sample_xpath <- paste0(".//", ns_prefix, ":sample")
    glstring_xpath <- paste0(".//", ns_prefix, ":glstring")
  } else {
    # No namespace declared; use unqualified element names.
    sample_xpath <- ".//sample"
    glstring_xpath <- ".//glstring"
  }

  # Filter for all the children in the HML file that represent a sample.
  samples <- xml_find_all(HML, sample_xpath, ns)

  # Get sample number and GL strings for each sample.
  GL_strings <- map(samples, function(node) {
    # Get sample ID.
    sampleID <- xml_attr(node, "id")
    # Get GL strings.
    glstring <- xml_text(xml_find_all(node, glstring_xpath, ns))
    # Combine to a tibble.
    tibble(sampleID, glstring)
  })

  # Combine to a single tibble.
  combined <- bind_rows(GL_strings)

  # Some implementations of HML put the same locus in multiple nodes; this combines them with "+" to form a compliant GL string
  reduced <- combined %>%
    mutate(locus = str_extract(glstring, "[^//*]+")) %>%
    mutate(glstring = paste0(glstring, collapse = "+"), .by = c(sampleID, locus)) %>%
    # Clean up values
    distinct(sampleID, glstring, locus) %>%
    filter(!is.na(sampleID) & !is.na(glstring)) %>%
    select(-locus)

  # Combine to a single GL string per sample
  summarise(reduced, GL_string = str_flatten(glstring, collapse = "^"), .by = sampleID)
}

globalVariables(c("glstring", "sampleID"))

library(testthat)
library(dplyr)
library(stringr)
library(xml2)
library(tibble)
library(purrr)
library(tidyr)

test_that("read_HML correctly extracts GL strings from HML_1.hml", {
  # Load the first test HML file bundled with the package.
  HML_1 <- system.file("extdata", "HML_1.hml", package = "immunogenetr")

  if (file.exists(HML_1)) {
    result <- read_HML(HML_1)

    # Should return a tibble.
    expect_s3_class(result, "tbl_df")
    # Should have the expected columns.
    expect_true("sampleID" %in% colnames(result))
    expect_true("GL_string" %in% colnames(result))
    # Should extract rows.
    expect_gt(nrow(result), 0)
    # All GL strings should contain "HLA-" prefix.
    expect_true(all(str_detect(result$GL_string, "HLA-")))
    # HML_1 has 5 samples.
    expect_equal(nrow(result), 5)
    # Sample IDs should be present and non-NA.
    expect_false(any(is.na(result$sampleID)))
  } else {
    skip("HML_1.hml test file does not exist.")
  }
})

test_that("read_HML correctly extracts GL strings from hml_2.hml", {
  # Load the second test HML file with a slightly different format.
  HML_2 <- system.file("extdata", "hml_2.hml", package = "immunogenetr")

  if (file.exists(HML_2)) {
    result <- read_HML(HML_2)

    # Should return a tibble with the same structure.
    expect_s3_class(result, "tbl_df")
    expect_true(all(c("sampleID", "GL_string") %in% colnames(result)))
    expect_gt(nrow(result), 0)
    # hml_2 also has 5 samples.
    expect_equal(nrow(result), 5)
    # All GL strings should contain "HLA-".
    expect_true(all(str_detect(result$GL_string, "HLA-")))
  } else {
    skip("hml_2.hml test file does not exist.")
  }
})

test_that("read_HML produces valid GL strings with locus separators", {
  # Verify that the output GL strings have proper "^" locus separators.
  HML_1 <- system.file("extdata", "HML_1.hml", package = "immunogenetr")

  if (file.exists(HML_1)) {
    result <- read_HML(HML_1)

    # Each GL string should contain "^" separating loci.
    expect_true(all(str_detect(result$GL_string, "\\^")))
    # Each GL string should contain "+" separating gene copies.
    expect_true(all(str_detect(result$GL_string, "\\+")))
  } else {
    skip("HML_1.hml test file does not exist.")
  }
})

test_that("read_HML results are consistent between both HML files", {
  # Both HML files represent the same 5 samples; GL strings should align.
  HML_1 <- system.file("extdata", "HML_1.hml", package = "immunogenetr")
  HML_2 <- system.file("extdata", "hml_2.hml", package = "immunogenetr")

  if (file.exists(HML_1) && file.exists(HML_2)) {
    result_1 <- read_HML(HML_1)
    result_2 <- read_HML(HML_2)

    # Both should have the same number of samples.
    expect_equal(nrow(result_1), nrow(result_2))
  } else {
    skip("HML test files do not exist.")
  }
})

test_that("read_HML errors on non-existent file", {
  # Should produce an error for a file that doesn't exist.
  expect_error(read_HML("nonexistent_file.hml"))
})

test_that("read_HML errors on invalid XML", {
  # Create a temporary file with invalid XML content.
  bad_file <- tempfile(fileext = ".hml")
  writeLines("This is not valid XML content <><>", bad_file)

  expect_error(read_HML(bad_file))
  # Clean up.
  unlink(bad_file)
})

test_that("read_HML handles HML files without XML namespace", {
  # Create a minimal HML file with no namespace declaration,
  # which exercises the else branch for unqualified XPath queries.
  no_ns_file <- tempfile(fileext = ".hml")
  writeLines(c(
    '<?xml version="1.0" encoding="UTF-8"?>',
    '<hml>',
    '  <sample id="DONOR001">',
    '    <typing gene-family="HLA" gene="HLA-A" allele-db="IMGT/HLA" allele-version="3.25.0">',
    '      <allele-assignment allele-db="IMGT/HLA" allele-version="3.25.0">',
    '        <glstring>HLA-A*01:01+HLA-A*02:01</glstring>',
    '      </allele-assignment>',
    '    </typing>',
    '    <typing gene-family="HLA" gene="HLA-B" allele-db="IMGT/HLA" allele-version="3.25.0">',
    '      <allele-assignment allele-db="IMGT/HLA" allele-version="3.25.0">',
    '        <glstring>HLA-B*07:02+HLA-B*08:01</glstring>',
    '      </allele-assignment>',
    '    </typing>',
    '  </sample>',
    '</hml>'
  ), no_ns_file)

  result <- read_HML(no_ns_file)

  # Should return a tibble with the expected structure.
  expect_s3_class(result, "tbl_df")
  expect_true(all(c("sampleID", "GL_string") %in% colnames(result)))
  # Should find 1 sample.
  expect_equal(nrow(result), 1)
  # GL string should contain both loci joined by "^".
  expect_true(str_detect(result$GL_string, "HLA-A"))
  expect_true(str_detect(result$GL_string, "HLA-B"))
  expect_true(str_detect(result$GL_string, "\\^"))
  # Sample ID should match.
  expect_equal(result$sampleID, "DONOR001")

  # Clean up.
  unlink(no_ns_file)
})

test_that("read_HML validates input", {
  # NULL input should error.
  expect_error(read_HML(NULL))
  # Numeric input should error.
  expect_error(read_HML(123))
})

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----message=FALSE------------------------------------------------------------
library(immunogenetr)
library(dplyr)

## -----------------------------------------------------------------------------
# HLA_typing_1 contains typing for 10 individuals across all classical HLA loci.
head(HLA_typing_1, 3)

## -----------------------------------------------------------------------------
HLA_typing_GL <- HLA_typing_1 %>%
  # Convert all typing columns (A1 through DPB1_2) into a GL string.
  mutate(
    GL_string = HLA_columns_to_GLstring(., HLA_typing_columns = A1:DPB1_2),
    .after = patient
  ) %>%
  # Keep only patient ID and the new GL string column.
  select(patient, GL_string)

# View the GL strings.
(HLA_typing_GL)

## -----------------------------------------------------------------------------
# Take the first patient's GL string and split it into locus columns.
# Note: GLstring_genes and GLstring_genes_expanded use pivot_longer on all
# columns, so only pass the GL string column (no other data types).
single_patient <- HLA_typing_GL[1, "GL_string", drop = FALSE]
GLstring_genes(single_patient, "GL_string")

## -----------------------------------------------------------------------------
GLstring_genes_expanded(single_patient, "GL_string")

## -----------------------------------------------------------------------------
# Patient 7 is the recipient, patient 9 is the donor.
recip_gl <- HLA_typing_GL %>% filter(patient == 7) %>% pull(GL_string)
donor_gl <- HLA_typing_GL %>% filter(patient == 9) %>% pull(GL_string)

## -----------------------------------------------------------------------------
# Check if there is an HLA-A mismatch in the graft-vs-host direction.
HLA_mismatch_logical(recip_gl, donor_gl, "HLA-A", direction = "GvH")

# Check host-vs-graft direction.
HLA_mismatch_logical(recip_gl, donor_gl, "HLA-A", direction = "HvG")

## -----------------------------------------------------------------------------
# Count bidirectional mismatches across several loci at once.
HLA_mismatch_number(
  recip_gl, donor_gl,
  c("HLA-A", "HLA-B", "HLA-C", "HLA-DRB1"),
  direction = "bidirectional"
)

## -----------------------------------------------------------------------------
# Identify the specific mismatched alleles in the HvG direction.
HLA_mismatched_alleles(recip_gl, donor_gl, "HLA-A", direction = "HvG")

## -----------------------------------------------------------------------------
# Count the number of matches (complement of mismatches).
HLA_match_number(
  recip_gl, donor_gl,
  c("HLA-A", "HLA-B", "HLA-C", "HLA-DRB1"),
  direction = "bidirectional"
)

## -----------------------------------------------------------------------------
# X-of-8 matching (A, B, C, DRB1 bidirectional).
HLA_match_summary_HCT(recip_gl, donor_gl,
  direction = "bidirectional",
  match_grade = "Xof8"
)

# X-of-10 matching (adds DQB1).
HLA_match_summary_HCT(recip_gl, donor_gl,
  direction = "bidirectional",
  match_grade = "Xof10"
)

## -----------------------------------------------------------------------------
# Patient 3 is the recipient; compare against all 10 donors.
recipient <- HLA_typing_GL %>%
  filter(patient == 3) %>%
  select(GL_string) %>%
  rename(GL_string_recip = GL_string)

donors <- HLA_typing_GL %>%
  rename(GL_string_donor = GL_string, donor = patient) %>%
  # Cross-join to pair recipient with each donor.
  cross_join(recipient) %>%
  # Calculate 8/8 match grade for each pair.
  mutate(
    match_8of8 = HLA_match_summary_HCT(
      GL_string_recip, GL_string_donor,
      direction = "bidirectional",
      match_grade = "Xof8"
    ),
    .after = donor
  ) %>%
  # Sort best matches first.
  arrange(desc(match_8of8))

donors %>% select(donor, match_8of8)

## -----------------------------------------------------------------------------
# Truncate a four-field allele to two fields.
HLA_truncate("HLA-A*02:01:01:01", fields = 2)

# Works on full GL strings too.
HLA_truncate("HLA-A*02:01:01:01+HLA-A*03:01:01:02^HLA-B*07:02:01:01+HLA-B*44:02:01:01",
  fields = 2
)

## -----------------------------------------------------------------------------
# Remove all prefixes to get just the allele fields.
HLA_prefix_remove("HLA-A*02:01")

# Keep the locus designation but remove "HLA-".
HLA_prefix_remove("HLA-A*02:01", keep_locus = TRUE)

# Add the full prefix back.
HLA_prefix_add("02:01", "HLA-A*")

# "HLA-" is added by default.
HLA_prefix_add("A*02:01")

## -----------------------------------------------------------------------------
gl <- "HLA-A*02:01:01+HLA-A*68:01^HLA-B*07:01+HLA-B*15:01"

# A two-field search correctly matches the three-field allele.
pattern <- GLstring_regex("HLA-A*02:01")
stringr::str_detect(gl, pattern)

# But won't falsely match a longer allele number.
stringr::str_detect("HLA-A*02:149:01", GLstring_regex("HLA-A*02:14"))

## -----------------------------------------------------------------------------
# GLstring_genes returns tidyverse-friendly names by default.
repaired <- GLstring_genes(single_patient, "GL_string")
names(repaired)

# Convert back to WHO format with asterisks.
who_names <- HLA_column_repair(repaired, format = "WHO", asterisk = TRUE)
names(who_names)

## -----------------------------------------------------------------------------
# immunogenetr ships with two example HML files.
hml_path <- system.file("extdata", "HML_1.hml", package = "immunogenetr")
hml_result <- read_HML(hml_path)
hml_result

#' @title HLA_validate
#'
#' @description Returns only HLA alleles in valid nomenclature, either serologic or molecular.
#' Simple numbers, such as "2" or "27" will be returned as-is. Suffixes that are not
#' WHO-recognized suffixes (L, S, C, A, Q, N) or G or P group designations will be removed.
#' For example "novel" at the end of the allele will be removed, while "n" at the end of the
#' allele will be retained. Other values, such as "blank" or "-" will be converted to NA values.
#' This function is helpful for cleaning up the typing of an entire table of HLA values.
#'
#'
#' @param data A string containing an HLA allele.
#'
#' @return A string with a valid HLA allele or NA if no valid allele was present.
#'
#' @examples
#' HLA_validate("HLA-A2")
#' HLA_validate("A*02:01:01:01N")
#' HLA_validate("A*02:01:01N")
#' HLA_validate("HLA-DRB1*02:03novel")
#' HLA_validate("HLA-DQB1*03:01v")
#' HLA_validate("HLA-DRB1*02:03P")
#' HLA_validate("HLA-DPB1*04:01:01G")
#' HLA_validate("2")
#' HLA_validate(2)
#' HLA_validate("B27")
#' HLA_validate("A*010101")
#' HLA_validate("-")
#' HLA_validate("blank")
#'
#' # The HLA_typing_LIS dataset contains a table with HLA typing spread across multiple columns:
#' print(HLA_typing_LIS)
#'
#' # Cleaning up the entire table. Note that blank values will be converted to "NA".
#' library(dplyr)
#' HLA_typing_LIS %>% mutate(
#'   across(
#'     mA1Cd.recipient:mDPB12cd.recipient,
#'     ~ HLA_validate(.)
#'   )
#' )
#' @export
#'
#' @importFrom dplyr %>%
#' @importFrom stringr str_extract

HLA_validate <- function(data) {
  data %>%
    str_extract("(HLA-)?([:alnum:]*)(\\*)?[:digit:]{1,}:?[:digit:]*:?[:digit:]*:?[:digit:]*([GPLSCAQNgplscaqn](?!.))*") %>%
    as.character()
}

library(testthat)
library(stringr)

test_that("HLA_validate correctly validates and cleans HLA alleles", {
  expect_equal(HLA_validate("HLA-A2"), "HLA-A2")
  expect_equal(HLA_validate("A*02:01:01:01N"), "A*02:01:01:01N")
  expect_equal(HLA_validate("A*02:01:01N"), "A*02:01:01N")
  expect_equal(HLA_validate("HLA-DRB1*02:03novel"), "HLA-DRB1*02:03")
  expect_equal(HLA_validate("HLA-DQB1*03:01v"), "HLA-DQB1*03:01")
  expect_equal(HLA_validate("HLA-DRB1*02:03P"), "HLA-DRB1*02:03P")
  expect_equal(HLA_validate("HLA-DPB1*04:01:01G"), "HLA-DPB1*04:01:01G")
  expect_equal(HLA_validate("2"), "2")
  expect_equal(HLA_validate(2), "2")
  expect_equal(HLA_validate("B27"), "B27")
  expect_equal(HLA_validate("A*010101"), "A*010101")
  expect_equal(HLA_validate("-"), NA_character_)
  expect_equal(HLA_validate("blank"), NA_character_)
})

#' @title HLA_mismatched_alleles
#'
#' @description A function to return a string of mismatches between recipient
#' and donor HLA genotypes represented as GL strings. The function finds
#' mismatches based on the direction of comparison specified in the inputs
#' and also handles homozygosity.
#'
#' `HLA_mismatch_alleles` and `HLA_mismatched_alleles` are synonyms.
#'
#' @param GL_string_recip A GL strings representing the recipient's HLA genotypes.
#' @param GL_string_donor A GL strings representing the donor's HLA genotypes.
#' @param loci A character vector specifying the loci to be considered for
#' mismatch calculation. HLA-DRB3/4/5 (and their serologic equivalents DR51/52/53)
#' are considered once locus for this function, and should be called in this argument
#' as "HLA-DRB3/4/5" or "HLA-DR51/52/53", respectively.
#' @param direction A character string indicating the direction of mismatch.
#' Options are "HvG" (host vs. graft), "GvH" (graft vs. host), "bidirectional"
#' (the max value of "HvG" and "GvH"), or "SOT" (host vs. graft, as is used for
#' mismatching in solid organ transplantation).
#' @param homozygous_count An integer specifying how to handle homozygosity.
#' Defaults to 2, where homozygous alleles are treated as duplicated for
#' mismatch calculations. Can be specified as 1, in which case homozygous
#' alleles are treated as single occurrences without duplication (in other words,
#' homozygous mismatches are only "counted" once).
#'
#' @return A character vector, where each element is a string summarizing the
#' mismatches for the specified loci. The strings are formatted as
#' comma-separated locus mismatch entries if multiple loci were supplied, or as
#' simple GL strings if a single locus was supplied.

#'
#' @examples
#' file <- HLA_typing_1[, -1]
#' GL_string <- HLA_columns_to_GLstring(file, HLA_typing_columns = everything())
#'
#' GL_string_recip <- GL_string[1]
#' GL_string_donor <- GL_string[2]
#'
#' loci <- c("HLA-A", "HLA-DRB3/4/5", "HLA-DPB1")
#' mismatches <- HLA_mismatched_alleles(GL_string_recip, GL_string_donor, loci, direction = "HvG")
#' print(mismatches)
#'
#' @export
#'
#' @importFrom stringr str_c
#' @importFrom tibble tibble
#' @importFrom magrittr  %>%
#' @importFrom dplyr mutate
#' @importFrom dplyr across
#' @importFrom tidyr replace_na
#' @importFrom stringr str_c
#' @importFrom tidyr unite

HLA_mismatched_alleles <- function(GL_string_recip, GL_string_donor, loci, direction, homozygous_count = 2) {
  # Validate inputs
  check_gl_string(GL_string_recip, "GL_string_recip")
  check_gl_string(GL_string_donor, "GL_string_donor")
  check_loci(loci)
  check_homozygous_count(homozygous_count)

  direction <- match.arg(direction, c("HvG", "GvH", "bidirectional", "SOT"))
  # "HvG" or "GvH" can use the output of `HLA_mismatch_base` directly.
  if (direction == "HvG" | direction == "GvH") {
    HLA_mismatch_base(GL_string_recip, GL_string_donor, loci, direction, homozygous_count)
    # "SOT" defaults to "HvG".
  } else if (direction == "SOT") {
    HLA_mismatch_base(GL_string_recip, GL_string_donor, loci, "HvG", homozygous_count)
    # "Bidirectional" will paste together the output of each direction.
  } else if (direction == "bidirectional") {
    HvG <- HLA_mismatch_base(GL_string_recip, GL_string_donor, loci, "HvG", homozygous_count)
    GvH <- HLA_mismatch_base(GL_string_recip, GL_string_donor, loci, "GvH", homozygous_count)
    # Combine the results from each direction
    bidirectional <- tibble("HvG" = HvG, "GvH" = GvH) %>%
      mutate(across(HvG:GvH, ~ replace_na(., "NA"))) %>%
      mutate(HvG = str_c("HvG;", HvG), GvH = str_c("GvH;", GvH)) %>%
      unite(HvG, GvH, col = "bidirectional", sep = "<>")

    return(bidirectional$bidirectional)
  }
}

globalVariables(c("HvG", "GvH"))

library(testthat)
library(dplyr)
library(stringr)

test_that("HLA_mismatched_alleles correctly identifies mismatched alleles", {
  GL_string_recip <- "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01^HLA-DRB1*13:01+HLA-DRB1*15:01"
  GL_string_donor <- "HLA-A*01:01+HLA-A*03:01^HLA-B*07:02+HLA-B*44:02^HLA-DRB1*13:01+HLA-DRB1*14:01"
  loci <- c("HLA-A", "HLA-B", "HLA-DRB1")

  result_HvG <- HLA_mismatched_alleles(GL_string_recip, GL_string_donor, loci, direction = "HvG")
  result_GvH <- HLA_mismatched_alleles(GL_string_recip, GL_string_donor, loci, direction = "GvH")
  result_bidirectional <- HLA_mismatched_alleles(GL_string_recip, GL_string_donor, loci, direction = "bidirectional")
  result_SOT <- HLA_mismatched_alleles(GL_string_recip, GL_string_donor, loci, direction = "SOT")

  expect_equal(result_HvG, "HLA-A=HLA-A*03:01, HLA-B=HLA-B*44:02, HLA-DRB1=HLA-DRB1*14:01")
  expect_equal(result_GvH, "HLA-A=HLA-A*02:01, HLA-B=HLA-B*08:01, HLA-DRB1=HLA-DRB1*15:01")
  expect_equal(result_SOT, "HLA-A=HLA-A*03:01, HLA-B=HLA-B*44:02, HLA-DRB1=HLA-DRB1*14:01")
  expect_equal(result_bidirectional, "HvG;HLA-A=HLA-A*03:01, HLA-B=HLA-B*44:02, HLA-DRB1=HLA-DRB1*14:01<>GvH;HLA-A=HLA-A*02:01, HLA-B=HLA-B*08:01, HLA-DRB1=HLA-DRB1*15:01")

  result_single_locus <- HLA_mismatched_alleles(GL_string_recip, GL_string_donor, "HLA-A", direction = "HvG")
  expect_equal(result_single_locus, "HLA-A*03:01")

  result_homozygous_1 <- HLA_mismatched_alleles(GL_string_recip, GL_string_donor, "HLA-B", direction = "HvG", homozygous_count = 1)
  expect_equal(result_homozygous_1, "HLA-B*44:02")

  result_homozygous_2 <- HLA_mismatched_alleles(GL_string_recip, GL_string_donor, "HLA-B", direction = "HvG", homozygous_count = 2)
  expect_equal(result_homozygous_2, "HLA-B*44:02")

  expect_error(HLA_mismatched_alleles(GL_string_recip, GL_string_donor, loci, direction = "invalid"))
})


# --- Input validation tests ---

test_that("HLA_mismatched_alleles rejects NULL inputs", {
  gl <- "HLA-A*01:01+HLA-A*02:01"
  expect_error(HLA_mismatched_alleles(NULL, gl, "HLA-A", "HvG"), "GL_string_recip")
  expect_error(HLA_mismatched_alleles(gl, NULL, "HLA-A", "HvG"), "GL_string_donor")
  expect_error(HLA_mismatched_alleles(gl, gl, NULL, "HvG"), "loci")
})

test_that("HLA_mismatched_alleles rejects non-character GL strings", {
  gl <- "HLA-A*01:01+HLA-A*02:01"
  expect_error(HLA_mismatched_alleles(123, gl, "HLA-A", "HvG"), "must be a character")
})

test_that("HLA_mismatched_alleles rejects invalid homozygous_count", {
  gl <- "HLA-A*01:01+HLA-A*02:01"
  expect_error(HLA_mismatched_alleles(gl, gl, "HLA-A", "HvG", homozygous_count = 0), "must be.*1.*or.*2")
})

test_that("HLA_mismatched_alleles returns NA for perfect match at single locus", {
  gl <- "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01"
  result <- HLA_mismatched_alleles(gl, gl, "HLA-A", "HvG")
  expect_true(is.na(result))
})

test_that("HLA_mismatched_alleles agrees with mismatch_table_2010 consensus (homozygous_count = 1)", {
  # Define symbolic alleles to substitute into the table.
  A <- "HLA-A*01:01"
  B <- "HLA-A*02:05"
  C <- "HLA-A*24:02"
  D <- "HLA-A*31:03"
  N <- "HLA-A*68:11N"

  # Build GL strings from the 2010 table's symbolic allele pairs.
  mismatch_table_2010_alleles <- mismatch_table_2010 %>%
    rename(Patient = patient, Donor = donor) %>%
    # Assign the example alleles based on symbolic codes.
    mutate(recipient_1 = case_when(
      str_detect(Patient, "^A") ~ A,
      str_detect(Patient, "^N") ~ N
    ), recipient_2 = case_when(
      str_detect(Patient, "A$") ~ A,
      str_detect(Patient, "B$") ~ B,
      str_detect(Patient, "N$") ~ N
    ), donor_1 = case_when(
      str_detect(Donor, "^A") ~ A,
      str_detect(Donor, "^B") ~ B,
      str_detect(Donor, "^C") ~ C,
      str_detect(Donor, "^N") ~ N
    ), donor_2 = case_when(
      str_detect(Donor, "A$") ~ A,
      str_detect(Donor, "B$") ~ B,
      str_detect(Donor, "C$") ~ C,
      str_detect(Donor, "D$") ~ D,
      str_detect(Donor, "N$") ~ N
    )) %>%
    # Turn the two alleles into a GL string.
    mutate(
      GL_string_recip = str_c(recipient_1, recipient_2, sep = "+"),
      GL_string_donor = str_c(donor_1, donor_2, sep = "+")
    ) %>%
    select(Patient, Donor, GL_string_recip, GL_string_donor, "#GvH", "#HvG") %>%
    rename(GvH_count = "#GvH", HvG_count = "#HvG") %>%
    # Calculate mismatched alleles for each direction with homozygous_count = 1.
    mutate(
      GvH_alleles = HLA_mismatched_alleles(GL_string_recip, GL_string_donor, "HLA-A", "GvH", homozygous_count = 1),
      HvG_alleles = HLA_mismatched_alleles(GL_string_recip, GL_string_donor, "HLA-A", "HvG", homozygous_count = 1)
    ) %>%
    # Verify: NA when mismatch count is 0, non-NA when count > 0.
    mutate(
      GvH_na_correct = (is.na(GvH_alleles) == (GvH_count == 0)),
      HvG_na_correct = (is.na(HvG_alleles) == (HvG_count == 0))
    ) %>%
    # Verify: number of alleles returned matches the mismatch count.
    # Count alleles by splitting on "+" delimiter.
    mutate(
      GvH_n_alleles = if_else(is.na(GvH_alleles), 0L, as.integer(str_count(GvH_alleles, "\\+") + 1)),
      HvG_n_alleles = if_else(is.na(HvG_alleles), 0L, as.integer(str_count(HvG_alleles, "\\+") + 1))
    ) %>%
    mutate(
      GvH_count_correct = (GvH_n_alleles == GvH_count),
      HvG_count_correct = (HvG_n_alleles == HvG_count)
    ) %>%
    # Check that all validations pass for every row.
    mutate(total_result = if_all(c(GvH_na_correct, HvG_na_correct, GvH_count_correct, HvG_count_correct))) %>%
    distinct(total_result) %>%
    pull(total_result)

  expect_equal(mismatch_table_2010_alleles, TRUE)
})

test_that("HLA_mismatched_alleles agrees with mismatch_table_2016 consensus (homozygous_count = 2)", {
  # Define symbolic alleles to substitute into the table.
  A <- "HLA-A*01:01"
  B <- "HLA-A*02:05"
  C <- "HLA-A*24:02"
  D <- "HLA-A*31:03"
  N <- "HLA-A*68:11N"

  # Build GL strings from the 2016 table's symbolic allele pairs.
  mismatch_table_2016_alleles <- mismatch_table_2016 %>%
    # Assign the example alleles based on symbolic codes.
    mutate(recipient_1 = case_when(
      str_detect(Patient, "^A") ~ A,
      str_detect(Patient, "^N") ~ N
    ), recipient_2 = case_when(
      str_detect(Patient, "A$") ~ A,
      str_detect(Patient, "B$") ~ B,
      str_detect(Patient, "N$") ~ N
    ), donor_1 = case_when(
      str_detect(Donor, "^A") ~ A,
      str_detect(Donor, "^B") ~ B,
      str_detect(Donor, "^C") ~ C,
      str_detect(Donor, "^N") ~ N
    ), donor_2 = case_when(
      str_detect(Donor, "A$") ~ A,
      str_detect(Donor, "B$") ~ B,
      str_detect(Donor, "C$") ~ C,
      str_detect(Donor, "D$") ~ D,
      str_detect(Donor, "N$") ~ N
    )) %>%
    # Turn the two alleles into a GL string.
    mutate(
      GL_string_recip = str_c(recipient_1, recipient_2, sep = "+"),
      GL_string_donor = str_c(donor_1, donor_2, sep = "+")
    ) %>%
    select(Patient, Donor, GL_string_recip, GL_string_donor, "#GvH", "#HvG") %>%
    rename(GvH_count = "#GvH", HvG_count = "#HvG") %>%
    # Calculate mismatched alleles for each direction with homozygous_count = 2 (default).
    mutate(
      GvH_alleles = HLA_mismatched_alleles(GL_string_recip, GL_string_donor, "HLA-A", "GvH", homozygous_count = 2),
      HvG_alleles = HLA_mismatched_alleles(GL_string_recip, GL_string_donor, "HLA-A", "HvG", homozygous_count = 2)
    ) %>%
    # Verify: NA when mismatch count is 0, non-NA when count > 0.
    mutate(
      GvH_na_correct = (is.na(GvH_alleles) == (GvH_count == 0)),
      HvG_na_correct = (is.na(HvG_alleles) == (HvG_count == 0))
    ) %>%
    # Verify: number of alleles returned matches the mismatch count.
    # Count alleles by splitting on "+" delimiter.
    mutate(
      GvH_n_alleles = if_else(is.na(GvH_alleles), 0L, as.integer(str_count(GvH_alleles, "\\+") + 1)),
      HvG_n_alleles = if_else(is.na(HvG_alleles), 0L, as.integer(str_count(HvG_alleles, "\\+") + 1))
    ) %>%
    mutate(
      GvH_count_correct = (GvH_n_alleles == GvH_count),
      HvG_count_correct = (HvG_n_alleles == HvG_count)
    ) %>%
    # Check that all validations pass for every row.
    mutate(total_result = if_all(c(GvH_na_correct, HvG_na_correct, GvH_count_correct, HvG_count_correct))) %>%
    distinct(total_result) %>%
    pull(total_result)

  expect_equal(mismatch_table_2016_alleles, TRUE)
})

test_that("HLA_mismatched_alleles works with vectorized inputs", {
  recip <- c(
    "HLA-A*01:01+HLA-A*02:01",
    "HLA-A*03:01+HLA-A*24:02"
  )
  donor <- c(
    "HLA-A*01:01+HLA-A*03:01",
    "HLA-A*03:01+HLA-A*24:02"
  )
  result <- HLA_mismatched_alleles(recip, donor, "HLA-A", "HvG")
  expect_length(result, 2)
  # First pair has a mismatch
  expect_equal(result[1], "HLA-A*03:01")
  # Second pair is a perfect match
  expect_true(is.na(result[2]))
})

#' @title GLstring_genes_expanded
#'
#' @description This function processes a specified column in a data frame
#' that contains GL strings. It separates the GL strings, identifies the HLA
#' loci, and transforms the data into a wider format with loci as column names.
#' It also creates multiple rows to separate each locus in the allele.
#'
#' @param data A data frame containing GL strings for HLA data.
#' @param gl_string The name of the column in the data frame that contains
#' GL strings.
#'
#' @return A data frame with expanded columns, where each row has a single
#' allele for a specific locus.
#'
#' @examples
#'
#' file <- HLA_typing_1[, -1]
#' GL_string <- data.frame("GL_string" = HLA_columns_to_GLstring(
#'   file,
#'   HLA_typing_columns = everything()
#' ))
#' GL_string <- GL_string[1, , drop = FALSE] # When considering first patient
#' result <- GLstring_genes_expanded(GL_string, "GL_string")
#' print(result)
#'
#' @export
#'
#' @importFrom dplyr mutate
#' @importFrom dplyr %>%
#' @importFrom tidyr pivot_longer
#' @importFrom tidyr separate_rows
#' @importFrom tidyr pivot_wider
#' @importFrom tidyr unnest


GLstring_genes_expanded <- function(data, gl_string) {
  # Validate input
  check_data_frame(data, "data")

  data %>%
    GLstring_genes(all_of(gl_string)) %>%
    pivot_longer(cols = everything(), names_to = "locus", values_to = "alleles") %>% # Pivot to long format
    mutate(locus = gsub("HLA-|HLA_", "", locus)) %>% # Normalize locus names
    separate_rows(alleles, sep = "\\+") %>% # Separate rows by "+"
    pivot_wider(names_from = locus, values_from = alleles, values_fn = list(alleles = list)) %>% # Pivot back to wide format
    unnest(cols = everything()) # Unnest the columns
}

globalVariables(c("alleles", "everything"))

library(testthat)
library(dplyr)
library(tidyr)
library(stringr)

test_that("GLstring_genes_expanded correctly expands GL strings into separate loci", {
  table <- data.frame(
    GL_string = "HLA-A*29:02+HLA-A*30:02^HLA-C*06:02+HLA-C*07:01^HLA-B*08:01+HLA-B*13:02^HLA-DRB4*01:03^HLA-DRB1*04:01+HLA-DRB1*07:01",
    stringsAsFactors = FALSE
  )

  result <- GLstring_genes_expanded(table, "GL_string")

  expect_s3_class(result, "data.frame")
  expect_gt(nrow(result), 0)
  expect_true(all(c("A", "C", "B", "DRB4", "DRB1") %in% colnames(result)))

  expect_equal(sum(result$A == "HLA-A*29:02"), 1)
  expect_equal(sum(result$A == "HLA-A*30:02"), 1)
  expect_equal(sum(result$C == "HLA-C*06:02"), 1)
  expect_equal(sum(result$C == "HLA-C*07:01"), 1)
  expect_equal(sum(result$B == "HLA-B*08:01"), 1)
  expect_equal(sum(result$B == "HLA-B*13:02"), 1)
  # Known behavior: DRB4 has a single allele (no "+") while other loci have two.
  # When unnest() expands list columns to equal lengths, the single DRB4 value
  # gets recycled to 2 rows. This is expected given the current pivot_wider + unnest
  # approach in GLstring_genes_expanded.
  expect_equal(sum(result$DRB4 == "HLA-DRB4*01:03"), 2)
  expect_equal(sum(result$DRB1 == "HLA-DRB1*04:01"), 1)
  expect_equal(sum(result$DRB1 == "HLA-DRB1*07:01"), 1)
})

test_that("GLstring_genes_expanded handles missing values correctly", {
  table <- data.frame(
    GL_string = c("HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01", NA),
    stringsAsFactors = FALSE
  )

  result <- GLstring_genes_expanded(table, "GL_string")

  expect_gt(nrow(result), 0)
  expect_true("A" %in% colnames(result))
  expect_true("B" %in% colnames(result))

  expect_true(any(is.na(result)))
})

#' @title GLstring_genes
#'
#' @description This function processes a specified column in a data frame
#' that contains GL strings. It separates the GL strings, identifies the HLA
#' loci, and transforms the data into a wider format with loci as column names.
#'
#' @param data A data frame
#' @param gl_string The name of the column in the data frame that contains
#' GL strings
#'
#' @return A data frame with GL strings separated, loci identified, and data
#' transformed to a wider format with loci as columns.
#'
#' @examples
#'
#' file <- HLA_typing_1[, -1]
#' GL_string <- data.frame("GL_string" = HLA_columns_to_GLstring(
#'   file,
#'   HLA_typing_columns = everything()
#' ))
#' GL_string <- GL_string[1, , drop = FALSE] # When considering first patient
#' result <- GLstring_genes(GL_string, "GL_string")
#' print(result)
#'
#' @export
#'
#' @importFrom dplyr select
#' @importFrom dplyr rename
#' @importFrom dplyr mutate
#' @importFrom dplyr %>%
#' @importFrom tidyr separate_longer_delim
#' @importFrom tidyr pivot_wider
#' @importFrom stringr str_extract


GLstring_genes <- function(data, gl_string) {
  # Validate input
  check_data_frame(data, "data")

  # Identify the columns to modify
  col2mod <- names(select(data, {{ gl_string }}))
  data %>%
    # Separate the GL string column by the delimiter "^" into multiple rows
    separate_longer_delim({{ col2mod }}, delim = "^") %>%
    # Rename the separated column to "gl_string"
    rename(gl_string = {{ col2mod }}) %>%
    # Extract the locus information from the GL string
    mutate(locus = str_extract(gl_string, "[[:alnum:]-]+(?=\\*)")) %>%
    # Transform the data from long to wide format, using locus names as new column names
    pivot_wider(names_from = locus, values_from = gl_string) %>%
    # Apply the HLA_column_repair function to the transformed data
    HLA_column_repair(.)
}

globalVariables(c("."))

library(testthat)
library(dplyr)
library(tidyr)
library(stringr)

test_that("GLstring_genes correctly processes GL strings", {
  # Basic test with three loci separated by "^".
  table <- data.frame(
    GL_string = "HLA-A*29:02+HLA-A*30:02^HLA-C*06:02+HLA-C*07:01^HLA-B*08:01+HLA-B*13:02",
    stringsAsFactors = FALSE
  )

  result <- GLstring_genes(table, "GL_string")

  # Should return a data frame.
  expect_s3_class(result, "data.frame")
  # Should have HLA_A, HLA_C, HLA_B columns (tidyverse-repaired names).
  expect_true(all(c("HLA_A", "HLA_C", "HLA_B") %in% colnames(result)))
  # Each locus column should contain the correct GL string fragment.
  expect_true("HLA-A*29:02+HLA-A*30:02" %in% result$HLA_A)
  expect_true("HLA-C*06:02+HLA-C*07:01" %in% result$HLA_C)
  expect_true("HLA-B*08:01+HLA-B*13:02" %in% result$HLA_B)
})

test_that("GLstring_genes handles a full HLA typing with many loci", {
  # GL string with standard HCT loci: A, B, C, DRB1, DQB1.
  gl <- paste(
    "HLA-A*02:01+HLA-A*03:01",
    "HLA-B*07:02+HLA-B*44:02",
    "HLA-C*07:02+HLA-C*05:01",
    "HLA-DRB1*15:01+HLA-DRB1*03:01",
    "HLA-DQB1*06:02+HLA-DQB1*02:01",
    sep = "^"
  )
  table <- data.frame(GL_string = gl, stringsAsFactors = FALSE)

  result <- GLstring_genes(table, "GL_string")

  # Should have all five locus columns.
  expect_true(all(c("HLA_A", "HLA_B", "HLA_C", "HLA_DRB1", "HLA_DQB1") %in% colnames(result)))
  # Verify one row of output.
  expect_equal(nrow(result), 1)
})

test_that("GLstring_genes handles multiple rows", {
  # Two patients with different typings.
  table <- data.frame(
    patient = c("P1", "P2"),
    GL_string = c(
      "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01",
      "HLA-A*03:01+HLA-A*11:01^HLA-B*15:01+HLA-B*35:01"
    ),
    stringsAsFactors = FALSE
  )

  result <- GLstring_genes(table, "GL_string")

  # Should have two rows.
  expect_equal(nrow(result), 2)
  # Should have patient column preserved.
  expect_true("patient" %in% colnames(result))
  # Should have HLA locus columns.
  expect_true(all(c("HLA_A", "HLA_B") %in% colnames(result)))
})

test_that("GLstring_genes handles GL strings with ambiguity (slash notation)", {

  # A GL string with allele-level ambiguity using "/" delimiter.
  gl <- "HLA-A*02:01/HLA-A*02:06+HLA-A*03:01^HLA-B*07:02+HLA-B*44:02"
  table <- data.frame(GL_string = gl, stringsAsFactors = FALSE)

  result <- GLstring_genes(table, "GL_string")

  # The ambiguity should be preserved within the HLA_A column.
  expect_true(grepl("/", result$HLA_A))
  expect_true("HLA_B" %in% colnames(result))
})

test_that("GLstring_genes handles a single-locus GL string", {
  # Only one locus, no "^" separator.
  table <- data.frame(
    GL_string = "HLA-A*01:01+HLA-A*02:01",
    stringsAsFactors = FALSE
  )

  result <- GLstring_genes(table, "GL_string")

  expect_s3_class(result, "data.frame")
  expect_true("HLA_A" %in% colnames(result))
  expect_equal(nrow(result), 1)
})

test_that("GLstring_genes handles DRB3/4/5 loci", {
  # GL string including DRB345 loci.
  gl <- paste(
    "HLA-DRB1*04:01+HLA-DRB1*07:01",
    "HLA-DRB4*01:03+HLA-DRB4*01:01",
    sep = "^"
  )
  table <- data.frame(GL_string = gl, stringsAsFactors = FALSE)

  result <- GLstring_genes(table, "GL_string")

  expect_true("HLA_DRB1" %in% colnames(result))
  expect_true("HLA_DRB4" %in% colnames(result))
})

test_that("GLstring_genes validates inputs", {
  # NULL data should error.
  expect_error(GLstring_genes(NULL, "GL_string"))
  # Non-data-frame should error.
  expect_error(GLstring_genes("not a data frame", "GL_string"))
})

library(testthat)
library(dplyr)
library(stringr)

test_that("GLstring_genotype_ambiguity correctly retains the first genotype ambiguity", {
  HLA_type <- data.frame(
    sample = c("sample1", "sample2"),
    HLA_A = c(
      "A*01:01+A*68:01|A*01:02+A*68:55|A*01:99+A*68:66",
      "A*02:01+A*03:01|A*02:02+A*03:03"
    ),
    HLA_B = c(
      "B*07:02+B*58:01|B*07:03+B*58:09",
      "B*08:01+B*15:01|B*08:02+B*15:17"
    ),
    stringsAsFactors = FALSE
  )

  result <- GLstring_genotype_ambiguity(HLA_type, columns = c("HLA_A", "HLA_B"))

  expect_s3_class(result, "data.frame")
  expect_equal(result$HLA_A[1], "A*01:01+A*68:01")
  expect_equal(result$HLA_A[2], "A*02:01+A*03:01")
  expect_equal(result$HLA_B[1], "B*07:02+B*58:01")
  expect_equal(result$HLA_B[2], "B*08:01+B*15:01")
})

test_that("GLstring_genotype_ambiguity errors when genes have not been separated", {
  # Create a data frame with a full GL string that still contains "^" (gene separator),
  # indicating genes haven't been split with GLstring_genes() first.
  unseparated <- data.frame(
    GL_string = "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01",
    stringsAsFactors = FALSE
  )

  # Should error because "^" is still present in the GL string

  expect_error(
    GLstring_genotype_ambiguity(unseparated, columns = "GL_string"),
    "Genes must be separated"
  )
})

test_that("GLstring_genotype_ambiguity keeps ambiguities when keep_ambiguities = TRUE", {
  # Data with multiple genotype ambiguities separated by "|"
  HLA_type <- data.frame(
    sample = c("sample1"),
    HLA_A = c("A*01:01+A*68:01|A*01:02+A*68:55|A*01:99+A*68:66"),
    stringsAsFactors = FALSE
  )

  # With keep_ambiguities = TRUE, should create an "_genotype_ambiguity" column
  result <- GLstring_genotype_ambiguity(HLA_type, columns = "HLA_A", keep_ambiguities = TRUE)

  # First genotype ambiguity should be in the original column

  expect_equal(result$HLA_A, "A*01:01+A*68:01")
  # Remaining ambiguities should be in the new column

  expect_equal(result$HLA_A_genotype_ambiguity, "A*01:02+A*68:55|A*01:99+A*68:66")
})

test_that("GLstring_genotype_ambiguity returns NA for ambiguity column when no ambiguity exists", {
  # Data with no genotype ambiguity (no "|" in the GL string)
  no_ambiguity <- data.frame(
    HLA_A = "A*01:01+A*68:01",
    stringsAsFactors = FALSE
  )

  # With keep_ambiguities = TRUE, ambiguity column should be NA
  result <- GLstring_genotype_ambiguity(no_ambiguity, columns = "HLA_A", keep_ambiguities = TRUE)

  expect_equal(result$HLA_A, "A*01:01+A*68:01")
  expect_true(is.na(result$HLA_A_genotype_ambiguity))
})

test_that("GLstring_genotype_ambiguity validates inputs", {
  # Non-data-frame input should error
  expect_error(
    GLstring_genotype_ambiguity("not a data frame", columns = "col1")
  )

  # Non-logical keep_ambiguities should error
  HLA_type <- data.frame(HLA_A = "A*01:01+A*02:01", stringsAsFactors = FALSE)
  expect_error(
    GLstring_genotype_ambiguity(HLA_type, columns = "HLA_A", keep_ambiguities = "yes")
  )
})

#' @title HLA_mismatch_number
#'
#' @description Calculates the number of mismatched HLA alleles between a
#' recipient and a donor across specified loci. Supports mismatch calculations
#' for host-vs-graft (HvG), graft-vs-host (GvH), or bidirectional.
#'
#' @param GL_string_recip A GL string representing the recipient's HLA genotype.
#' @param GL_string_donor A GL string representing the donor's HLA genotype.
#' @param loci A character vector specifying the loci to be considered for
#' mismatch calculation. HLA-DRB3/4/5 (and their serologic equivalents DR51/52/53)
#' are considered once locus for this function, and should be called in this argument
#' as "HLA-DRB3/4/5" or "HLA-DR51/52/53", respectively.
#' @param direction A character string indicating the direction of mismatch.
#' Options are "HvG" (host vs. graft), "GvH" (graft vs. host), "bidirectional"
#' (the max value of "HvG" and "GvH"), or "SOT" (host vs. graft, as is used for
#' mismatching in solid organ transplantation).
#' @param homozygous_count An integer specifying how to count homozygous mismatches.
#' Defaults to 2, where homozygous mismatches are treated as two mismatches,
#' regardless if one or two alleles are supplied in the GL string (in cases
#' where one allele is supplied, it is duplicated by the function). If
#' specified as 1, homozygous mismatches are only counted once, regardless of
#' whether one or two alleles are supplied in the GL string (in cases where
#' two alleles are supplied, the second identical allele is deleted).
#'
#' @return An integer value or a character string:
#' - If `loci` includes only one locus, the function returns an integer
#' mismatch count for that locus.
#' - If `loci` includes multiple loci, the function returns a character
#' string in the format "Locus1=Count1, Locus2=Count2, ...".
#'
#' @examples
#'
#' file <- HLA_typing_1[, -1]
#' GL_string <- HLA_columns_to_GLstring(file, HLA_typing_columns = everything())
#'
#' GL_string_recip <- GL_string[1]
#' GL_string_donor <- GL_string[2]
#'
#' loci <- c("HLA-A", "HLA-DRB3/4/5", "HLA-DPB1")
#'
#' # Calculate mismatch numbers (Host vs. Graft)
#' HLA_mismatch_number(GL_string_recip, GL_string_donor, loci, direction = "HvG")
#'
#' # Calculate mismatch numbers (Graft vs. Host)
#' HLA_mismatch_number(GL_string_recip, GL_string_donor, loci, direction = "GvH")
#'
#' # Calculate mismatch numbers (Bidirectional)
#' HLA_mismatch_number(GL_string_recip, GL_string_donor,
#'   loci,
#'   direction = "bidirectional"
#' )
#'
#' @export
#'
#' @importFrom dplyr join_by
#' @importFrom dplyr mutate
#' @importFrom dplyr summarise
#' @importFrom dplyr left_join
#' @importFrom dplyr na_if
#' @importFrom stringr str_count
#' @importFrom stringr str_flatten
#' @importFrom tidyr separate_longer_delim
#' @importFrom tidyr separate_wider_delim
#' @importFrom tidyr replace_na
#' @importFrom tibble tibble
#' @importFrom tidyr unite


HLA_mismatch_number <- function(GL_string_recip, GL_string_donor, loci, direction, homozygous_count = 2) {
  # Validate inputs
  check_gl_string(GL_string_recip, "GL_string_recip")
  check_gl_string(GL_string_donor, "GL_string_donor")
  check_loci(loci)
  check_homozygous_count(homozygous_count)

  direction <- match.arg(direction, c("HvG", "GvH", "bidirectional", "SOT"))

  # Helper to count mismatched alleles from HLA_mismatch_base output.
  # The regex "(\\+|$)" matches each allele (separated by "+" or at end of string).
  count_mismatches <- function(mismatch_str) {
    replace_na(str_count(mismatch_str, "(\\+|$)"), 0)
  }

  # Determine which direction(s) need to be computed.
  # HvG and SOT only need the HvG direction; GvH only needs GvH.
  # Bidirectional needs both to take the max.
  need_HvG <- (direction == "HvG" | direction == "SOT" | direction == "bidirectional")
  need_GvH <- (direction == "GvH" | direction == "bidirectional")

  # Code to determine mismatch numbers if a single locus was supplied.
  if (length(loci) == 1) {
    # Only compute the direction(s) we actually need.
    if (need_HvG) {
      HvG <- count_mismatches(HLA_mismatch_base(GL_string_recip, GL_string_donor, loci, "HvG", homozygous_count))
    }
    if (need_GvH) {
      GvH <- count_mismatches(HLA_mismatch_base(GL_string_recip, GL_string_donor, loci, "GvH", homozygous_count))
    }
    # Return the result for the requested direction.
    if (direction == "HvG" | direction == "SOT") {
      return(HvG)
    } else if (direction == "GvH") {
      return(GvH)
    } else {
      # Bidirectional: return the max of both directions.
      return(pmax(HvG, GvH, na.rm = TRUE))
    }
  } else {
    # Code to determine mismatch numbers if multiple loci were supplied.
    # Helper to build a mismatch count table from HLA_mismatch_base output.
    build_mm_table <- function(base_direction, col_name) {
      tibble(raw = HLA_mismatch_base(GL_string_recip, GL_string_donor, loci, base_direction, homozygous_count)) %>%
        # Add a row number to combine data at the end.
        mutate(case = row_number()) %>%
        # Separate the loci.
        separate_longer_delim(raw, delim = ", ") %>%
        separate_wider_delim(raw, delim = "=", names = c("locus", "mismatches")) %>%
        # Recode NA values to ensure accurate counting.
        mutate(mismatches = na_if(mismatches, "NA")) %>%
        # Count number of mismatches.
        mutate(!!col_name := count_mismatches(mismatches)) %>%
        # Clean up table.
        select(-mismatches)
    }

    # Only build the table(s) we actually need.
    if (direction == "bidirectional") {
      # Bidirectional needs both directions, joined together.
      HvG_table <- build_mm_table("HvG", "HvG_number")
      GvH_table <- build_mm_table("GvH", "GvH_number")
      # Join and take the max of both directions.
      MM_table <- HvG_table %>%
        left_join(GvH_table, join_by(locus, case)) %>%
        mutate(bidirectional = pmax(HvG_number, GvH_number, na.rm = TRUE))
      result_col <- "bidirectional"
    } else if (direction == "HvG" | direction == "SOT") {
      # Only need HvG direction.
      MM_table <- build_mm_table("HvG", "HvG_number")
      result_col <- "HvG_number"
    } else {
      # Only need GvH direction.
      MM_table <- build_mm_table("GvH", "GvH_number")
      result_col <- "GvH_number"
    }

    # Format the result as "Locus1=Count1, Locus2=Count2, ..." strings.
    MM_table <- MM_table %>%
      select(locus, case, all_of(result_col)) %>%
      unite(locus, all_of(result_col), col = "MM", sep = "=") %>%
      summarise(MM = str_flatten(MM, collapse = ", "), .by = case)
    return(MM_table$MM)
  }
}

globalVariables(c(
  "mismatches", "case", "HvG_number", "GvH_number", "MM", "bidirectional", ":="
))

library(testthat)
library(dplyr)
library(stringr)

test_that("HLA_mismatch_number correctly calculates mismatch counts", {
  GL_string_recip <- "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01^HLA-DRB1*13:01+HLA-DRB1*15:01"
  GL_string_donor <- "HLA-A*01:01+HLA-A*03:01^HLA-B*07:02+HLA-B*44:02^HLA-DRB1*13:01+HLA-DRB1*14:01"
  loci <- c("HLA-A", "HLA-B", "HLA-DRB1")

  result_HvG <- HLA_mismatch_number(GL_string_recip, GL_string_donor, loci, direction = "HvG")
  result_GvH <- HLA_mismatch_number(GL_string_recip, GL_string_donor, loci, direction = "GvH")
  result_bidirectional <- HLA_mismatch_number(GL_string_recip, GL_string_donor, loci, direction = "bidirectional")
  result_SOT <- HLA_mismatch_number(GL_string_recip, GL_string_donor, loci, direction = "SOT")

  expect_match(result_HvG, "HLA-A=1, HLA-B=1, HLA-DRB1=1")
  expect_match(result_GvH, "HLA-A=1, HLA-B=1, HLA-DRB1=1")
  expect_match(result_bidirectional, "HLA-A=1, HLA-B=1, HLA-DRB1=1")
  expect_match(result_SOT, "HLA-A=1, HLA-B=1, HLA-DRB1=1")

  result_single_locus <- HLA_mismatch_number(GL_string_recip, GL_string_donor, "HLA-A", direction = "HvG")
  expect_equal(result_single_locus, 1)

  result_homozygous_1 <- HLA_mismatch_number(GL_string_recip, GL_string_donor, "HLA-B", direction = "HvG", homozygous_count = 1)
  expect_equal(result_homozygous_1, 1)

  GL_string_recip <- "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01^HLA-DRB1*13:01+HLA-DRB1*15:01"
  GL_string_donor <- "HLA-A*01:01+HLA-A*03:01^HLA-B*44:02+HLA-B*44:02^HLA-DRB1*13:01+HLA-DRB1*14:01"
  result_homozygous_2 <- HLA_mismatch_number(GL_string_recip, GL_string_donor, "HLA-B", direction = "HvG", homozygous_count = 2)
  expect_equal(result_homozygous_2, 2)

  GL_string_recip_match <- "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01^HLA-DRB1*13:01+HLA-DRB1*15:01"
  GL_string_donor_match <- "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01^HLA-DRB1*13:01+HLA-DRB1*15:01"
  result_no_mismatch <- HLA_mismatch_number(GL_string_recip_match, GL_string_donor_match, loci, direction = "bidirectional")
  expect_match(result_no_mismatch, "HLA-A=0, HLA-B=0, HLA-DRB1=0")

  # Set the alleles
  A <- "HLA-A*01:01"
  B <- "HLA-A*02:05"
  C <- "HLA-A*24:02"
  D <- "HLA-A*31:03"
  N <- "HLA-A*68:11N"

  # Test the 2010 mismatch table.
  (mismatch_table_2010_test <- mismatch_table_2010
    %>% rename(Patient = patient, Donor = donor)
    # Assign the example alleles.
    %>% mutate(recipient_1 = case_when(
      str_detect(Patient, "^A") ~ A,
      str_detect(Patient, "^N") ~ N
    ), recipient_2 = case_when(
      str_detect(Patient, "A$") ~ A,
      str_detect(Patient, "B$") ~ B,
      str_detect(Patient, "N$") ~ N
    ), donor_1 = case_when(
      str_detect(Donor, "^A") ~ A,
      str_detect(Donor, "^B") ~ B,
      str_detect(Donor, "^C") ~ C,
      str_detect(Donor, "^N") ~ N
    ), donor_2 = case_when(
      str_detect(Donor, "A$") ~ A,
      str_detect(Donor, "B$") ~ B,
      str_detect(Donor, "C$") ~ C,
      str_detect(Donor, "D$") ~ D,
      str_detect(Donor, "N$") ~ N
    ))
    # Turn the two alleles into a GL string.
    %>% mutate(
      GL_string_recip = str_c(recipient_1, recipient_2, sep = "+"),
      GL_string_donor = str_c(donor_1, donor_2, sep = "+")
    )
    %>% select(Patient, Donor, GL_string_recip, GL_string_donor, "#GvH", "#HvG", "#Max")
    # Calculate MM numbers for each row - use homozygous_count = 1 for the 2010 table.
    %>% mutate(
      GvH_hc1 = HLA_mismatch_number(GL_string_recip, GL_string_donor, "HLA-A", "GvH", 1),
      HvG_hc1 = HLA_mismatch_number(GL_string_recip, GL_string_donor, "HLA-A", "HvG", 1),
      bidirectional_hc1 = HLA_mismatch_number(GL_string_recip, GL_string_donor, "HLA-A", "bidirectional", 1)
    )
    %>%
    rename(GvH = "#GvH", HvG = "#HvG", Max = "#Max") %>%
    # Check that the calculated results match the consensus results.
    mutate(
      GvH_hc1_result = (GvH == GvH_hc1),
      HvG_hc1_result = (HvG == HvG_hc1),
      bidirectional_hc1_result = (Max == bidirectional_hc1)
    ) %>%
    # Check that for each row all values are correct.
    mutate(total_result = if_all(GvH_hc1_result:bidirectional_hc1_result)) %>%
    # Summarize to a single value if all were true.
    distinct(total_result) %>%
    pull(total_result)
  )

  expect_equal(mismatch_table_2010_test, TRUE)

  # Test the 2016 mismatch table
  (mismatch_table_2016_test <- mismatch_table_2016
    # Assign the example alleles.
    %>% mutate(recipient_1 = case_when(
      str_detect(Patient, "^A") ~ A,
      str_detect(Patient, "^N") ~ N
    ), recipient_2 = case_when(
      str_detect(Patient, "A$") ~ A,
      str_detect(Patient, "B$") ~ B,
      str_detect(Patient, "N$") ~ N
    ), donor_1 = case_when(
      str_detect(Donor, "^A") ~ A,
      str_detect(Donor, "^B") ~ B,
      str_detect(Donor, "^C") ~ C,
      str_detect(Donor, "^N") ~ N
    ), donor_2 = case_when(
      str_detect(Donor, "A$") ~ A,
      str_detect(Donor, "B$") ~ B,
      str_detect(Donor, "C$") ~ C,
      str_detect(Donor, "D$") ~ D,
      str_detect(Donor, "N$") ~ N
    ))
    # Turn the two alleles into a GL string.
    %>% mutate(
      GL_string_recip = str_c(recipient_1, recipient_2, sep = "+"),
      GL_string_donor = str_c(donor_1, donor_2, sep = "+")
    )
    %>% select(Patient, Donor, GL_string_recip, GL_string_donor, "#GvH", "#HvG", "#Max")
    # Calculate MM numbers for each row - use homozygous_count = 2 for the 2016 table.
    %>% mutate(
      GvH_hc2 = HLA_mismatch_number(GL_string_recip, GL_string_donor, "HLA-A", "GvH", 2),
      HvG_hc2 = HLA_mismatch_number(GL_string_recip, GL_string_donor, "HLA-A", "HvG", 2),
      bidirectional_hc2 = HLA_mismatch_number(GL_string_recip, GL_string_donor, "HLA-A", "bidirectional", 2)
    )
    %>%
    rename(GvH = "#GvH", HvG = "#HvG", Max = "#Max") %>%
    # Check that the calculated results match the consensus results.
    mutate(
      GvH_hc2_result = (GvH == GvH_hc2),
      HvG_hc2_result = (HvG == HvG_hc2),
      bidirectional_hc2_result = (Max == bidirectional_hc2)
    ) %>%
    # Check that for each row all values are correct.
    mutate(total_result = if_all(GvH_hc2_result:bidirectional_hc2_result)) %>%
    # Summarize to a single value if all were true.
    distinct(total_result) %>%
    pull(total_result)
  )

  expect_equal(mismatch_table_2016_test, TRUE)
})


# --- Input validation tests ---

test_that("HLA_mismatch_number rejects NULL inputs", {
  gl <- "HLA-A*01:01+HLA-A*02:01"
  expect_error(HLA_mismatch_number(NULL, gl, "HLA-A", "HvG"), "GL_string_recip")
  expect_error(HLA_mismatch_number(gl, NULL, "HLA-A", "HvG"), "GL_string_donor")
  expect_error(HLA_mismatch_number(gl, gl, NULL, "HvG"), "loci")
})

test_that("HLA_mismatch_number rejects non-character GL strings", {
  gl <- "HLA-A*01:01+HLA-A*02:01"
  expect_error(HLA_mismatch_number(123, gl, "HLA-A", "HvG"), "must be a character")
})

test_that("HLA_mismatch_number rejects invalid direction", {
  gl <- "HLA-A*01:01+HLA-A*02:01"
  expect_error(HLA_mismatch_number(gl, gl, "HLA-A", "invalid"))
})

test_that("HLA_mismatch_number rejects invalid homozygous_count", {
  gl <- "HLA-A*01:01+HLA-A*02:01"
  expect_error(HLA_mismatch_number(gl, gl, "HLA-A", "HvG", homozygous_count = 3), "must be.*1.*or.*2")
})

test_that("HLA_mismatch_number returns 0 for perfect match at single locus", {
  gl <- "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01"
  result <- HLA_mismatch_number(gl, gl, "HLA-A", "bidirectional")
  expect_equal(result, 0)
})

test_that("HLA_mismatch_number works with vectorized inputs", {
  recip <- c(
    "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01",
    "HLA-A*03:01+HLA-A*24:02^HLA-B*15:01+HLA-B*40:01"
  )
  donor <- c(
    "HLA-A*01:01+HLA-A*03:01^HLA-B*07:02+HLA-B*08:01",
    "HLA-A*03:01+HLA-A*24:02^HLA-B*15:01+HLA-B*40:01"
  )
  # Single locus: should return integer vector of length 2
  result <- HLA_mismatch_number(recip, donor, "HLA-A", "HvG")
  expect_length(result, 2)
  expect_equal(result[1], 1)
  expect_equal(result[2], 0)

  # Multiple loci: should return character vector of length 2
  result_multi <- HLA_mismatch_number(recip, donor, c("HLA-A", "HLA-B"), "HvG")
  expect_length(result_multi, 2)
  expect_match(result_multi[1], "HLA-A=1, HLA-B=0")
  expect_match(result_multi[2], "HLA-A=0, HLA-B=0")
})

#' @title HLA_truncate
#'
#' @description This function truncates HLA typing values in molecular nomenclature
#' (for example from 4 fields to 2 fields). The truncation is based on the number
#' of fields specified and optionally retains any WHO-recognized suffixes
#' (L, S, C, A, Q, or N) or G and P group designations (G or P). This function
#' will work on individual alleles (e.g. "HLA-A*02:01:01:01") or on all alleles
#' in a GL string (e.g. "HLA-A*02:01:01:01+HLA-A*68:01:01^HLA-DRB1*01:01:01+HLA-DRB1*03:01:01").
#'
#' Note: depending on arguments used, this function can output HLA alleles that do not
#' exist in the IPD-IMGT/HLA database. For example, truncating the allele "DRB4*01:03:01:02N"
#' to 2 fields would result in "DRB4*01:03N," which does not exist in the IPD-IMGT/HLA database.
#' Users should take care in setting the parameters for this function.
#'
#'
#' @param data A string containing an HLA allele or a GL string.
#' @param fields An integer specifying the number of fields to retain in the
#' truncated values. Default is 2.
#' @param keep_suffix A logical value indicating whether to retain any
#' WHO-recognized suffixes. Default is TRUE.
#' @param keep_G_P_group A logical value indicating whether to retain any
#' G or P group designations. Default is FALSE.
#' @param remove_duplicates A logical value indicating whether to remove duplicated
#' values from a GL string after truncation. Default is FALSE.
#'
#' @return A string with the HLA typing truncated according to
#' the specified number of fields and optional suffix retention.
#'
#' @examples
#'
#' # The Haplotype_frequencies dataset contains a table with HLA typing spread across multiple columns:
#' print(Haplotype_frequencies)
#'
#' # The `HLA_truncate` function can be used to truncate the typing results to 2 fields:
#' library(dplyr)
#' Haplotype_frequencies %>% mutate(
#'   across(
#'     "HLA-A":"HLA-DPB1",
#'     ~ HLA_truncate(
#'       .,
#'       fields = 2,
#'       keep_suffix = TRUE,
#'       keep_G_P_group = FALSE
#'     )
#'   )
#' )
#'
#' @export
#'
#' @importFrom dplyr mutate
#' @importFrom dplyr across
#' @importFrom dplyr select
#' @importFrom dplyr %>%
#' @importFrom stringr str_extract
#' @importFrom tidyr replace_na
#' @importFrom dplyr na_if
#' @importFrom tidyr separate_wider_delim
#' @importFrom tidyr unite


HLA_truncate <- function(data, fields = 2, keep_suffix = TRUE, keep_G_P_group = FALSE, remove_duplicates = FALSE) {
  # Validate inputs
  check_gl_string(data, "data")
  check_fields(fields)
  check_logical_flag(keep_suffix, "keep_suffix")
  check_logical_flag(keep_G_P_group, "keep_G_P_group")
  check_logical_flag(remove_duplicates, "remove_duplicates")

  # Expand the GL string
  alleles <- GLstring_expand_longer(data) %>%
    # Extract any WHO-recognized suffixes
    mutate(suffix = replace_na(str_extract(value, "(?<=[:digit:])[LSCAQNlscaqn]$"), "")) %>%
    # Extract any P or G group designation
    mutate(GP = replace_na(str_extract(value, "(?<=[:digit:])[PGpg]$"), "")) %>%
    # Separate HLA prefix if available
    separate_wider_delim(value, delim = "-", names = c("prefix", "rest"), too_few = "align_end") %>%
    separate_wider_delim(rest, delim = "*", names = c("gene", "molecular_type"), too_few = "align_start") %>%
    # Separate molecular fields
    separate_wider_delim(molecular_type, delim = ":", names = c("one", "two", "three", "four"), too_few = "align_start") %>%
    # Keep only numbers in each field, in case there were non-standard suffixes.
    mutate(across(one:four, ~ str_extract(., "[:digit:]+")))

  # Determine which field columns to keep and which to drop based on requested fields.
  all_field_cols <- c("one", "two", "three", "four")
  keep_cols <- all_field_cols[seq_len(fields)]
  drop_cols <- setdiff(all_field_cols, keep_cols)

  # Drop the unneeded field columns.
  if (length(drop_cols) > 0) {
    alleles <- alleles %>% select(-all_of(drop_cols))
  }

  # Reunite the prefix and gene name (e.g. "HLA" + "A" -> "HLA-A").
  truncated <- alleles %>%
    unite(gene, prefix:gene, sep = "-", na.rm = TRUE)

  # Reunite the retained field columns into the allele code.
  if (fields == 1) {
    # Single field: unite directly with the gene using "*" separator.
    truncated <- truncated %>%
      unite(gene, gene, one, sep = "*", na.rm = TRUE)
  } else {
    # Multiple fields: join fields with ":", then unite with gene using "*".
    truncated <- truncated %>%
      unite(code, all_of(keep_cols), sep = ":", na.rm = TRUE) %>%
      mutate(code = na_if(code, "")) %>%
      unite(gene, gene, code, sep = "*", na.rm = TRUE)
  }

  # Retain or drop suffix and P/G group designation.
  if (keep_suffix) {
    truncated <- truncated %>% unite(gene, gene, suffix, sep = "", na.rm = TRUE)
  } else {
    truncated <- truncated %>% select(-suffix)
  }
  if (keep_G_P_group) {
    truncated <- truncated %>% unite(gene, gene, GP, sep = "", na.rm = TRUE)
  } else {
    truncated <- truncated %>% select(-GP)
  }

  # Combine everything back to a GL string.
  truncated %>%
    rename(value = gene) %>%
    ambiguity_table_to_GLstring(., remove_duplicates = remove_duplicates)
}


globalVariables(c(
  "rest", "molecular_type", "one", "four", "three",
  "two", "gene", "prefix", "code", "suffix", "GP"
))

library(testthat)
library(dplyr)
library(stringr)

test_that("HLA_truncate correctly truncates HLA alleles", {
  allele_4_field <- "HLA-A*01:01:01:02N"
  allele_3_field <- "HLA-B*07:02:01"
  allele_2_field <- "HLA-C*06:02"
  allele_g_group <- "HLA-DRB1*04:01:01G"

  expect_equal(HLA_truncate(allele_4_field), "HLA-A*01:01N")
  expect_equal(HLA_truncate(allele_3_field, fields = 3), "HLA-B*07:02:01")
  expect_equal(HLA_truncate(allele_2_field, fields = 1), "HLA-C*06")
  expect_equal(HLA_truncate(allele_g_group, keep_G_P_group = TRUE), "HLA-DRB1*04:01G")
  expect_equal(HLA_truncate(allele_4_field, keep_suffix = FALSE), "HLA-A*01:01")
  expect_equal(HLA_truncate(allele_4_field, fields = 1, keep_suffix = FALSE), "HLA-A*01")

  gl_string <- "HLA-A*02:01:01:01+HLA-A*68:01:01^HLA-B*07:02+HLA-B*44:02:01^HLA-DRB1*04:01:01+HLA-DRB1*13:01"
  expected_truncated <- "HLA-A*02:01+HLA-A*68:01^HLA-B*07:02+HLA-B*44:02^HLA-DRB1*04:01+HLA-DRB1*13:01"
  expect_equal(HLA_truncate(gl_string), expected_truncated)

  GL_string_duplicates <- "HLA-A*02:01:03/HLA-A*02:01:06"
  expect_equal(HLA_truncate(GL_string_duplicates, remove_duplicates = FALSE), "HLA-A*02:01/HLA-A*02:01")
  expect_equal(HLA_truncate(GL_string_duplicates, remove_duplicates = TRUE), "HLA-A*02:01")
})

#' @title HLA_mismatch_logical
#'
#' @description Determines if there are any mismatches between recipient and
#' donor HLA alleles for the specified loci. Returns `TRUE` if mismatches are
#' present, and `FALSE` otherwise.
#'
#' @param GL_string_recip A GL strings representing the recipient's HLA genotypes.
#' @param GL_string_donor A GL strings representing the donor's HLA genotypes.
#' @param loci A character vector specifying the loci to be considered for
#' mismatch calculation. HLA-DRB3/4/5 (and their serologic equivalents DR51/52/53)
#' are considered once locus for this function, and should be called in this argument
#' as "HLA-DRB3/4/5" or "HLA-DR51/52/53", respectively.
#' @param direction A character string indicating the direction of mismatch.
#' Options are "HvG" (host vs. graft), "GvH" (graft vs. host), "bidirectional"
#' (if either "HvG" or "GvH" is TRUE), or "SOT" (host vs. graft, as is used for
#' mismatching in solid organ transplantation).
#'
#' @return A logical value (`TRUE` or `FALSE`):
#' - `TRUE` if there are mismatches between recipient and donor HLA alleles.
#' - `FALSE` if there are no mismatches.
#'
#' @examples
#'
#' file <- HLA_typing_1[, -1]
#' GL_string <- HLA_columns_to_GLstring(file, HLA_typing_columns = everything())
#'
#' GL_string_recip <- GL_string[1]
#' GL_string_donor <- GL_string[2]
#'
#' loci <- c("HLA-A", "HLA-DRB3/4/5", "HLA-DPB1")
#' mismatches <- HLA_mismatch_logical(GL_string_recip, GL_string_donor, loci, direction = "HvG")
#' print(mismatches)
#'
#' @export
#'

HLA_mismatch_logical <- function(GL_string_recip, GL_string_donor, loci, direction) {
  # Validate inputs
  check_gl_string(GL_string_recip, "GL_string_recip")
  check_gl_string(GL_string_donor, "GL_string_donor")
  check_loci(loci)

  direction <- match.arg(direction, c("HvG", "GvH", "bidirectional", "SOT"))

  # Determine which direction(s) need to be computed.
  # HvG and SOT only need the HvG direction; GvH only needs GvH.
  # Bidirectional needs both to take the logical OR.
  need_HvG <- (direction == "HvG" | direction == "SOT" | direction == "bidirectional")
  need_GvH <- (direction == "GvH" | direction == "bidirectional")

  # Code to determine mismatch if a single locus was supplied.
  if (length(loci) == 1) {
    # Only compute the direction(s) we actually need.
    if (need_HvG) {
      HvG <- !is.na(HLA_mismatch_base(GL_string_recip, GL_string_donor, loci, "HvG"))
    }
    if (need_GvH) {
      GvH <- !is.na(HLA_mismatch_base(GL_string_recip, GL_string_donor, loci, "GvH"))
    }
    # Return the result for the requested direction.
    if (direction == "HvG" | direction == "SOT") {
      return(HvG)
    } else if (direction == "GvH") {
      return(GvH)
    } else {
      # Bidirectional: TRUE if either direction has a mismatch.
      return(HvG | GvH)
    }
  } else {
    # Code to determine mismatch if multiple loci were supplied.
    # Helper to build a logical mismatch table from HLA_mismatch_base output.
    build_mm_table <- function(base_direction, col_name) {
      tibble(raw = HLA_mismatch_base(GL_string_recip, GL_string_donor, loci, base_direction)) %>%
        # Add a row number to combine data at the end.
        mutate(case = row_number()) %>%
        # Separate the loci.
        separate_longer_delim(raw, delim = ", ") %>%
        separate_wider_delim(raw, delim = "=", names = c("locus", "mismatches")) %>%
        # Recode NA values to ensure accurate matching.
        mutate(mismatches = na_if(mismatches, "NA")) %>%
        # Determine if any mismatches are present.
        mutate(!!col_name := !is.na(mismatches)) %>%
        # Clean up table.
        select(-mismatches)
    }

    # Only build the table(s) we actually need.
    if (direction == "bidirectional") {
      # Bidirectional needs both directions, joined together.
      HvG_table <- build_mm_table("HvG", "HvG_MM")
      GvH_table <- build_mm_table("GvH", "GvH_MM")
      # Join and take the logical OR of both directions.
      MM_table <- HvG_table %>%
        left_join(GvH_table, join_by(locus, case)) %>%
        mutate(bidirectional = HvG_MM | GvH_MM)
      result_col <- "bidirectional"
    } else if (direction == "HvG" | direction == "SOT") {
      # Only need HvG direction.
      MM_table <- build_mm_table("HvG", "HvG_MM")
      result_col <- "HvG_MM"
    } else {
      # Only need GvH direction.
      MM_table <- build_mm_table("GvH", "GvH_MM")
      result_col <- "GvH_MM"
    }

    # Format the result as "Locus1=TRUE/FALSE, Locus2=TRUE/FALSE, ..." strings.
    MM_table <- MM_table %>%
      select(locus, case, all_of(result_col)) %>%
      unite(locus, all_of(result_col), col = "MM", sep = "=") %>%
      summarise(MM = str_flatten(MM, collapse = ", "), .by = case)

    return(MM_table$MM)
  }
}

globalVariables(c("HvG_MM", "GvH_MM", ":="))

library(testthat)
library(dplyr)
library(stringr)

test_that("HLA_mismatch_logical correctly identifies presence of mismatches", {
  GL_string_recip <- "HLA-A*03:01+HLA-A*74:01^HLA-DRB3*03:01^HLA-DRB5*02:21"
  GL_string_donor <- "HLA-A*03:02+HLA-A*20:01^HLA-DRB3*03:01"

  result_HvG <- HLA_mismatch_logical(GL_string_recip, GL_string_donor, loci = "HLA-A", direction = "HvG")
  expect_true(result_HvG)

  result_GvH <- HLA_mismatch_logical(GL_string_recip, GL_string_donor, loci = "HLA-A", direction = "GvH")
  expect_true(result_GvH)

  result_DRB3 <- HLA_mismatch_logical(GL_string_recip, GL_string_donor, loci = "HLA-DRB3/4/5", direction = "HvG")
  expect_false(result_DRB3)

  result_bidirectional <- HLA_mismatch_logical(GL_string_recip, GL_string_donor, loci = "HLA-A", direction = "bidirectional")
  expect_true(result_bidirectional)

  result_multi_loci <- HLA_mismatch_logical(GL_string_recip, GL_string_donor, loci = c("HLA-A", "HLA-DRB3/4/5"), direction = "HvG")
  expect_match(result_multi_loci, "HLA-A=TRUE")
  expect_match(result_multi_loci, "HLA-DRB3/4/5=FALSE")

  result_SOT <- HLA_mismatch_logical(GL_string_recip, GL_string_donor, loci = "HLA-A", direction = "SOT")
  expect_true(result_SOT)

  GL_string_donor_same <- "HLA-A*03:01+HLA-A*74:01^HLA-DRB3*03:01^HLA-DRB5*02:21"
  result_no_mismatch <- HLA_mismatch_logical(GL_string_recip, GL_string_donor_same, loci = c("HLA-A", "HLA-DRB3/4/5"), direction = "bidirectional")
  expect_match(result_no_mismatch, "HLA-A=FALSE")
  expect_match(result_no_mismatch, "HLA-DRB3/4/5=FALSE")
})


# --- Input validation tests ---

test_that("HLA_mismatch_logical rejects NULL inputs", {
  gl <- "HLA-A*01:01+HLA-A*02:01"
  expect_error(HLA_mismatch_logical(NULL, gl, "HLA-A", "HvG"), "GL_string_recip")
  expect_error(HLA_mismatch_logical(gl, NULL, "HLA-A", "HvG"), "GL_string_donor")
  expect_error(HLA_mismatch_logical(gl, gl, NULL, "HvG"), "loci")
})

test_that("HLA_mismatch_logical rejects non-character GL strings", {
  gl <- "HLA-A*01:01+HLA-A*02:01"
  expect_error(HLA_mismatch_logical(123, gl, "HLA-A", "HvG"), "must be a character")
})

test_that("HLA_mismatch_logical rejects invalid direction", {
  gl <- "HLA-A*01:01+HLA-A*02:01"
  expect_error(HLA_mismatch_logical(gl, gl, "HLA-A", "invalid"))
})

test_that("HLA_mismatch_logical returns FALSE for perfect match", {
  gl <- "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01"
  # Single locus
  expect_false(HLA_mismatch_logical(gl, gl, "HLA-A", "bidirectional"))
  # Multiple loci
  result <- HLA_mismatch_logical(gl, gl, c("HLA-A", "HLA-B"), "bidirectional")
  expect_match(result, "HLA-A=FALSE")
  expect_match(result, "HLA-B=FALSE")
})

test_that("HLA_mismatch_logical agrees with mismatch_table_2016 consensus (homozygous_count = 2)", {
  # Note: HLA_mismatch_logical does not expose a homozygous_count parameter —
  # it always uses the default of 2 via HLA_mismatch_base. Therefore,
  # only the 2016 table (which expects homozygous_count = 2) is tested here.
  # The 2010 table (homozygous_count = 1) is tested in test-HLA_mismatch_number.R.
  # Define symbolic alleles to substitute into the table.
  A <- "HLA-A*01:01"
  B <- "HLA-A*02:05"
  C <- "HLA-A*24:02"
  D <- "HLA-A*31:03"
  N <- "HLA-A*68:11N"

  # Build GL strings from the 2016 table's symbolic allele pairs.
  mismatch_table_2016_logical <- mismatch_table_2016 %>%
    # Assign the example alleles based on symbolic codes.
    mutate(recipient_1 = case_when(
      str_detect(Patient, "^A") ~ A,
      str_detect(Patient, "^N") ~ N
    ), recipient_2 = case_when(
      str_detect(Patient, "A$") ~ A,
      str_detect(Patient, "B$") ~ B,
      str_detect(Patient, "N$") ~ N
    ), donor_1 = case_when(
      str_detect(Donor, "^A") ~ A,
      str_detect(Donor, "^B") ~ B,
      str_detect(Donor, "^C") ~ C,
      str_detect(Donor, "^N") ~ N
    ), donor_2 = case_when(
      str_detect(Donor, "A$") ~ A,
      str_detect(Donor, "B$") ~ B,
      str_detect(Donor, "C$") ~ C,
      str_detect(Donor, "D$") ~ D,
      str_detect(Donor, "N$") ~ N
    )) %>%
    # Turn the two alleles into a GL string.
    mutate(
      GL_string_recip = str_c(recipient_1, recipient_2, sep = "+"),
      GL_string_donor = str_c(donor_1, donor_2, sep = "+")
    ) %>%
    select(Patient, Donor, GL_string_recip, GL_string_donor, "#GvH", "#HvG", "#Max") %>%
    rename(GvH = "#GvH", HvG = "#HvG", Max = "#Max") %>%
    # Calculate logical mismatch for each direction.
    mutate(
      GvH_logical = HLA_mismatch_logical(GL_string_recip, GL_string_donor, "HLA-A", "GvH"),
      HvG_logical = HLA_mismatch_logical(GL_string_recip, GL_string_donor, "HLA-A", "HvG"),
      bidir_logical = HLA_mismatch_logical(GL_string_recip, GL_string_donor, "HLA-A", "bidirectional")
    ) %>%
    # Derive expected logical values from the consensus mismatch counts.
    mutate(
      GvH_expected = (GvH > 0),
      HvG_expected = (HvG > 0),
      bidir_expected = (Max > 0)
    ) %>%
    # Check that calculated results match expected results.
    mutate(
      GvH_result = (GvH_logical == GvH_expected),
      HvG_result = (HvG_logical == HvG_expected),
      bidir_result = (bidir_logical == bidir_expected)
    ) %>%
    # Check that for each row all values are correct.
    mutate(total_result = if_all(GvH_result:bidir_result)) %>%
    # Summarize to a single value if all were true.
    distinct(total_result) %>%
    pull(total_result)

  expect_equal(mismatch_table_2016_logical, TRUE)
})

test_that("HLA_mismatch_logical handles GvH direction with multiple loci", {
  # This specifically tests the multi-loci GvH branch (lines 102-104),
  # which was not covered by existing tests that only used GvH with a single locus.
  GL_string_recip <- "HLA-A*03:01+HLA-A*74:01^HLA-B*07:02+HLA-B*08:01"
  GL_string_donor <- "HLA-A*03:02+HLA-A*20:01^HLA-B*07:02+HLA-B*08:01"

  result <- HLA_mismatch_logical(GL_string_recip, GL_string_donor,
                                  loci = c("HLA-A", "HLA-B"), direction = "GvH")

  # HLA-A should be mismatched (donor has different alleles).
  expect_match(result, "HLA-A=TRUE")
  # HLA-B should match (same alleles).
  expect_match(result, "HLA-B=FALSE")
})

test_that("HLA_mismatch_logical works with vectorized inputs", {
  recip <- c(
    "HLA-A*01:01+HLA-A*02:01",
    "HLA-A*03:01+HLA-A*24:02"
  )
  donor <- c(
    "HLA-A*01:01+HLA-A*03:01",
    "HLA-A*03:01+HLA-A*24:02"
  )
  # Single locus: should return logical vector of length 2
  result <- HLA_mismatch_logical(recip, donor, "HLA-A", "HvG")
  expect_length(result, 2)
  expect_true(result[1])
  expect_false(result[2])
})

#' Synthetic HLA typing data for 10 individuals for the HLA-A, B, C, DRB1, DRB3/4/5,
#' DQB1, DQA1, DPB1 and DPA1 loci.
#'
#' @docType data
#'
#' @usage data(HLA_typing_1)
#'
#' @format
#' A tibble:
#' \describe{
#'   \item{patient}{Patients 1-10}
#'   \item{A1:DPB1_2}{HLA allele names}
#'     ...
#' }
#'
#' @keywords datasets
#'
#' @source Synthetic data from Nicholas K. Brown.
#'
"HLA_typing_1"

#' @title GLstring_to_ambiguity_table
#'
#' @description A function that expands a GL string into a longer, more detailed
#' format (also known as an ambiguity table) by separating the string into its
#' components resulting from its hierarchical set of operators, including gene
#' locations, loci, genotypes, haplotypes, and alleles. The function processes
#' each level of the GL string and assigns identifiers for each hierarchical
#' component. The resulting table can be assembled back into a GL string using
#' the function `ambiguity_table_to_GLstring`. This function is an alias of
#' `GLstring_expand_longer`.
#'
#' @param GL_string A GL string that encodes HLA alleles and their potential
#' ambiguities
#'
#' @return A tibble that contains the expanded GL string with separate columns
#' for possible gene locations, loci, genotype ambiguities, genotypes, haplotypes,
#' and alleles, each with associated identifiers
#'
#' @examples
#' file <- HLA_typing_1[, -1]
#' GL_string <- HLA_columns_to_GLstring(file, HLA_typing_columns = everything())
#' result <- GLstring_to_ambiguity_table(GL_string[1])
#' print(result)
#'
#' @export
#'
#'


GLstring_to_ambiguity_table <- function(GL_string) {
  GLstring_expand_longer(GL_string)
}

#' @keywords internal
"_PACKAGE"

#' @section GL string functions:
#' Functions for working with genotype list (GL) strings:
#' \itemize{
#'   \item \code{\link{HLA_columns_to_GLstring}}: Convert tabular HLA data to GL strings
#'   \item \code{\link{GLstring_genes}}: Split a GL string into locus columns
#'   \item \code{\link{GLstring_genes_expanded}}: Split a GL string into one allele per row
#'   \item \code{\link{GLstring_expand_longer}}: Expand a GL string into an ambiguity table
#'   \item \code{\link{ambiguity_table_to_GLstring}}: Collapse an ambiguity table back to a GL string
#'   \item \code{\link{GLstring_genotype_ambiguity}}: Extract genotype ambiguity from GL strings
#'   \item \code{\link{GLstring_gene_copies_combine}}: Combine gene copies in a GL string
#'   \item \code{\link{GLstring_regex}}: Create regex patterns for accurate GL string searching
#' }
#'
#' @section Matching and mismatching:
#' Functions for calculating HLA matching between recipient and donor:
#' \itemize{
#'   \item \code{\link{HLA_mismatch_base}}: Core mismatch engine (used internally by other functions)
#'   \item \code{\link{HLA_mismatch_logical}}: Is there a mismatch? (TRUE/FALSE)
#'   \item \code{\link{HLA_mismatch_number}}: How many mismatches?
#'   \item \code{\link{HLA_mismatched_alleles}}: Which alleles are mismatched?
#'   \item \code{\link{HLA_match_number}}: How many matches?
#'   \item \code{\link{HLA_match_summary_HCT}}: Standard match grades for hematopoietic cell transplantation
#' }
#'
#' @section Allele name utilities:
#' Functions for manipulating HLA allele names:
#' \itemize{
#'   \item \code{\link{HLA_truncate}}: Truncate alleles to a specified number of fields
#'   \item \code{\link{HLA_prefix_add}}: Add HLA and locus prefixes
#'   \item \code{\link{HLA_prefix_remove}}: Remove HLA and locus prefixes
#'   \item \code{\link{HLA_validate}}: Validate and clean HLA allele names
#'   \item \code{\link{HLA_column_repair}}: Convert column names between WHO and tidyverse formats
#' }
#'
#' @section File I/O:
#' \itemize{
#'   \item \code{\link{read_HML}}: Read GL strings from HML (HLA Markup Language) files
#' }
#'
#' @section Datasets:
#' \itemize{
#'   \item \code{\link{HLA_typing_1}}: Synthetic HLA typing for 10 individuals
#'   \item \code{\link{HLA_typing_LIS}}: Synthetic LIS-formatted HLA typing
#'   \item \code{\link{HLA_dictionary}}: HLA allele dictionary (2008)
#'   \item \code{\link{Haplotype_frequencies}}: HLA haplotype frequencies
#'   \item \code{\link{mismatch_table_2010}}: Consensus mismatch reference table (2010)
#'   \item \code{\link{mismatch_table_2016}}: Consensus mismatch reference table (2016)
#' }
NULL

#' @title GLstring_expand_longer
#'
#' @description A function that expands a GL string into a longer, more detailed
#' format (also known as an ambiguity table) by separating the string into its
#' components resulting from its hierarchical set of operators, including gene
#' locations, loci, genotypes, haplotypes, and alleles. The function processes
#' each level of the GL string and assigns identifiers for each hierarchical
#' component. The resulting table can be assembled back into a GL string using
#' the function `ambiguity_table_to_GLstring`.
#'
#' @param GL_string A GL string that encodes HLA alleles and their potential
#' ambiguities
#'
#' @return A tibble that contains the expanded GL string with separate columns
#' for possible gene locations, loci, genotype ambiguities, genotypes, haplotypes,
#' and alleles, each with associated identifiers
#'
#' @examples
#' file <- HLA_typing_1[, -1]
#' GL_string <- HLA_columns_to_GLstring(file, HLA_typing_columns = everything())
#' result <- GLstring_expand_longer(GL_string[1])
#' print(result)
#'
#' @export
#'
#' @importFrom dplyr mutate
#' @importFrom dplyr row_number
#' @importFrom dplyr %>%
#' @importFrom tidyr separate_longer_delim
#' @importFrom tibble as_tibble
#'


GLstring_expand_longer <- function(GL_string) {
  # Validate input
  check_gl_string(GL_string, "GL_string")

  as_tibble(GL_string) %>%
    # Assign a unique identifier for each entry for the function
    mutate(entry = row_number()) %>%
    # Separate GL string precedence 0: possible gene locations
    separate_longer_delim(value, delim = "?") %>%
    # Assign a unique identifier for each possible gene location
    mutate(possible_gene_location = row_number(), .by = entry) %>%
    # Separate GL string precedence 1: gene/locus
    separate_longer_delim(value, delim = "^") %>%
    # Identify locus
    mutate(locus = row_number(), .by = c(entry, possible_gene_location)) %>%
    # Separate GL string precedence 2: genotype list
    separate_longer_delim(value, delim = "|") %>%
    # Identify genotype ambiguities
    mutate(genotype_ambiguity = row_number(), .by = c(entry, possible_gene_location, locus)) %>%
    # Separate GL string precedence 3: genotype
    separate_longer_delim(value, delim = "+") %>%
    # Identify genotypes
    mutate(genotype = row_number(), .by = c(entry, possible_gene_location, locus, genotype_ambiguity)) %>%
    # Separate GL string precedence 4: haplotype
    separate_longer_delim(value, delim = "~") %>%
    # Identify haplotypes
    mutate(haplotype = row_number(), .by = c(entry, possible_gene_location, locus, genotype_ambiguity, genotype)) %>%
    # Separate GL string precedence 5: allele list
    separate_longer_delim(value, delim = "/") %>%
    # Identify alleles
    mutate(allele = row_number(), .by = c(entry, possible_gene_location, locus, genotype_ambiguity, genotype, haplotype))
}

globalVariables(c(
  "value", "entry", "possible_gene_location", "locus",
  "genotype_ambiguity", "genotype", "haplotype", "allele"
))

library(testthat)
library(dplyr)
library(tidyr)
library(stringr)

test_that("GLstring_expand_longer expands a GL string correctly", {
  GL_string <- "HLA-A*01:01:01:01/HLA-A*01:02/HLA-A*01:03/HLA-A*01:95+HLA-A*24:02:01:01|HLA-A*01:01:01:01/HLA-A*01:03+HLA-A*24:03:01:01^HLA-B*07:01:01+B*15:01:01/B*15:02:01|B*07:03+B*15:99:01^HLA-DRB1*03:01:02~HLA-DRB5*01:01:01+HLA-KIR2DL5A*0010101+HLA-KIR2DL5A*0010201?HLA-KIR2DL5B*0010201+HLA-KIR2DL5B*0010301"
  result <- GLstring_expand_longer(GL_string)

  expect_s3_class(result, "tbl_df")
  expect_true("locus" %in% colnames(result))
  expect_true("allele" %in% colnames(result))
  expect_gt(nrow(result), 0)
})

#' Consensus mismatch numbers for every possible allele combination
#' at a single locus, from the 2016 publication.
#'
#' @docType data
#'
#' @usage data(mismatch_table_2016)
#'
#' @format
#' A tibble:
#' \describe{
#'   \item{Patient:Donor}{allele combinations for patient and donor}
#'   \item{#GvH:#Max}{consensus mismatch number for each direction}
#'     ...
#' }
#'
#' @keywords datasets
#'
#' @source <https://www.nature.com/articles/bmt2010132>
#'
"mismatch_table_2016"

#' @title HLA_column_repair
#'
#' @description This function will change column names that have the official
#'  HLA nomenclature (e.g. "HLA-A*" or "HLA-A") to a format more easily selected
#'  in tidyverse functions (e.g. "HLA_A"). The dash and asterisk are special
#'  characters in R, and makes selecting columns by name difficult. This function
#'  will also allow for easily changing back to WHO-compliant nomenclature
#'  (e.g. "HLA-A*").
#'
#' @param data A data frame
#' @param format Either "tidyverse" or "WHO".
#' @param asterisk Logical value to return column with an asterisk.
#'
#' @return A data frame object with column names renamed in the specified format.
#'
#' @examples
#' HLA_type <- data.frame(
#'   "HLA-A*" = c("01:01", "02:01"),
#'   "HLA-B*" = c("07:02", "08:01"),
#'   "HLA-C*" = c("03:04", "04:01"),
#'   stringsAsFactors = FALSE
#' )
#'
#' HLA_column_repair(HLA_type, format = "tidyverse")
#'
#' @export
#'
#' @importFrom cli cli_abort
#' @importFrom dplyr rename_with
#' @importFrom dplyr %>%
#' @importFrom dplyr starts_with
#' @importFrom stringr str_replace


HLA_column_repair <- function(data, format = "tidyverse", asterisk = FALSE) {
  # Validate inputs
  check_data_frame(data, "data")
  check_logical_flag(asterisk, "asterisk")

  # Step 1: turn "HLA-A" to "HLA_A" or vice versa.
  if (format == "tidyverse") {
    step_1 <- data %>% rename_with(~ str_replace(., "HLA\\-", "HLA_"))
  } else if (format == "WHO") {
    step_1 <- data %>% rename_with(~ str_replace(., "HLA_", "HLA-"))
  } else {
    cli_abort("{.arg format} must be {.val tidyverse} or {.val WHO}.")
  }
  # Step 2: remove asterisk from all columns.
  step_2 <- step_1 %>% rename_with(~ str_replace(., "\\*$", ""))
  # Step 3: add asterisk back if wanted.
  if (asterisk == TRUE) {
    step_3 <- step_2 %>% rename_with(~ str_replace(., "$", "*"), starts_with("HLA"))
  } else {
    step_3 <- step_2
  }
  return(step_3)
}

library(testthat)
library(immunogenetr)
library(dplyr)

test_that("HLA_column_repair() returns a data frame", {
  # Set up test data with WHO-style column names including asterisks.
  output_table <- HLA_column_repair(
    tibble(
      "HLA-A*" = c("01:01", "02:01"),
      "HLA-B*" = c("07:02", "08:01"),
      "HLA-C*" = c("03:04", "04:01")
    )
  )
  expect_s3_class(output_table, "data.frame")
})

test_that("HLA_column_repair() converts WHO format to tidyverse format", {
  # Test data with WHO-style dash and asterisk in column names.
  who_data <- tibble(
    "HLA-A*" = c("01:01", "02:01"),
    "HLA-B*" = c("07:02", "08:01"),
    "HLA-DRB1*" = c("03:01", "04:01")
  )

  # Default format is "tidyverse": replaces dash with underscore and removes asterisk.
  result <- HLA_column_repair(who_data, format = "tidyverse")
  expect_true(all(c("HLA_A", "HLA_B", "HLA_DRB1") %in% colnames(result)))
  # Verify no dashes or asterisks remain in HLA column names.
  hla_cols <- grep("^HLA", colnames(result), value = TRUE)
  expect_false(any(grepl("-", hla_cols)))
  expect_false(any(grepl("\\*", hla_cols)))
})

test_that("HLA_column_repair() converts tidyverse format to WHO format", {
  # Test data with tidyverse-style underscore column names.
  tidy_data <- tibble(
    "HLA_A" = c("01:01", "02:01"),
    "HLA_B" = c("07:02", "08:01"),
    "HLA_C" = c("03:04", "04:01")
  )

  # WHO format: replaces underscore with dash.
  result <- HLA_column_repair(tidy_data, format = "WHO")
  expect_true(all(c("HLA-A", "HLA-B", "HLA-C") %in% colnames(result)))
  # Verify no underscores in HLA column names.
  hla_cols <- grep("^HLA", colnames(result), value = TRUE)
  expect_false(any(grepl("_", hla_cols)))
})

test_that("HLA_column_repair() correctly handles the asterisk parameter", {
  # Test data without asterisks.
  tidy_data <- tibble(
    "HLA_A" = c("01:01"),
    "HLA_B" = c("07:02")
  )

  # asterisk = TRUE should add asterisk to all HLA columns.
  result_with <- HLA_column_repair(tidy_data, format = "tidyverse", asterisk = TRUE)
  expect_true(all(c("HLA_A*", "HLA_B*") %in% colnames(result_with)))

  # asterisk = FALSE (default) should not add asterisk.
  result_without <- HLA_column_repair(tidy_data, format = "tidyverse", asterisk = FALSE)
  expect_false(any(grepl("\\*", colnames(result_without))))
})

test_that("HLA_column_repair() adds asterisk in WHO format", {
  # Test combining WHO format with asterisk.
  tidy_data <- tibble(
    "HLA_A" = c("01:01"),
    "HLA_DRB1" = c("03:01")
  )

  # WHO format + asterisk should produce "HLA-A*" and "HLA-DRB1*".
  result <- HLA_column_repair(tidy_data, format = "WHO", asterisk = TRUE)
  expect_true(all(c("HLA-A*", "HLA-DRB1*") %in% colnames(result)))
})

test_that("HLA_column_repair() removes existing asterisks before optionally re-adding", {
  # Test data that already has asterisks.
  who_data <- tibble(
    "HLA-A*" = c("01:01"),
    "HLA-B*" = c("07:02")
  )

  # tidyverse with asterisk = FALSE: asterisk should be removed.
  result_no_ast <- HLA_column_repair(who_data, format = "tidyverse", asterisk = FALSE)
  expect_true(all(c("HLA_A", "HLA_B") %in% colnames(result_no_ast)))

  # tidyverse with asterisk = TRUE: asterisk should be present.
  result_ast <- HLA_column_repair(who_data, format = "tidyverse", asterisk = TRUE)
  expect_true(all(c("HLA_A*", "HLA_B*") %in% colnames(result_ast)))
})

test_that("HLA_column_repair() preserves non-HLA columns unchanged", {
  # Data with a mix of HLA and non-HLA columns.
  mixed_data <- tibble(
    patient = c("P1", "P2"),
    "HLA-A*" = c("01:01", "02:01"),
    age = c(30, 45)
  )

  result <- HLA_column_repair(mixed_data, format = "tidyverse")
  # Non-HLA columns should be unchanged.
  expect_true("patient" %in% colnames(result))
  expect_true("age" %in% colnames(result))
  # HLA column should be converted.
  expect_true("HLA_A" %in% colnames(result))
})

test_that("HLA_column_repair() preserves data values", {
  # Verify that column renaming doesn't alter the data itself.
  who_data <- tibble(
    "HLA-A*" = c("01:01", "02:01"),
    "HLA-B*" = c("07:02", "08:01")
  )

  result <- HLA_column_repair(who_data, format = "tidyverse")
  expect_equal(result$HLA_A, c("01:01", "02:01"))
  expect_equal(result$HLA_B, c("07:02", "08:01"))
})

test_that("HLA_column_repair() errors on invalid format parameter", {
  test_data <- tibble("HLA-A*" = c("01:01"))
  # Invalid format should produce an error.
  expect_error(HLA_column_repair(test_data, format = "invalid"))
})

test_that("HLA_column_repair() validates inputs", {
  # NULL data should error.
  expect_error(HLA_column_repair(NULL))
  # Non-data-frame should error.
  expect_error(HLA_column_repair("not a data frame"))
  # Non-logical asterisk should error.
  test_data <- tibble("HLA-A" = c("01:01"))
  expect_error(HLA_column_repair(test_data, asterisk = "yes"))
})

test_that("HLA_column_repair() handles columns without HLA prefix", {
  # Data with no HLA columns; should return unchanged.
  no_hla_data <- tibble(
    patient = c("P1"),
    gene = c("TP53")
  )
  result <- HLA_column_repair(no_hla_data, format = "tidyverse")
  expect_equal(colnames(result), c("patient", "gene"))
})

# Internal validation helper functions for immunogenetr
# These are not exported; they provide consistent input checking across the package.

#' @importFrom cli cli_abort
#' @importFrom rlang caller_env

# Validate that a GL string argument is a non-empty character vector.
# Allows NA values within the vector (handled downstream), but rejects
# NULL, non-character types, and zero-length input.
check_gl_string <- function(x, arg_name = "data", call = caller_env()) {
  # NULL check
  if (is.null(x)) {
    cli_abort("{.arg {arg_name}} must be a character vector, not {.cls NULL}.", call = call)
  }
  # Allow bare NA (logical) by treating it as NA_character_; a scalar NA or

  # vector of all NA is a valid "no data" input that downstream code handles.
  if (is.logical(x) && all(is.na(x))) {
    return(invisible(x))
  }
  # Type check
  if (!is.character(x)) {
    cli_abort("{.arg {arg_name}} must be a character vector, not {.cls {class(x)}}.", call = call)
  }
  # Length check
  if (length(x) == 0) {
    cli_abort("{.arg {arg_name}} must have length >= 1, not 0.", call = call)
  }
  invisible(x)
}

# Validate that a data argument is a non-empty data frame.
check_data_frame <- function(x, arg_name = "data", call = caller_env()) {
  # NULL check
  if (is.null(x)) {
    cli_abort("{.arg {arg_name}} must be a data frame, not {.cls NULL}.", call = call)
  }
  # Type check
  if (!is.data.frame(x)) {
    cli_abort("{.arg {arg_name}} must be a data frame, not {.cls {class(x)}}.", call = call)
  }
  # Row check
  if (nrow(x) == 0) {
    cli_abort("{.arg {arg_name}} must have at least one row.", call = call)
  }
  invisible(x)
}

# Validate that a loci argument is a non-empty character vector of locus names.
check_loci <- function(x, arg_name = "loci", call = caller_env()) {
  # NULL check
  if (is.null(x)) {
    cli_abort("{.arg {arg_name}} must be a character vector of locus names, not {.cls NULL}.", call = call)
  }
  # Type check
  if (!is.character(x)) {
    cli_abort("{.arg {arg_name}} must be a character vector, not {.cls {class(x)}}.", call = call)
  }
  # Length check
  if (length(x) == 0) {
    cli_abort("{.arg {arg_name}} must have length >= 1, not 0.", call = call)
  }
  invisible(x)
}

# Validate that a scalar logical argument is TRUE or FALSE (not NA, NULL, or non-logical).
check_logical_flag <- function(x, arg_name, call = caller_env()) {
  if (is.null(x) || !is.logical(x) || length(x) != 1 || is.na(x)) {
    cli_abort("{.arg {arg_name}} must be {.val TRUE} or {.val FALSE}.", call = call)
  }
  invisible(x)
}

# Validate that homozygous_count is 1 or 2.
check_homozygous_count <- function(x, call = caller_env()) {
  if (is.null(x) || length(x) != 1 || !x %in% c(1, 2)) {
    cli_abort("{.arg homozygous_count} must be {.val 1} or {.val 2}.", call = call)
  }
  invisible(x)
}

# Validate that fields is an integer between 1 and 4.
check_fields <- function(x, call = caller_env()) {
  if (is.null(x) || length(x) != 1 || !x %in% 1:4) {
    cli_abort("{.arg fields} must be an integer between 1 and 4.", call = call)
  }
  invisible(x)
}

library(testthat)

# ===========================================================================
# Tests for check_gl_string
# ===========================================================================

test_that("check_gl_string accepts valid character vectors", {
  # Single GL string
  expect_invisible(check_gl_string("HLA-A*01:01+HLA-A*02:01"))
  # Vector of GL strings

  expect_invisible(check_gl_string(c("HLA-A*01:01", "HLA-B*07:02")))
  # Character vector with an embedded NA (valid; downstream handles NAs)
  expect_invisible(check_gl_string(c("HLA-A*01:01", NA_character_)))
  # Bare NA (logical) is allowed as a special case
  expect_invisible(check_gl_string(NA))
})

test_that("check_gl_string rejects NULL", {
  expect_error(check_gl_string(NULL), "must be a character vector, not.*NULL")
})

test_that("check_gl_string rejects non-character types", {
  # Numeric input
  expect_error(check_gl_string(123), "must be a character vector, not.*numeric")
  # Logical TRUE/FALSE (not bare NA)
  expect_error(check_gl_string(TRUE), "must be a character vector, not.*logical")
  # List input
  expect_error(check_gl_string(list("HLA-A*01:01")), "must be a character vector, not.*list")
  # Data frame input
  expect_error(check_gl_string(data.frame(x = "a")), "must be a character vector")
})

test_that("check_gl_string rejects zero-length character vector", {
  expect_error(check_gl_string(character(0)), "must have length >= 1, not 0")
})

test_that("check_gl_string reports the correct argument name", {
  # Default argument name
  expect_error(check_gl_string(NULL), "data")
  # Custom argument name
  expect_error(check_gl_string(NULL, arg_name = "GL_string_recip"), "GL_string_recip")
})

# ===========================================================================
# Tests for check_data_frame
# ===========================================================================

test_that("check_data_frame accepts valid data frames", {
  # Standard data frame
  expect_invisible(check_data_frame(data.frame(x = 1, y = 2)))
  # Tibble
  expect_invisible(check_data_frame(tibble::tibble(x = 1, y = 2)))
})

test_that("check_data_frame rejects NULL", {
  expect_error(check_data_frame(NULL), "must be a data frame, not.*NULL")
})

test_that("check_data_frame rejects non-data-frame types", {
  # Character vector
  expect_error(check_data_frame("not a dataframe"), "must be a data frame, not.*character")
  # Numeric vector
  expect_error(check_data_frame(c(1, 2, 3)), "must be a data frame, not.*numeric")
  # List (not a data frame)
  expect_error(check_data_frame(list(x = 1)), "must be a data frame, not.*list")
  # Matrix
  expect_error(check_data_frame(matrix(1:4, 2, 2)), "must be a data frame, not.*matrix")
})

test_that("check_data_frame rejects empty data frames (0 rows)", {
  expect_error(check_data_frame(data.frame(x = character(0))), "must have at least one row")
})

test_that("check_data_frame reports the correct argument name", {
  expect_error(check_data_frame(NULL), "data")
  expect_error(check_data_frame(NULL, arg_name = "my_table"), "my_table")
})

# ===========================================================================
# Tests for check_loci
# ===========================================================================

test_that("check_loci accepts valid locus vectors", {
  # Single locus
  expect_invisible(check_loci("HLA-A"))
  # Multiple loci
  expect_invisible(check_loci(c("HLA-A", "HLA-B", "HLA-DRB1")))
  # DRB3/4/5 combined locus
  expect_invisible(check_loci("HLA-DRB3/4/5"))
})

test_that("check_loci rejects NULL", {
  expect_error(check_loci(NULL), "must be a character vector of locus names, not.*NULL")
})

test_that("check_loci rejects non-character types", {
  # Numeric input
  expect_error(check_loci(1), "must be a character vector, not.*numeric")
  # Logical input
  expect_error(check_loci(TRUE), "must be a character vector, not.*logical")
})

test_that("check_loci rejects zero-length character vector", {
  expect_error(check_loci(character(0)), "must have length >= 1, not 0")
})

# ===========================================================================
# Tests for check_logical_flag
# ===========================================================================

test_that("check_logical_flag accepts TRUE and FALSE", {
  expect_invisible(check_logical_flag(TRUE, "my_flag"))
  expect_invisible(check_logical_flag(FALSE, "my_flag"))
})

test_that("check_logical_flag rejects NULL", {
  expect_error(check_logical_flag(NULL, "my_flag"), "must be.*TRUE.*or.*FALSE")
})

test_that("check_logical_flag rejects NA", {
  expect_error(check_logical_flag(NA, "my_flag"), "must be.*TRUE.*or.*FALSE")
})

test_that("check_logical_flag rejects non-logical types", {
  # String "TRUE"
  expect_error(check_logical_flag("TRUE", "my_flag"), "must be.*TRUE.*or.*FALSE")
  # Numeric 1
  expect_error(check_logical_flag(1, "my_flag"), "must be.*TRUE.*or.*FALSE")
})

test_that("check_logical_flag rejects logical vectors of length > 1", {
  expect_error(check_logical_flag(c(TRUE, FALSE), "my_flag"), "must be.*TRUE.*or.*FALSE")
})

test_that("check_logical_flag reports the correct argument name", {
  expect_error(check_logical_flag(NULL, "keep_suffix"), "keep_suffix")
})

# ===========================================================================
# Tests for check_homozygous_count
# ===========================================================================

test_that("check_homozygous_count accepts 1 and 2", {
  expect_invisible(check_homozygous_count(1))
  expect_invisible(check_homozygous_count(2))
})

test_that("check_homozygous_count rejects NULL", {
  expect_error(check_homozygous_count(NULL), "must be.*1.*or.*2")
})

test_that("check_homozygous_count rejects invalid values", {
  # Zero
  expect_error(check_homozygous_count(0), "must be.*1.*or.*2")
  # Three
  expect_error(check_homozygous_count(3), "must be.*1.*or.*2")
  # Negative
  expect_error(check_homozygous_count(-1), "must be.*1.*or.*2")
})

test_that("check_homozygous_count rejects vectors of length > 1", {
  expect_error(check_homozygous_count(c(1, 2)), "must be.*1.*or.*2")
})

# ===========================================================================
# Tests for check_fields
# ===========================================================================

test_that("check_fields accepts integers 1 through 4", {
  expect_invisible(check_fields(1))
  expect_invisible(check_fields(2))
  expect_invisible(check_fields(3))
  expect_invisible(check_fields(4))
})

test_that("check_fields rejects NULL", {
  expect_error(check_fields(NULL), "must be an integer between 1 and 4")
})

test_that("check_fields rejects out-of-range values", {
  # Zero
  expect_error(check_fields(0), "must be an integer between 1 and 4")
  # Five
  expect_error(check_fields(5), "must be an integer between 1 and 4")
  # Negative
  expect_error(check_fields(-1), "must be an integer between 1 and 4")
})

test_that("check_fields rejects non-numeric types", {
  # String that doesn't coerce to a valid value
  expect_error(check_fields("abc"), "must be an integer between 1 and 4")
})

test_that("check_fields rejects vectors of length > 1", {
  expect_error(check_fields(c(1, 2)), "must be an integer between 1 and 4")
})

#' @title HLA_match_summary_HCT
#'
#' @description Calculates the match summary for either the HLA-A, B, C and DRB1
#' loci (out-of-8 matching) or the HLA-A, B, C, DRB1 and DQB1 loci (out-of-10 matching),
#' as is commonly used for hematopoietic cell transplantation (HCT). Homozygous
#' mismatches are counted twice. Bidirectional matching is the default, but can
#' be overridden with the "direction" argument.
#'
#' @param GL_string_recip A GL string representing the recipient's HLA genotype,
#' and minimally containing the HLA-A, B, C and DRB1 loci (for Xof8 matching)
#' or the HLA-A, B, C, DRB1 and DQB1 loci (for Xof10 matching).
#' @param GL_string_donor A GL string representing the donor's HLA genotype,
#' and minimally containing the HLA-A, B, C and DRB1 loci (for Xof8 matching)
#' or the HLA-A, B, C, DRB1 and DQB1 loci (for Xof10 matching).
#' @param direction "GvH", "HvG" or "bidirectional". Default is "bidirectional".
#' @param match_grade "Xof8" for HLA-A, B, C and DRB1 matching or "Xof10" for
#' HLA-A, B, C, DRB1 and DQB1 matching.
#' @param scope "locus" or "genotype". Default is "locus". When "locus" (the default),
#' bidirectional matching takes the minimum match count at each locus independently
#' before summing. When "genotype", the GvH and HvG match summaries are calculated
#' separately across all loci, and the maximum of the two totals is returned.
#' Only affects results when direction is "bidirectional"; ignored otherwise.
#'
#' @return An integer value of the match grade summary.
#'
#' @examples
#' # Example recipient and donor GL strings
#' file <- HLA_typing_1[, -1]
#' GL_string <- HLA_columns_to_GLstring(file, HLA_typing_columns = everything())
#'
#' GL_string_recip <- GL_string[1]
#' GL_string_donor <- GL_string[2]
#'
#' # Calculate mismatch numbers
#' HLA_match_summary_HCT(GL_string_recip, GL_string_donor,
#'   direction = "bidirectional", match_grade = "Xof8"
#' )
#'
#' # Genotype-level bidirectional matching (max of GvH and HvG totals)
#' HLA_match_summary_HCT(GL_string_recip, GL_string_donor,
#'   direction = "bidirectional", match_grade = "Xof8", scope = "genotype"
#' )
#'
#' @importFrom purrr map_int
#' @importFrom stringr str_extract_all
#' @export
#'

HLA_match_summary_HCT <- function(GL_string_recip, GL_string_donor, direction = "bidirectional", match_grade, scope = "locus") {
  # Validate inputs
  check_gl_string(GL_string_recip, "GL_string_recip")
  check_gl_string(GL_string_donor, "GL_string_donor")
  match_grade <- match.arg(match_grade, c("Xof8", "Xof10"))
  direction <- match.arg(direction, c("HvG", "GvH", "bidirectional"))
  scope <- match.arg(scope, c("locus", "genotype"))

  # Determine loci based on match grade
  loci <- if (match_grade == "Xof8") {
    c("HLA-A", "HLA-B", "HLA-C", "HLA-DRB1")
  } else {
    c("HLA-A", "HLA-B", "HLA-C", "HLA-DRB1", "HLA-DQB1")
  }

  # When scope is "genotype" and direction is "bidirectional",
  # calculate GvH and HvG match summaries separately, then take the maximum.
  # Each recursive call is already vectorized via map_int() below.
  if (scope == "genotype" && direction == "bidirectional") {
    # Calculate GvH match summary across all loci
    gvh_result <- HLA_match_summary_HCT(
      GL_string_recip, GL_string_donor,
      direction = "GvH", match_grade = match_grade, scope = "locus"
    )
    # Calculate HvG match summary across all loci
    hvg_result <- HLA_match_summary_HCT(
      GL_string_recip, GL_string_donor,
      direction = "HvG", match_grade = match_grade, scope = "locus"
    )
    # Return the maximum of the two directional totals (vectorized via pmax)
    return(pmax(gvh_result, hvg_result))
  }

  # Calculate match numbers across loci (returns "HLA-A=2, HLA-B=1, ..." per pair)
  match_strings <- HLA_match_number(
    GL_string_recip, GL_string_donor, loci, direction = direction
  )

  # Extract the numeric match counts from each string and sum them
  map_int(str_extract_all(match_strings, "(?<=\\=)\\d+"), ~ sum(as.integer(.x)))
}

library(testthat)
library(dplyr)
library(stringr)

test_that("HLA_match_summary_HCT correctly calculates match summary for HCT", {
  GL_string_recip <- "HLA-A*29:02^HLA-C*06:02+HLA-C*07:01^HLA-B*08:01+HLA-B*13:02^HLA-DRB1*04:01+HLA-DRB1*07:01^HLA-DQB1*02:02+HLA-DQB1*03:02"
  GL_string_donor <- "HLA-A*02:01+HLA-A*29:02^HLA-C*06:01+HLA-C*07:02^HLA-B*08:01+HLA-B*13:03^HLA-DRB1*04:01+HLA-DRB1*07:01^HLA-DQB1*02:02+HLA-DQB1*03:02"

  result_Xof8_HvG <- HLA_match_summary_HCT(GL_string_recip, GL_string_donor, direction = "HvG", match_grade = "Xof8")
  result_Xof8_GvH <- HLA_match_summary_HCT(GL_string_recip, GL_string_donor, direction = "GvH", match_grade = "Xof8")
  result_Xof8_bidirectional <- HLA_match_summary_HCT(GL_string_recip, GL_string_donor, direction = "bidirectional", match_grade = "Xof8")

  expect_type(result_Xof8_HvG, "integer")
  expect_type(result_Xof8_GvH, "integer")
  expect_type(result_Xof8_bidirectional, "integer")

  expect_equal(result_Xof8_HvG, 4)
  expect_equal(result_Xof8_GvH, 5)
  expect_equal(result_Xof8_bidirectional, 4)

  result_Xof10_HvG <- HLA_match_summary_HCT(GL_string_recip, GL_string_donor, direction = "HvG", match_grade = "Xof10")
  result_Xof10_GvH <- HLA_match_summary_HCT(GL_string_recip, GL_string_donor, direction = "GvH", match_grade = "Xof10")
  result_Xof10_bidirectional <- HLA_match_summary_HCT(GL_string_recip, GL_string_donor, direction = "bidirectional", match_grade = "Xof10")

  expect_type(result_Xof10_HvG, "integer")
  expect_type(result_Xof10_GvH, "integer")
  expect_type(result_Xof10_bidirectional, "integer")

  expect_equal(result_Xof10_HvG, 6)
  expect_equal(result_Xof10_GvH, 7)
  expect_equal(result_Xof10_bidirectional, 6)

  perfect_match <- HLA_match_summary_HCT(GL_string_recip, GL_string_recip, direction = "bidirectional", match_grade = "Xof10")
  expect_equal(perfect_match, 10)

  GL_string_mismatch <- "HLA-A*24:01^HLA-C*05:01+HLA-C*08:01^HLA-B*51:01+HLA-B*52:01^HLA-DRB1*15:01+HLA-DRB1*16:01^HLA-DQB1*05:01+HLA-DQB1*06:01"
  mismatch_result <- HLA_match_summary_HCT(GL_string_recip, GL_string_mismatch, direction = "bidirectional", match_grade = "Xof10")
  expect_equal(mismatch_result, 0)

  expect_error(HLA_match_summary_HCT(GL_string_recip, GL_string_donor, direction = "invalid"))
})


# --- Input validation tests ---

test_that("HLA_match_summary_HCT rejects NULL inputs", {
  gl <- "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01^HLA-C*03:04+HLA-C*07:01^HLA-DRB1*04:01+HLA-DRB1*07:01"
  expect_error(HLA_match_summary_HCT(NULL, gl, match_grade = "Xof8"), "GL_string_recip")
  expect_error(HLA_match_summary_HCT(gl, NULL, match_grade = "Xof8"), "GL_string_donor")
})

test_that("HLA_match_summary_HCT rejects non-character GL strings", {
  gl <- "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01^HLA-C*03:04+HLA-C*07:01^HLA-DRB1*04:01+HLA-DRB1*07:01"
  expect_error(HLA_match_summary_HCT(123, gl, match_grade = "Xof8"), "must be a character")
})

test_that("HLA_match_summary_HCT rejects invalid match_grade", {
  gl <- "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01^HLA-C*03:04+HLA-C*07:01^HLA-DRB1*04:01+HLA-DRB1*07:01"
  expect_error(HLA_match_summary_HCT(gl, gl, match_grade = "Xof6"))
})

test_that("HLA_match_summary_HCT returns 8 for perfect Xof8 match", {
  gl <- "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01^HLA-C*03:04+HLA-C*07:01^HLA-DRB1*04:01+HLA-DRB1*07:01"
  result <- HLA_match_summary_HCT(gl, gl, direction = "bidirectional", match_grade = "Xof8")
  expect_equal(result, 8)
})

# --- Scope parameter tests ---

test_that("HLA_match_summary_HCT scope='locus' returns same results as default", {
  # Locus scope should match default behavior exactly.
  GL_string_recip <- "HLA-A*29:02^HLA-C*06:02+HLA-C*07:01^HLA-B*08:01+HLA-B*13:02^HLA-DRB1*04:01+HLA-DRB1*07:01^HLA-DQB1*02:02+HLA-DQB1*03:02"
  GL_string_donor <- "HLA-A*02:01+HLA-A*29:02^HLA-C*06:01+HLA-C*07:02^HLA-B*08:01+HLA-B*13:03^HLA-DRB1*04:01+HLA-DRB1*07:01^HLA-DQB1*02:02+HLA-DQB1*03:02"

  # Explicitly setting scope = "locus" should match not setting it at all
  expect_equal(
    HLA_match_summary_HCT(GL_string_recip, GL_string_donor, direction = "bidirectional", match_grade = "Xof8", scope = "locus"),
    HLA_match_summary_HCT(GL_string_recip, GL_string_donor, direction = "bidirectional", match_grade = "Xof8")
  )
  expect_equal(
    HLA_match_summary_HCT(GL_string_recip, GL_string_donor, direction = "bidirectional", match_grade = "Xof10", scope = "locus"),
    HLA_match_summary_HCT(GL_string_recip, GL_string_donor, direction = "bidirectional", match_grade = "Xof10")
  )
})

test_that("HLA_match_summary_HCT scope='genotype' calculates max of GvH and HvG totals", {
  GL_string_recip <- "HLA-A*29:02^HLA-C*06:02+HLA-C*07:01^HLA-B*08:01+HLA-B*13:02^HLA-DRB1*04:01+HLA-DRB1*07:01^HLA-DQB1*02:02+HLA-DQB1*03:02"
  GL_string_donor <- "HLA-A*02:01+HLA-A*29:02^HLA-C*06:01+HLA-C*07:02^HLA-B*08:01+HLA-B*13:03^HLA-DRB1*04:01+HLA-DRB1*07:01^HLA-DQB1*02:02+HLA-DQB1*03:02"

  # Genotype scope with bidirectional should return max of GvH and HvG totals
  # GvH Xof8 = 5, HvG Xof8 = 4, so genotype bidirectional Xof8 = max(5, 4) = 5
  result_Xof8_genotype <- HLA_match_summary_HCT(GL_string_recip, GL_string_donor,
    direction = "bidirectional", match_grade = "Xof8", scope = "genotype"
  )
  expect_type(result_Xof8_genotype, "integer")
  expect_equal(result_Xof8_genotype, 5)

  # GvH Xof10 = 7, HvG Xof10 = 6, so genotype bidirectional Xof10 = max(7, 6) = 7
  result_Xof10_genotype <- HLA_match_summary_HCT(GL_string_recip, GL_string_donor,
    direction = "bidirectional", match_grade = "Xof10", scope = "genotype"
  )
  expect_type(result_Xof10_genotype, "integer")
  expect_equal(result_Xof10_genotype, 7)
})

test_that("HLA_match_summary_HCT scope='genotype' is >= scope='locus' for bidirectional", {
  # Genotype scope should always be >= locus scope for bidirectional matching,
  # since locus takes the per-locus minimum before summing.
  GL_string_recip <- "HLA-A*29:02^HLA-C*06:02+HLA-C*07:01^HLA-B*08:01+HLA-B*13:02^HLA-DRB1*04:01+HLA-DRB1*07:01^HLA-DQB1*02:02+HLA-DQB1*03:02"
  GL_string_donor <- "HLA-A*02:01+HLA-A*29:02^HLA-C*06:01+HLA-C*07:02^HLA-B*08:01+HLA-B*13:03^HLA-DRB1*04:01+HLA-DRB1*07:01^HLA-DQB1*02:02+HLA-DQB1*03:02"

  locus_result <- HLA_match_summary_HCT(GL_string_recip, GL_string_donor,
    direction = "bidirectional", match_grade = "Xof8", scope = "locus"
  )
  genotype_result <- HLA_match_summary_HCT(GL_string_recip, GL_string_donor,
    direction = "bidirectional", match_grade = "Xof8", scope = "genotype"
  )
  expect_true(genotype_result >= locus_result)
})

test_that("HLA_match_summary_HCT scope='genotype' with non-bidirectional direction is ignored", {
  # When direction is not "bidirectional", scope should not affect the result.
  GL_string_recip <- "HLA-A*29:02^HLA-C*06:02+HLA-C*07:01^HLA-B*08:01+HLA-B*13:02^HLA-DRB1*04:01+HLA-DRB1*07:01"
  GL_string_donor <- "HLA-A*02:01+HLA-A*29:02^HLA-C*06:02+HLA-C*07:01^HLA-B*08:01^HLA-DRB1*04:01+HLA-DRB1*07:01"

  # GvH results should be identical regardless of scope
  expect_equal(
    HLA_match_summary_HCT(GL_string_recip, GL_string_donor, direction = "GvH", match_grade = "Xof8", scope = "genotype"),
    HLA_match_summary_HCT(GL_string_recip, GL_string_donor, direction = "GvH", match_grade = "Xof8", scope = "locus")
  )
  # HvG results should be identical regardless of scope
  expect_equal(
    HLA_match_summary_HCT(GL_string_recip, GL_string_donor, direction = "HvG", match_grade = "Xof8", scope = "genotype"),
    HLA_match_summary_HCT(GL_string_recip, GL_string_donor, direction = "HvG", match_grade = "Xof8", scope = "locus")
  )
})

test_that("HLA_match_summary_HCT scope='genotype' returns 7 for prompt example data", {
  # This test uses the exact data from the development prompt.
  GL_string_recip <- "HLA-A*29:02^HLA-C*06:02+HLA-C*07:01^HLA-B*08:01+HLA-B*13:02^HLA-DRB1*04:01+HLA-DRB1*07:01"
  GL_string_donor <- "HLA-A*02:01+HLA-A*29:02^HLA-C*06:02+HLA-C*07:01^HLA-B*08:01^HLA-DRB1*04:01+HLA-DRB1*07:01"

  result <- HLA_match_summary_HCT(GL_string_recip, GL_string_donor,
    direction = "bidirectional", match_grade = "Xof8", scope = "genotype"
  )
  expect_equal(result, 7)
})

test_that("HLA_match_summary_HCT scope='genotype' returns perfect score for identical genotypes", {
  gl <- "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01^HLA-C*03:04+HLA-C*07:01^HLA-DRB1*04:01+HLA-DRB1*07:01"
  # Perfect match: GvH = 8, HvG = 8, max(8, 8) = 8
  result <- HLA_match_summary_HCT(gl, gl, direction = "bidirectional", match_grade = "Xof8", scope = "genotype")
  expect_equal(result, 8)
})

test_that("HLA_match_summary_HCT scope='genotype' works with vectorized inputs", {
  recip <- c(
    "HLA-A*29:02^HLA-C*06:02+HLA-C*07:01^HLA-B*08:01+HLA-B*13:02^HLA-DRB1*04:01+HLA-DRB1*07:01",
    "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01^HLA-C*03:04+HLA-C*07:01^HLA-DRB1*04:01+HLA-DRB1*07:01"
  )
  donor <- c(
    "HLA-A*02:01+HLA-A*29:02^HLA-C*06:02+HLA-C*07:01^HLA-B*08:01^HLA-DRB1*04:01+HLA-DRB1*07:01",
    "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01^HLA-C*03:04+HLA-C*07:01^HLA-DRB1*04:01+HLA-DRB1*07:01"
  )
  result <- HLA_match_summary_HCT(recip, donor,
    direction = "bidirectional", match_grade = "Xof8", scope = "genotype"
  )
  expect_length(result, 2)
  # First pair should return 7 (from prompt example)
  expect_equal(result[1], 7)
  # Second pair is a perfect match
  expect_equal(result[2], 8)
})

test_that("HLA_match_summary_HCT rejects invalid scope", {
  gl <- "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01^HLA-C*03:04+HLA-C*07:01^HLA-DRB1*04:01+HLA-DRB1*07:01"
  expect_error(HLA_match_summary_HCT(gl, gl, match_grade = "Xof8", scope = "invalid"))
})


test_that("HLA_match_summary_HCT works with vectorized inputs", {
  recip <- c(
    "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01^HLA-C*03:04+HLA-C*07:01^HLA-DRB1*04:01+HLA-DRB1*07:01",
    "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01^HLA-C*03:04+HLA-C*07:01^HLA-DRB1*04:01+HLA-DRB1*07:01"
  )
  donor <- c(
    "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01^HLA-C*03:04+HLA-C*07:01^HLA-DRB1*04:01+HLA-DRB1*07:01",
    "HLA-A*03:01+HLA-A*24:02^HLA-B*15:01+HLA-B*40:01^HLA-C*01:02+HLA-C*02:02^HLA-DRB1*11:01+HLA-DRB1*13:01"
  )
  result <- HLA_match_summary_HCT(recip, donor, direction = "bidirectional", match_grade = "Xof8")
  expect_length(result, 2)
  # First pair is a perfect match
  expect_equal(result[1], 8)
  # Second pair is a complete mismatch
  expect_equal(result[2], 0)
})

test_that("HLA_match_summary_HCT works with vectorized inputs for GvH direction", {
  recip <- c(
    "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01^HLA-C*03:04+HLA-C*07:01^HLA-DRB1*04:01+HLA-DRB1*07:01",
    "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01^HLA-C*03:04+HLA-C*07:01^HLA-DRB1*04:01+HLA-DRB1*07:01"
  )
  donor <- c(
    "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01^HLA-C*03:04+HLA-C*07:01^HLA-DRB1*04:01+HLA-DRB1*07:01",
    "HLA-A*03:01+HLA-A*24:02^HLA-B*15:01+HLA-B*40:01^HLA-C*01:02+HLA-C*02:02^HLA-DRB1*11:01+HLA-DRB1*13:01"
  )
  result <- HLA_match_summary_HCT(recip, donor, direction = "GvH", match_grade = "Xof8")
  expect_length(result, 2)
  # First pair is a perfect match
  expect_equal(result[1], 8)
  # Second pair is a complete mismatch
  expect_equal(result[2], 0)
})

test_that("HLA_match_summary_HCT works with vectorized inputs for HvG direction", {
  recip <- c(
    "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01^HLA-C*03:04+HLA-C*07:01^HLA-DRB1*04:01+HLA-DRB1*07:01",
    "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01^HLA-C*03:04+HLA-C*07:01^HLA-DRB1*04:01+HLA-DRB1*07:01"
  )
  donor <- c(
    "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01^HLA-C*03:04+HLA-C*07:01^HLA-DRB1*04:01+HLA-DRB1*07:01",
    "HLA-A*03:01+HLA-A*24:02^HLA-B*15:01+HLA-B*40:01^HLA-C*01:02+HLA-C*02:02^HLA-DRB1*11:01+HLA-DRB1*13:01"
  )
  result <- HLA_match_summary_HCT(recip, donor, direction = "HvG", match_grade = "Xof8")
  expect_length(result, 2)
  # First pair is a perfect match
  expect_equal(result[1], 8)
  # Second pair is a complete mismatch
  expect_equal(result[2], 0)
})

test_that("HLA_match_summary_HCT works with vectorized inputs and explicit scope='locus'", {
  recip <- c(
    "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01^HLA-C*03:04+HLA-C*07:01^HLA-DRB1*04:01+HLA-DRB1*07:01",
    "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01^HLA-C*03:04+HLA-C*07:01^HLA-DRB1*04:01+HLA-DRB1*07:01"
  )
  donor <- c(
    "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01^HLA-C*03:04+HLA-C*07:01^HLA-DRB1*04:01+HLA-DRB1*07:01",
    "HLA-A*03:01+HLA-A*24:02^HLA-B*15:01+HLA-B*40:01^HLA-C*01:02+HLA-C*02:02^HLA-DRB1*11:01+HLA-DRB1*13:01"
  )
  # Explicit scope = "locus" should match the default behavior
  result_locus <- HLA_match_summary_HCT(recip, donor, direction = "bidirectional", match_grade = "Xof8", scope = "locus")
  result_default <- HLA_match_summary_HCT(recip, donor, direction = "bidirectional", match_grade = "Xof8")
  expect_equal(result_locus, result_default)
  expect_length(result_locus, 2)
  expect_equal(result_locus[1], 8)
  expect_equal(result_locus[2], 0)
})

#' @title HLA_prefix_add
#'
#' @description This function adds a specified prefix to the beginning of
#' each HLA type, and works on a single allele or all alleles in a GL string.
#' Useful for adding HLA or gene prefixes.
#'
#' @param data A string with a single HLA allele, a GL string of HLA alleles,
#' or a character vector containing either of the previous.
#' @param prefix A character string to be added as a prefix to the alleles.
#' Default is "HLA-".
#'
#' @return A vector with the specified prefix added to the values.
#'
#' @examples
#' # The HLA_typing_LIS dataset contains a table as might be found in a clinical
#' # laboratory information system:
#' print(HLA_typing_LIS)
#'
#' # The `HLA_prefix_add` function can be used to add the correct HLA prefixes to the table:
#' library(dplyr)
#' HLA_typing_LIS %>% mutate(
#'   across(
#'     mA1Cd.recipient:mA2Cd.recipient,
#'     ~ HLA_prefix_add(., "HLA-A*")
#'   )
#' )
#'
#' @export
#'
#' @importFrom stringr str_replace
#' @importFrom dplyr mutate
#' @importFrom dplyr %>%

HLA_prefix_add <- function(data, prefix = "HLA-") {
  # Validate input
  check_gl_string(data, "data")

  # Expand the allele of GL string
  data %>%
    GLstring_expand_longer() %>%
    # Add string to beginning of typing
    mutate(value = str_replace(value, "^", prefix)) %>%
    # Collapse the GL string
    ambiguity_table_to_GLstring()
}

library(testthat)
library(stringr)

test_that("HLA_prefix_add correctly adds prefixes", {
  result_single <- HLA_prefix_add("01:01", "HLA-A*")
  expect_equal(result_single, "HLA-A*01:01")

  result_multiple <- HLA_prefix_add(c("02:01", "03:02", "11:01"), "HLA-B*")
  expect_equal(result_multiple, c("HLA-B*02:01", "HLA-B*03:02", "HLA-B*11:01"))

  result_numeric <- HLA_prefix_add("2", "HLA-A")
  expect_equal(result_numeric, "HLA-A2")

  result_empty <- HLA_prefix_add("", "HLA-C*")
  expect_equal(result_empty, "HLA-C*")

  result_na <- HLA_prefix_add(NA, "HLA-")
  expect_true(is.na(result_na))

  result_default <- HLA_prefix_add("DRB1*04:01")
  expect_equal(result_default, "HLA-DRB1*04:01")

  result_no_prefix <- HLA_prefix_add("DQB1*06:02", "")
  expect_equal(result_no_prefix, "DQB1*06:02")

  result_special_chars <- HLA_prefix_add("**01:01", "HLA-")
  expect_equal(result_special_chars, "HLA-**01:01")

  result_gl_string <- HLA_prefix_add("A*02:01/A*02:02+A*68:01^B*57:01+B*07:02", "HLA-")
  expect_equal(result_gl_string, "HLA-A*02:01/HLA-A*02:02+HLA-A*68:01^HLA-B*57:01+HLA-B*07:02")

  result_gl_string_vector <- HLA_prefix_add(c("A*02:01/A*02:02+A*68:01^B*57:01+B*07:02", "A*02:01", "B*07:01/B*07:02+B*15:01"), "HLA-")
  expect_equal(result_gl_string_vector, c("HLA-A*02:01/HLA-A*02:02+HLA-A*68:01^HLA-B*57:01+HLA-B*07:02", "HLA-A*02:01", "HLA-B*07:01/HLA-B*07:02+HLA-B*15:01"))
})

#' @title GLstring_genotype_ambiguity
#'
#' @description This function processes GL strings in the specified columns of
#' a data frame to retain only the first genotype ambiguity, optionally
#' retaining the remaining ambiguities in a separate column with "_ambiguity"
#' appended. The function ensures that genes have been separated from the GL
#' strings prior to execution; otherwise, an error will be thrown if a "^" is
#' detected in the GL strings.
#'
#' @param data A data frame
#' @param columns The names of the columns in the data frame that contain GL strings
#' @param keep_ambiguities A logical value indicating whether to retain the
#' remaining ambiguities in separate columns with "_genotype_ambiguity" appended
#' to the original column names. Default is FALSE.
#'
#' @return A data frame with the first genotype ambiguity retained in the
#' original columns. If \code{keep_ambiguities} is TRUE, the remaining
#' ambiguities are placed in separate columns.
#'
#' @examples
#' HLA_type <- data.frame(
#'   sample = c("sample1", "sample2"),
#'   HLA_A = c("A*01:01+A*68:01|A*01:02+A*68:55|A*01:99+A*68:66", "A*02:01+A*03:01|A*02:02+A*03:03"),
#'   HLA_B = c("B*07:02+B*58:01|B*07:03+B*58:09", "B*08:01+B*15:01|B*08:02+B*15:17"),
#'   stringsAsFactors = FALSE
#' )
#'
#' GLstring_genotype_ambiguity(HLA_type, columns = c("HLA_A", "HLA_B"), keep_ambiguities = TRUE)
#'
#' @export
#'
#' @importFrom dplyr select
#' @importFrom dplyr mutate
#' @importFrom dplyr across
#' @importFrom dplyr summarize
#' @importFrom dplyr %>%
#' @importFrom dplyr contains
#' @importFrom dplyr na_if
#' @importFrom stringr str_detect
#' @importFrom stringr str_extract
#' @importFrom stringr str_replace_all
#' @importFrom cli cli_abort
#' @importFrom stringr str_replace


GLstring_genotype_ambiguity <- function(data, columns, keep_ambiguities = FALSE) {
  # Validate inputs
  check_data_frame(data, "data")
  check_logical_flag(keep_ambiguities, "keep_ambiguities")

  # Identify the columns to modify
  cols2mod <- names(select(data, {{ columns }}))

  # Check for "^" in any of the selected columns, which indicates genes
  # haven't been separated from the GL string yet.
  if (any(str_detect(unlist(data[cols2mod]), "\\^"), na.rm = TRUE)) {
    cli_abort("Genes must be separated before {.fn GLstring_genotype_ambiguity} can be used. Process GL strings with {.fn GLstring_genes} first.")
  }

  # Copy GL string to a new ambiguity column.
  data %>%
    mutate(across({{ cols2mod }},
      ~ as.character(.),
      .names = "{col}_genotype_ambiguity"
    )) %>%
    # Keep the first genotype ambiguity in the original columns.
    mutate(across({{ cols2mod }}, ~ str_extract(., "[^|]+"))) %>%
    # Remove the first genotype and its trailing pipe from the ambiguity columns,
    # then convert empty strings to NA.
    mutate(across(ends_with("_genotype_ambiguity"), ~ na_if(str_replace(., "^[^|]+\\|?", ""), ""))) %>%
    # Drop the ambiguity columns if not wanted
    {
      if (keep_ambiguities) . else select(., -contains("ambiguity"))
    }
}

globalVariables(c("X", "Y", "ends_with"))

#' Ten HLA-A~C~B~DRB3/4/5~DRB1~DQA1~DQB1~DPA1~DPB1 haplotypes and their
#' frequencies in the global population.
#'
#' @docType data
#'
#' @usage data(Haplotype_frequencies)
#'
#' @format
#' A tibble:
#' \describe{
#'   \item{HLA-A:HLA-DPB1}{HLA allele names}
#'   \item{Global_Fq}{haplotype frequencies}
#'   \item{Global_n}{haplotype observations}
#'   \item{Global_rank}{haplotype rank}
#'     ...
#' }
#'
#' @keywords datasets
#'
#' @source <https://doi.org/10.1016/j.humimm.2021.04.007>
#'
"Haplotype_frequencies"

#' @title HLA_prefix_remove
#'
#' @description This function removes HLA and optionally locus prefixes from a string of HLA typing:
#' "HLA-A2" changes to "A2" or "2". By default, HLA and locus prefixes are removed. This function
#' also works on each allele in a GL string.
#'
#' @param data A string with a single HLA allele, a GL string of HLA alleles,
#' or a character vector containing either of the previous.
#' @param keep_locus A logical value indicating whether to retain any locus values.
#' The default value is FALSE.
#'
#' @return A vector modified to remove HLA and optionally locus prefixes.
#'
#' @examples
#' # The HLA_typing_1 dataset contains a table with HLA typing spread across multiple columns:
#' print(HLA_typing_1)
#'
#' # The `HLA_prefix_remove` function can be used to get each column to have only the
#' # colon-separated fields:
#' library(dplyr)
#' HLA_typing_1 %>% mutate(
#'   across(
#'     A1:DPB1_2,
#'     ~ HLA_prefix_remove(.)
#'   )
#' )
#'
#' @export
#'
#' @importFrom dplyr %>%
#' @importFrom dplyr mutate
#' @importFrom stringr str_replace

HLA_prefix_remove <- function(data, keep_locus = FALSE) {
  # Validate inputs
  check_gl_string(data, "data")
  check_logical_flag(keep_locus, "keep_locus")

  # Removes any HLA and locus prefixes from typing results.
  result <- data %>%
    GLstring_expand_longer() %>%
    # Remove "HLA-" prefix from each allele.
    mutate(value = str_replace(value, "HLA-", ""))

  if (!keep_locus) {
    # Remove locus prefix in two passes:
    # Pass 1: Remove alphabetic prefix anchored at the start (e.g. "DRB" from "DRB1*04:01", "A" from "A2").
    # Pass 2: Remove any remaining digits followed by an asterisk (e.g. "1*" from "1*04:01").
    # These cannot be combined because pass 2 must be unanchored to catch the asterisk
    # after pass 1 has already stripped the alpha characters.
    result <- result %>%
      mutate(value = str_replace(value, "^[:alpha:]+", "")) %>%
      mutate(value = str_replace(value, "[:digit:]*\\*", ""))
  }

  # Reassemble the GL string.
  ambiguity_table_to_GLstring(result)
}

library(testthat)
library(stringr)

test_that("HLA_prefix_remove correctly removes prefixes", {
  result_single <- HLA_prefix_remove("HLA-A2")
  expect_equal(result_single, "2")

  result_multiple <- HLA_prefix_remove(c("HLA-A*02:01", "HLA-B*07:02", "HLA-Cw6"))
  expect_equal(result_multiple, c("02:01", "07:02", "6"))

  result_no_prefix <- HLA_prefix_remove("A*02:01")
  expect_equal(result_no_prefix, "02:01")

  result_locus_only <- HLA_prefix_remove("HLA-DRB1*15:01")
  expect_equal(result_locus_only, "15:01")

  result_mixed <- HLA_prefix_remove(c("A66", "B27", "HLA-Cw4", "HLA-DR52"))
  expect_equal(result_mixed, c("66", "27", "4", "52"))

  result_empty <- HLA_prefix_remove("")
  expect_equal(result_empty, "")

  result_na <- HLA_prefix_remove(NA)
  expect_true(is.na(result_na))

  result_special_chars <- HLA_prefix_remove("HLA-**01:01")
  expect_equal(result_special_chars, "*01:01")

  result_GL_string <- HLA_prefix_remove(c("HLA-A*02:01", "HLA-A2", "HLA-A*02:01/HLA-A*02:02+HLA-A*68:01^HLA-B*57:01+HLA-B*07:02"), keep_locus = TRUE)
  expect_equal(result_GL_string, c("A*02:01", "A2", "A*02:01/A*02:02+A*68:01^B*57:01+B*07:02"))

  result_GL_string_2 <- HLA_prefix_remove(c("HLA-A*02:01", "HLA-A2", "HLA-A*02:01/HLA-A*02:02+HLA-A*68:01^HLA-B*57:01+HLA-B*07:02"), keep_locus = FALSE)
  expect_equal(result_GL_string_2, c("02:01", "2", "02:01/02:02+68:01^57:01+07:02"))
})

#' @title HLA_columns_to_GLstring
#'
#' @description A function to take HLA typing data spread across different columns,
#' as is often found in wild-caught data, and transform it to a GL string. If column names
#' have anything besides the locus name and a number (e.g. "mA1Cd" instead of just "A1"),
#' the function will have trouble determining the locus from the column name. The `prefix_to_remove`
#' and `suffix_to_remove` arguments can be used to clean up the column names. See the example for
#' how these arguments are used.
#'
#' @param data A data frame with each row including an HLA typing result, with
#' individual columns containing a single allele.
#' @param HLA_typing_columns A list of columns containing the HLA alleles. Tidyselect is supported.
#' @param prefix_to_remove An optional string of characters to remove from the
#' locus names. The goal is to get the column names to the locus and a number. For example,
#' columns named "mDRB11Cd" and "mDRB12Cd" should use the `prefix_to_remove` value of "m".
#' @param suffix_to_remove An optional string of characters to remove from the
#' locus names. Using the example above, the `suffix_to_remove` value will be "Cd".
#'
#' @return A list of GL strings in the order of the original data frame.
#'
#' @examples
#' # The HLA_typing_LIS dataset contains a table as might be found in a clinical laboratory
#' # information system:
#' print(HLA_typing_LIS)
#'
#' # The `HLA_columns_to_GLString` function can be used to coerce typing spread across
#' # multiple columns into a GL string:
#' library(dplyr)
#' HLA_typing_LIS %>%
#'   mutate(
#'     GL_string = HLA_columns_to_GLstring(
#'       ., # Note that if this function is used inside a `mutate` call "." will have to be
#'       # used as the first argument to extract data from the working data frame.
#'       HLA_typing_columns = mA1Cd.recipient:mDPB12cd.recipient,
#'       prefix_to_remove = "m",
#'       suffix_to_remove = "Cd.recipient"
#'     ),
#'     .after = patient
#'   ) %>%
#'   select(patient, GL_string)
#'
#' # Using the base pipe:
#' HLA_typing_LIS |>
#'   mutate(
#'     GL_string = HLA_columns_to_GLstring(
#'       HLA_typing_LIS, # If using the base pipe, the first argument will have to be
#'       # the working data frame.
#'       HLA_typing_columns = mA1Cd.recipient:mDPB12cd.recipient,
#'       prefix_to_remove = "m",
#'       suffix_to_remove = "Cd.recipient"
#'     ),
#'     .after = patient
#'   ) |>
#'   select(patient, GL_string)
#'
#' @export
#'
#' @importFrom dplyr mutate
#' @importFrom dplyr %>%
#' @importFrom dplyr select
#' @importFrom dplyr case_when
#' @importFrom dplyr if_else
#' @importFrom dplyr coalesce
#' @importFrom dplyr filter
#' @importFrom dplyr distinct
#' @importFrom dplyr pull
#' @importFrom dplyr summarise
#' @importFrom glue glue
#' @importFrom stringr regex
#' @importFrom stringr str_c
#' @importFrom stringr str_escape
#' @importFrom stringr str_extract
#' @importFrom stringr str_replace
#' @importFrom stringr str_detect
#' @importFrom stringr str_glue
#' @importFrom stringr str_flatten
#' @importFrom tidyr pivot_longer
#' @importFrom tidyselect all_of
#' @importFrom cli cli_abort

HLA_columns_to_GLstring <- function(data, HLA_typing_columns, prefix_to_remove = "", suffix_to_remove = "") {
  # Validate inputs
  check_data_frame(data, "data")

  # Set up prefix and suffix regex to remove unwanted parts
  prefix_regex <- regex(str_c("^", str_escape(prefix_to_remove)), ignore_case = TRUE)
  suffix_regex <- regex(str_c(str_escape(suffix_to_remove), "$"), ignore_case = TRUE)

  # Identify the columns to modify
  col2mod <- names(select(data, {{ HLA_typing_columns }}))

  # Named vector mapping molecular locus names to their serologic equivalents.
  # Used as a fast lookup instead of a case_when with str_detect calls.
  serologic_map <- c(
    "HLA-A" = "HLA-A",
    "HLA-B" = "HLA-B",
    "HLA-Bw" = "HLA-Bw",
    "HLA-C" = "HLA-Cw",
    "HLA-Cw" = "HLA-Cw",
    "HLA-DRB1" = "HLA-DR",
    "HLA-DRB345" = "HLA-DR",
    "HLA-DRB3" = "HLA-DR",
    "HLA-DRB4" = "HLA-DR",
    "HLA-DRB5" = "HLA-DR",
    "HLA-DR" = "HLA-DR",
    "HLA-DQB1" = "HLA-DQ",
    "HLA-DQA1" = "HLA-DQA",
    "HLA-DQ" = "HLA-DQ",
    "HLA-DPA1" = "HLA-DPA",
    "HLA-DPB1" = "HLA-DP",
    "unknown" = "unknown"
  )

  # Processing Step 1
  step1 <- data %>%
    mutate(row_for_function = row_number()) %>%
    # Pivoting longer to get each allele on a separate row.
    pivot_longer(cols = all_of(col2mod), names_to = "names", values_to = "allele") %>%
    # Determine if typing is molecular by presence of ":" or leading zero in allele,
    # or the loci DQA1, DPB1, and DPA1, which are always in molecular format.
    mutate(molecular = str_detect(allele, ":|^0") | str_detect(names, "DQA1|DPB1|DPA1")) %>%
    # Remove prefixes and suffixes from column names.
    mutate(
      truncated_names = str_replace(names, prefix_regex, ""),
      truncated_names = str_replace(truncated_names, suffix_regex, ""),
      truncated_names = HLA_prefix_remove(truncated_names, keep_locus = TRUE)
    ) %>%
    # Use the HLA_validate function to clean up the typing.
    mutate(allele = HLA_validate(allele)) %>%
    # Determine if allele itself specifies a DRB3/4/5 locus (for generic DRB345 columns).
    mutate(
      DRB_locus_raw = case_when(
        # When the allele starts with "DRB3*", "DRB4*" or "DRB5*"
        str_detect(allele, regex("DRB[345]\\*", ignore_case = TRUE)) ~ str_c("HLA-DRB", str_extract(allele, "(?<=DRB)[345]")),
        # When the allele does not have a "DRB" prefix, instead only has a number prefix
        str_detect(allele, regex("^[345]\\*", ignore_case = TRUE)) ~ str_c("HLA-DRB", str_extract(allele, "^[345]")),
        TRUE ~ NA_character_
      )
    ) %>%
    # Remove any prefixes in alleles.
    mutate(allele = HLA_prefix_remove(allele)) %>%
    # Lowercase column names once for case-insensitive prefix matching.
    mutate(lower_names = tolower(truncated_names)) %>%
    # Determine locus name from column name using prefix matching.
    # Order matters: more specific prefixes must come before less specific ones
    # (e.g. "bw" before "b", "drb345" before "drb3", "dqb1" before "dq").
    mutate(locus_from_name = case_when(
      startsWith(lower_names, "a") ~ "HLA-A",
      startsWith(lower_names, "bw") ~ "HLA-Bw",
      startsWith(lower_names, "b") ~ "HLA-B",
      startsWith(lower_names, "cw") ~ "HLA-Cw",
      startsWith(lower_names, "c") ~ "HLA-C",
      startsWith(lower_names, "drb345") ~ "HLA-DRB345",
      startsWith(lower_names, "drb1") ~ "HLA-DRB1",
      startsWith(lower_names, "drb3") ~ "HLA-DRB3",
      startsWith(lower_names, "drb4") ~ "HLA-DRB4",
      startsWith(lower_names, "drb5") ~ "HLA-DRB5",
      startsWith(lower_names, "dr") ~ "HLA-DR",
      startsWith(lower_names, "dqb1") ~ "HLA-DQB1",
      startsWith(lower_names, "dqa1") ~ "HLA-DQA1",
      startsWith(lower_names, "dq") ~ "HLA-DQ",
      startsWith(lower_names, "dpa1") ~ "HLA-DPA1",
      startsWith(lower_names, "dpb1") ~ "HLA-DPB1",
      TRUE ~ "unknown"
    )) %>%
    # Look up the serologic name from the molecular locus name.
    mutate(serologic_name = serologic_map[locus_from_name]) %>%
    # Determine if the DR typing is one of the DR51/52/53 loci.
    mutate(DRB345 = if_else(str_detect(allele, "^5") & locus_from_name == "HLA-DR", TRUE, FALSE)) %>%
    # Handle the final locus name for high-resolution DRB alleles.
    mutate(molecular_locus = coalesce(DRB_locus_raw, locus_from_name)) %>%
    # Record "XX" if there is no typing at any of the selected loci.
    mutate(allele = if (all(is.na(allele))) "XX" else allele, .by = c(row_for_function, molecular_locus)) %>%
    # Remove any blank values
    filter(!is.na(allele)) %>%
    # Remove any "XX" values from the DRB3/4/5 loci
    filter(!(allele == "XX" & str_detect(molecular_locus, "DRB[345]")))

  # Set up error detection for any loci that could not be determined
  error_table <- step1 %>% filter(locus_from_name == "unknown")
  error_column_names <- error_table %>%
    select(names) %>%
    distinct() %>%
    dplyr::pull(names)

  if (nrow(error_table) != 0) {
    # Print the columns that caused the error to assist in debugging
    cli_abort("The column(s) {.val {error_column_names}} could not be parsed to determine HLA loci.")
  }

  # Assemble the final type
  step2 <- step1 %>%
    mutate(final_type = if_else(
      molecular,
      str_glue("{molecular_locus}*{allele}"),
      str_glue("{serologic_name}{allele}")
    )) %>%
    # Group alleles within the same locus
    summarise(final_type_2 = str_flatten(final_type, collapse = "+"), .by = c(row_for_function, molecular_locus, DRB345)) %>%
    # Assemble GL string with each locus separated by "^"
    summarise(GL_string = str_flatten(final_type_2, collapse = "^"), .by = row_for_function)

  # Return GL Strings
  return(step2 %>% dplyr::pull(GL_string))
}

globalVariables(c(
  ".", "truncated_names", "lower_names", "locus_from_name", "DRB_locus",
  "row_for_function", "molecular_locus", "molecular", "serologic_name",
  "final_type", "DRB345", "final_type_2", "GL_string", "DRB_locus_raw"
))

#' @title HLA_match_number
#'
#' @description Calculates the number of HLA matches as two minus the number of
#' mismatches from `HLA_mismatch_number`. Homozygous mismatches are counted twice.
#' Supports match calculations for host-vs-graft (HvG), graft-vs-host (GvH),
#' or bidirectional. Bidirectional matching is the default, but can be overridden
#' using the "direction" argument.
#'
#' @param GL_string_recip A GL string representing the recipient's HLA genotype.
#' @param GL_string_donor A GL string representing the donor's HLA genotype.
#' @param loci A character vector specifying the loci to be considered for
#' mismatch calculation. HLA-DRB3/4/5 (and their serologic equivalents DR51/52/53)
#' are considered once locus for this function, and should be called in this argument
#' as "HLA-DRB3/4/5" or "HLA-DR51/52/53", respectively.
#' @param direction A character string indicating the direction of match.
#' Options are "HvG" (host vs. graft), "GvH" (graft vs. host), "bidirectional"
#' (the minimum value of "HvG" and "GvH").
#'
#' @return An integer value or a character string:
#' - If `loci` includes only one locus, the function returns an integer
#' mismatch count for that locus.
#' - If `loci` includes multiple loci, the function returns a character
#' string in the format "Locus1=Count1, Locus2=Count2, ...".
#'
#' @examples
#'
#' file <- HLA_typing_1[, -1]
#' GL_string <- HLA_columns_to_GLstring(file, HLA_typing_columns = everything())
#' GL_string_recip <- GL_string[1]
#' GL_string_donor <- GL_string[2]
#'
#' loci <- c("HLA-A", "HLA-B")
#'
#' # Calculate mismatch numbers (Host vs. Graft)
#' HLA_match_number(GL_string_recip, GL_string_donor, loci, direction = "HvG")
#'
#' # Calculate mismatch numbers (Graft vs. Host)
#' HLA_match_number(GL_string_recip, GL_string_donor, loci, direction = "GvH")
#'
#' # Calculate mismatch numbers (Bidirectional)
#' HLA_match_number(GL_string_recip, GL_string_donor,
#'   loci,
#'   direction = "bidirectional"
#' )
#'
#' @export
#'

HLA_match_number <- function(GL_string_recip, GL_string_donor, loci, direction = "bidirectional") {
  # Validate inputs
  check_gl_string(GL_string_recip, "GL_string_recip")
  check_gl_string(GL_string_donor, "GL_string_donor")
  check_loci(loci)

  direction <- match.arg(direction, c("HvG", "GvH", "bidirectional"))

  # Single locus: return integer match count (2 - mismatch count)
  if (length(loci) == 1) {
    # Calculate matches as 2 minus the mismatch count for the given direction.
    match_table <- tibble(mismatch = HLA_mismatch_number(
      GL_string_recip, GL_string_donor, loci, direction
    )) %>%
      mutate(match = 2 - mismatch)
    return(match_table$match)
  } else {
  # Multiple loci: return "Locus1=Count1, Locus2=Count2, ..." string
  match_table <- tibble(mismatch = HLA_mismatch_number(
    GL_string_recip, GL_string_donor, loci, direction
  )) %>%
    # Add a row number to combine data at the end.
    mutate(case = row_number()) %>%
    # Separate the loci.
    separate_longer_delim(mismatch, delim = ", ") %>%
    separate_wider_delim(mismatch, delim = "=", names = c("locus", "mismatches")) %>%
    # Recode mismatches as integers
    mutate(mismatches = as.integer(mismatches)) %>%
    # Calculate matches as 2 - mismatch.
    mutate(matches = 2 - mismatches) %>%
    # Clean up table.
    select(-mismatches) %>%
    unite(locus, matches, col = "Matches", sep = "=") %>%
    summarise(Matches = str_flatten(Matches, collapse = ", "), .by = case)

  return(match_table$Matches)
  }
}

globalVariables(c("mismatch", "matches", "Matches"))

library(testthat)
library(dplyr)
library(stringr)

test_that("HLA_match_number correctly calculates HLA match numbers across different scenarios", {
  GL_string_recip <- "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01^HLA-DRB1*04:01+HLA-DRB1*07:01"
  GL_string_donor <- "HLA-A*01:01+HLA-A*03:01^HLA-B*07:02+HLA-B*44:02^HLA-DRB1*04:01+HLA-DRB1*13:01"
  loci <- c("HLA-A", "HLA-B", "HLA-DRB1")

  result_HvG <- HLA_match_number(GL_string_recip, GL_string_donor, loci, direction = "HvG")
  result_GvH <- HLA_match_number(GL_string_recip, GL_string_donor, loci, direction = "GvH")
  result_bidirectional <- HLA_match_number(GL_string_recip, GL_string_donor, loci, direction = "bidirectional")

  expect_match(result_HvG, "HLA-A=1, HLA-B=1, HLA-DRB1=1")
  expect_match(result_GvH, "HLA-A=1, HLA-B=1, HLA-DRB1=1")
  expect_match(result_bidirectional, "HLA-A=1, HLA-B=1, HLA-DRB1=1")

  perfect_match <- HLA_match_number(GL_string_recip, GL_string_recip, loci, direction = "bidirectional")
  expect_match(perfect_match, "HLA-A=2, HLA-B=2, HLA-DRB1=2")

  GL_string_mismatch <- "HLA-A*24:01+HLA-A*25:01^HLA-B*51:01+HLA-B*52:01^HLA-DRB1*15:01+HLA-DRB1*16:01"
  mismatch_result <- HLA_match_number(GL_string_recip, GL_string_mismatch, loci, direction = "bidirectional")
  expect_match(mismatch_result, "HLA-A=0, HLA-B=0, HLA-DRB1=0")

  result_A <- HLA_match_number(GL_string_recip, GL_string_donor, "HLA-A", direction = "HvG")
  result_B <- HLA_match_number(GL_string_recip, GL_string_donor, "HLA-B", direction = "HvG")
  result_DRB1 <- HLA_match_number(GL_string_recip, GL_string_donor, "HLA-DRB1", direction = "HvG")

  expect_equal(result_A, 1)
  expect_equal(result_B, 1)
  expect_equal(result_DRB1, 1)

  expect_error(HLA_match_number(GL_string_recip, GL_string_donor, loci, direction = "invalid"))
})


# --- Input validation tests ---

test_that("HLA_match_number rejects NULL inputs", {
  gl <- "HLA-A*01:01+HLA-A*02:01"
  expect_error(HLA_match_number(NULL, gl, "HLA-A", "HvG"), "GL_string_recip")
  expect_error(HLA_match_number(gl, NULL, "HLA-A", "HvG"), "GL_string_donor")
  expect_error(HLA_match_number(gl, gl, NULL, "HvG"), "loci")
})

test_that("HLA_match_number rejects non-character GL strings", {
  gl <- "HLA-A*01:01+HLA-A*02:01"
  expect_error(HLA_match_number(123, gl, "HLA-A", "HvG"), "must be a character")
})

test_that("HLA_match_number returns 2 for perfect match at single locus", {
  gl <- "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01"
  result <- HLA_match_number(gl, gl, "HLA-A", "bidirectional")
  expect_equal(result, 2)
})

test_that("HLA_match_number returns 0 for complete mismatch at single locus", {
  recip <- "HLA-A*01:01+HLA-A*02:01"
  donor <- "HLA-A*03:01+HLA-A*24:02"
  result <- HLA_match_number(recip, donor, "HLA-A", "bidirectional")
  expect_equal(result, 0)
})

test_that("HLA_match_number agrees with mismatch_table_2016 consensus (homozygous_count = 2)", {
  # Note: HLA_match_number does not expose a homozygous_count parameter —
  # it always uses the default of 2 via HLA_mismatch_number. Therefore,
  # only the 2016 table (which expects homozygous_count = 2) is tested here.
  # The 2010 table (homozygous_count = 1) is tested in test-HLA_mismatch_number.R.
  # Define symbolic alleles to substitute into the table.
  A <- "HLA-A*01:01"
  B <- "HLA-A*02:05"
  C <- "HLA-A*24:02"
  D <- "HLA-A*31:03"
  N <- "HLA-A*68:11N"

  # Build GL strings from the 2016 table's symbolic allele pairs.
  mismatch_table_2016_match <- mismatch_table_2016 %>%
    # Assign the example alleles based on symbolic codes.
    mutate(recipient_1 = case_when(
      str_detect(Patient, "^A") ~ A,
      str_detect(Patient, "^N") ~ N
    ), recipient_2 = case_when(
      str_detect(Patient, "A$") ~ A,
      str_detect(Patient, "B$") ~ B,
      str_detect(Patient, "N$") ~ N
    ), donor_1 = case_when(
      str_detect(Donor, "^A") ~ A,
      str_detect(Donor, "^B") ~ B,
      str_detect(Donor, "^C") ~ C,
      str_detect(Donor, "^N") ~ N
    ), donor_2 = case_when(
      str_detect(Donor, "A$") ~ A,
      str_detect(Donor, "B$") ~ B,
      str_detect(Donor, "C$") ~ C,
      str_detect(Donor, "D$") ~ D,
      str_detect(Donor, "N$") ~ N
    )) %>%
    # Turn the two alleles into a GL string.
    mutate(
      GL_string_recip = str_c(recipient_1, recipient_2, sep = "+"),
      GL_string_donor = str_c(donor_1, donor_2, sep = "+")
    ) %>%
    select(Patient, Donor, GL_string_recip, GL_string_donor, "#GvH", "#HvG", "#Max") %>%
    rename(GvH = "#GvH", HvG = "#HvG", Max = "#Max") %>%
    # Calculate match numbers for each direction (match = 2 - mismatch).
    mutate(
      GvH_match = HLA_match_number(GL_string_recip, GL_string_donor, "HLA-A", "GvH"),
      HvG_match = HLA_match_number(GL_string_recip, GL_string_donor, "HLA-A", "HvG"),
      bidir_match = HLA_match_number(GL_string_recip, GL_string_donor, "HLA-A", "bidirectional")
    ) %>%
    # Derive expected match numbers from the consensus mismatch counts.
    mutate(
      GvH_expected = 2L - GvH,
      HvG_expected = 2L - HvG,
      bidir_expected = 2L - Max
    ) %>%
    # Check that calculated results match expected results.
    mutate(
      GvH_result = (GvH_match == GvH_expected),
      HvG_result = (HvG_match == HvG_expected),
      bidir_result = (bidir_match == bidir_expected)
    ) %>%
    # Check that for each row all values are correct.
    mutate(total_result = if_all(GvH_result:bidir_result)) %>%
    # Summarize to a single value if all were true.
    distinct(total_result) %>%
    pull(total_result)

  expect_equal(mismatch_table_2016_match, TRUE)
})

test_that("HLA_match_number works with vectorized inputs", {
  recip <- c(
    "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01",
    "HLA-A*03:01+HLA-A*24:02^HLA-B*15:01+HLA-B*40:01"
  )
  donor <- c(
    "HLA-A*01:01+HLA-A*03:01^HLA-B*07:02+HLA-B*08:01",
    "HLA-A*03:01+HLA-A*24:02^HLA-B*15:01+HLA-B*40:01"
  )
  # Single locus: should return integer vector of length 2
  result <- HLA_match_number(recip, donor, "HLA-A", "HvG")
  expect_length(result, 2)
  expect_equal(result[1], 1)
  expect_equal(result[2], 2)

  # Multiple loci: should return character vector of length 2
  result_multi <- HLA_match_number(recip, donor, c("HLA-A", "HLA-B"), "bidirectional")
  expect_length(result_multi, 2)
  expect_match(result_multi[1], "HLA-A=1, HLA-B=2")
  expect_match(result_multi[2], "HLA-A=2, HLA-B=2")
})

library(testthat)
library(dplyr)
library(stringr)

test_that("HLA_mismatch_alleles correctly identifies mismatched alleles", {
  GL_string_recip <- "HLA-A*01:01+HLA-A*02:01^HLA-B*07:02+HLA-B*08:01^HLA-DRB1*13:01+HLA-DRB1*15:01"
  GL_string_donor <- "HLA-A*01:01+HLA-A*03:01^HLA-B*07:02+HLA-B*44:02^HLA-DRB1*13:01+HLA-DRB1*14:01"
  loci <- c("HLA-A", "HLA-B", "HLA-DRB1")

  result_HvG <- HLA_mismatch_alleles(GL_string_recip, GL_string_donor, loci, direction = "HvG")
  result_GvH <- HLA_mismatch_alleles(GL_string_recip, GL_string_donor, loci, direction = "GvH")
  result_bidirectional <- HLA_mismatch_alleles(GL_string_recip, GL_string_donor, loci, direction = "bidirectional")
  result_SOT <- HLA_mismatch_alleles(GL_string_recip, GL_string_donor, loci, direction = "SOT")

  expect_equal(result_HvG, "HLA-A=HLA-A*03:01, HLA-B=HLA-B*44:02, HLA-DRB1=HLA-DRB1*14:01")
  expect_equal(result_GvH, "HLA-A=HLA-A*02:01, HLA-B=HLA-B*08:01, HLA-DRB1=HLA-DRB1*15:01")
  expect_equal(result_SOT, "HLA-A=HLA-A*03:01, HLA-B=HLA-B*44:02, HLA-DRB1=HLA-DRB1*14:01")
  expect_equal(result_bidirectional, "HvG;HLA-A=HLA-A*03:01, HLA-B=HLA-B*44:02, HLA-DRB1=HLA-DRB1*14:01<>GvH;HLA-A=HLA-A*02:01, HLA-B=HLA-B*08:01, HLA-DRB1=HLA-DRB1*15:01")

  result_single_locus <- HLA_mismatch_alleles(GL_string_recip, GL_string_donor, "HLA-A", direction = "HvG")
  expect_equal(result_single_locus, "HLA-A*03:01")

  result_homozygous_1 <- HLA_mismatch_alleles(GL_string_recip, GL_string_donor, "HLA-B", direction = "HvG", homozygous_count = 1)
  expect_equal(result_homozygous_1, "HLA-B*44:02")

  result_homozygous_2 <- HLA_mismatch_alleles(GL_string_recip, GL_string_donor, "HLA-B", direction = "HvG", homozygous_count = 2)
  expect_equal(result_homozygous_2, "HLA-B*44:02")

  expect_error(HLA_mismatch_alleles(GL_string_recip, GL_string_donor, loci, direction = "invalid"))
})

---
title: "Getting Started with immunogenetr"
output:
  rmarkdown::html_vignette:
    df_print: kable
vignette: >
  %\VignetteIndexEntry{Getting Started with immunogenetr}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = FALSE,
  comment = "#>"
)

# Render data frames and tibbles as HTML tables in chunk output, instead of
# the default monospace console-style print. The kable HTML is wrapped in a
# pandoc raw-HTML block (```{=html} ... ```) so pandoc passes the content
# through verbatim. Without this, pandoc's markdown_in_html_blocks extension
# re-parses the cell contents as markdown, turning "^...^" in GL strings
# into <sup>...</sup>.
knit_print.data.frame <- function(x, ...) {
  html <- as.character(knitr::kable(x, format = "html"))
  knitr::asis_output(paste0("\n```{=html}\n", html, "\n```\n"))
}
registerS3method("knit_print", "data.frame", knit_print.data.frame, envir = asNamespace("knitr"))
registerS3method("knit_print", "tbl_df",     knit_print.data.frame, envir = asNamespace("knitr"))
registerS3method("knit_print", "tbl",        knit_print.data.frame, envir = asNamespace("knitr"))
```

## Overview

immunogenetr is a comprehensive toolkit for clinical HLA informatics, built on
tidyverse principles. It uses the genotype list string
(GL string, <https://glstring.org/>) as its core data structure for storing and
computing HLA genotype data.

This vignette walks through the main workflows:

1. Converting tabular HLA data to GL strings
2. Splitting GL strings back into individual loci
3. Calculating mismatches between recipient and donor
4. Summarizing HLA matching for transplantation
5. Working with HLA allele names (truncation, prefixes, regex)
6. Reading HML files

## Setup

```{r message=FALSE}
library(immunogenetr)
library(dplyr)
```

## Converting tabular HLA data to GL strings

Clinical HLA data is typically stored in a tabular format, with each allele in
its own column. immunogenetr includes the `HLA_typing_1` dataset as an example:

```{r}
# HLA_typing_1 contains typing for 10 individuals across all classical HLA loci.
head(HLA_typing_1, 3)
```

The `HLA_columns_to_GLstring()` function converts these columns into a single
GL string per individual. When used inside `mutate()`, pass `.` as the first
argument to reference the working data frame:
```{r}
HLA_typing_GL <- HLA_typing_1 %>%
  # Convert all typing columns (A1 through DPB1_2) into a GL string.
  mutate(
    GL_string = HLA_columns_to_GLstring(., HLA_typing_columns = A1:DPB1_2),
    .after = patient
  ) %>%
  # Keep only patient ID and the new GL string column.
  select(patient, GL_string)

# View the GL strings.
(HLA_typing_GL)
```

Each GL string encodes the full genotype: alleles within a gene copy are
separated by `/` (ambiguity), gene copies by `+`, and loci by `^`.

## Splitting GL strings into loci

To go the other direction, `GLstring_genes()` splits a GL string back into
separate columns by locus:

```{r}
# Take the first patient's GL string and split it into locus columns.
# Note: GLstring_genes and GLstring_genes_expanded use pivot_longer on all
# columns, so only pass the GL string column (no other data types).
single_patient <- HLA_typing_GL[1, "GL_string", drop = FALSE]
GLstring_genes(single_patient, "GL_string")
```

For a fully expanded view with one allele per row, use
`GLstring_genes_expanded()`:

```{r}
GLstring_genes_expanded(single_patient, "GL_string")
```

## Calculating HLA mismatches

The mismatch functions are the core of immunogenetr. They all take a recipient
GL string, a donor GL string, one or more loci, and a direction.

Let's set up a recipient/donor pair:

```{r}
# Patient 7 is the recipient, patient 9 is the donor.
recip_gl <- HLA_typing_GL %>% filter(patient == 7) %>% pull(GL_string)
donor_gl <- HLA_typing_GL %>% filter(patient == 9) %>% pull(GL_string)
```

### Is there a mismatch? (`HLA_mismatch_logical`)

```{r}
# Check if there is an HLA-A mismatch in the graft-vs-host direction.
HLA_mismatch_logical(recip_gl, donor_gl, "HLA-A", direction = "GvH")

# Check host-vs-graft direction.
HLA_mismatch_logical(recip_gl, donor_gl, "HLA-A", direction = "HvG")
```

### How many mismatches? (`HLA_mismatch_number`)

```{r}
# Count bidirectional mismatches across several loci at once.
HLA_mismatch_number(
  recip_gl, donor_gl,
  c("HLA-A", "HLA-B", "HLA-C", "HLA-DRB1"),
  direction = "bidirectional"
)
```

### Which alleles are mismatched? (`HLA_mismatched_alleles`)

```{r}
# Identify the specific mismatched alleles in the HvG direction.
HLA_mismatched_alleles(recip_gl, donor_gl, "HLA-A", direction = "HvG")
```

### Match count (`HLA_match_number`)

```{r}
# Count the number of matches (complement of mismatches).
HLA_match_number(
  recip_gl, donor_gl,
  c("HLA-A", "HLA-B", "HLA-C", "HLA-DRB1"),
  direction = "bidirectional"
)
```

## HLA match summaries for transplantation

The `HLA_match_summary_HCT()` function provides standard match grades used in
hematopoietic cell transplantation:

```{r}
# X-of-8 matching (A, B, C, DRB1 bidirectional).
HLA_match_summary_HCT(recip_gl, donor_gl,
  direction = "bidirectional",
  match_grade = "Xof8"
)

# X-of-10 matching (adds DQB1).
HLA_match_summary_HCT(recip_gl, donor_gl,
  direction = "bidirectional",
  match_grade = "Xof10"
)
```

### Finding the best donor

A common workflow is comparing one recipient against multiple potential donors:

```{r}
# Patient 3 is the recipient; compare against all 10 donors.
recipient <- HLA_typing_GL %>%
  filter(patient == 3) %>%
  select(GL_string) %>%
  rename(GL_string_recip = GL_string)

donors <- HLA_typing_GL %>%
  rename(GL_string_donor = GL_string, donor = patient) %>%
  # Cross-join to pair recipient with each donor.
  cross_join(recipient) %>%
  # Calculate 8/8 match grade for each pair.
  mutate(
    match_8of8 = HLA_match_summary_HCT(
      GL_string_recip, GL_string_donor,
      direction = "bidirectional",
      match_grade = "Xof8"
    ),
    .after = donor
  ) %>%
  # Sort best matches first.
  arrange(desc(match_8of8))

donors %>% select(donor, match_8of8)
```

## Working with HLA allele names

### Truncation

`HLA_truncate()` reduces allele resolution to a specified number of fields:

```{r}
# Truncate a four-field allele to two fields.
HLA_truncate("HLA-A*02:01:01:01", fields = 2)

# Works on full GL strings too.
HLA_truncate("HLA-A*02:01:01:01+HLA-A*03:01:01:02^HLA-B*07:02:01:01+HLA-B*44:02:01:01",
  fields = 2
)
```

### Prefix management

`HLA_prefix_remove()` and `HLA_prefix_add()` manage the `HLA-` and locus
prefixes:

```{r}
# Remove all prefixes to get just the allele fields.
HLA_prefix_remove("HLA-A*02:01")

# Keep the locus designation but remove "HLA-".
HLA_prefix_remove("HLA-A*02:01", keep_locus = TRUE)

# Add the full prefix back.
HLA_prefix_add("02:01", "HLA-A*")

# "HLA-" is added by default.
HLA_prefix_add("A*02:01")
```

### Regex for GL string searching

`GLstring_regex()` creates regex patterns that accurately search within GL
strings, preventing partial matches across field boundaries:

```{r}
gl <- "HLA-A*02:01:01+HLA-A*68:01^HLA-B*07:01+HLA-B*15:01"

# A two-field search correctly matches the three-field allele.
pattern <- GLstring_regex("HLA-A*02:01")
stringr::str_detect(gl, pattern)

# But won't falsely match a longer allele number.
stringr::str_detect("HLA-A*02:149:01", GLstring_regex("HLA-A*02:14"))
```

## Column name repair

When working in the tidyverse, column names with dashes and asterisks are
inconvenient. `HLA_column_repair()` converts between WHO-standard (`HLA-A*`)
and tidyverse-friendly (`HLA_A`) formats:

```{r}
# GLstring_genes returns tidyverse-friendly names by default.
repaired <- GLstring_genes(single_patient, "GL_string")
names(repaired)

# Convert back to WHO format with asterisks.
who_names <- HLA_column_repair(repaired, format = "WHO", asterisk = TRUE)
names(who_names)
```

## Reading HML files

The `read_HML()` function extracts GL strings from HML (HLA Markup Language)
files, which are a standard format for reporting HLA typing results from
next-generation sequencing:

```{r}
# immunogenetr ships with two example HML files.
hml_path <- system.file("extdata", "HML_1.hml", package = "immunogenetr")
hml_result <- read_HML(hml_path)
hml_result
```

## Disclaimer
This library is intended for research use. Any application making use of this
package in a clinical setting will need to be independently validated according
to local regulations.

---
title: "Getting Started with immunogenetr"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started with immunogenetr}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Overview

immunogenetr is a comprehensive toolkit for clinical HLA informatics, built on
tidyverse principles. It uses the genotype list string
(GL string, <https://glstring.org/>) as its core data structure for storing and
computing HLA genotype data.

This vignette walks through the main workflows:

1. Converting tabular HLA data to GL strings
2. Splitting GL strings back into individual loci
3. Calculating mismatches between recipient and donor
4. Summarizing HLA matching for transplantation
5. Working with HLA allele names (truncation, prefixes, regex)
6. Reading HML files

## Setup

```{r message=FALSE}
library(immunogenetr)
library(dplyr)
```

## Converting tabular HLA data to GL strings

Clinical HLA data is typically stored in a tabular format, with each allele in
its own column. immunogenetr includes the `HLA_typing_1` dataset as an example:

```{r}
# HLA_typing_1 contains typing for 10 individuals across all classical HLA loci.
head(HLA_typing_1, 3)
```

The `HLA_columns_to_GLstring()` function converts these columns into a single
GL string per individual. When used inside `mutate()`, pass `.` as the first
argument to reference the working data frame:
```{r}
HLA_typing_GL <- HLA_typing_1 %>%
  # Convert all typing columns (A1 through DPB1_2) into a GL string.
  mutate(
    GL_string = HLA_columns_to_GLstring(., HLA_typing_columns = A1:DPB1_2),
    .after = patient
  ) %>%
  # Keep only patient ID and the new GL string column.
  select(patient, GL_string)

# View the GL strings.
(HLA_typing_GL)
```

Each GL string encodes the full genotype: alleles within a gene copy are
separated by `/` (ambiguity), gene copies by `+`, and loci by `^`.

## Splitting GL strings into loci

To go the other direction, `GLstring_genes()` splits a GL string back into
separate columns by locus:

```{r}
# Take the first patient's GL string and split it into locus columns.
# Note: GLstring_genes and GLstring_genes_expanded use pivot_longer on all
# columns, so only pass the GL string column (no other data types).
single_patient <- HLA_typing_GL[1, "GL_string", drop = FALSE]
GLstring_genes(single_patient, "GL_string")
```

For a fully expanded view with one allele per row, use
`GLstring_genes_expanded()`:

```{r}
GLstring_genes_expanded(single_patient, "GL_string")
```

## Calculating HLA mismatches

The mismatch functions are the core of immunogenetr. They all take a recipient
GL string, a donor GL string, one or more loci, and a direction.

Let's set up a recipient/donor pair:

```{r}
# Patient 7 is the recipient, patient 9 is the donor.
recip_gl <- HLA_typing_GL %>% filter(patient == 7) %>% pull(GL_string)
donor_gl <- HLA_typing_GL %>% filter(patient == 9) %>% pull(GL_string)
```

### Is there a mismatch? (`HLA_mismatch_logical`)

```{r}
# Check if there is an HLA-A mismatch in the graft-vs-host direction.
HLA_mismatch_logical(recip_gl, donor_gl, "HLA-A", direction = "GvH")

# Check host-vs-graft direction.
HLA_mismatch_logical(recip_gl, donor_gl, "HLA-A", direction = "HvG")
```

### How many mismatches? (`HLA_mismatch_number`)

```{r}
# Count bidirectional mismatches across several loci at once.
HLA_mismatch_number(
  recip_gl, donor_gl,
  c("HLA-A", "HLA-B", "HLA-C", "HLA-DRB1"),
  direction = "bidirectional"
)
```

### Which alleles are mismatched? (`HLA_mismatched_alleles`)

```{r}
# Identify the specific mismatched alleles in the HvG direction.
HLA_mismatched_alleles(recip_gl, donor_gl, "HLA-A", direction = "HvG")
```

### Match count (`HLA_match_number`)

```{r}
# Count the number of matches (complement of mismatches).
HLA_match_number(
  recip_gl, donor_gl,
  c("HLA-A", "HLA-B", "HLA-C", "HLA-DRB1"),
  direction = "bidirectional"
)
```

## HLA match summaries for transplantation

The `HLA_match_summary_HCT()` function provides standard match grades used in
hematopoietic cell transplantation:

```{r}
# X-of-8 matching (A, B, C, DRB1 bidirectional).
HLA_match_summary_HCT(recip_gl, donor_gl,
  direction = "bidirectional",
  match_grade = "Xof8"
)

# X-of-10 matching (adds DQB1).
HLA_match_summary_HCT(recip_gl, donor_gl,
  direction = "bidirectional",
  match_grade = "Xof10"
)
```

### Finding the best donor

A common workflow is comparing one recipient against multiple potential donors:

```{r}
# Patient 3 is the recipient; compare against all 10 donors.
recipient <- HLA_typing_GL %>%
  filter(patient == 3) %>%
  select(GL_string) %>%
  rename(GL_string_recip = GL_string)

donors <- HLA_typing_GL %>%
  rename(GL_string_donor = GL_string, donor = patient) %>%
  # Cross-join to pair recipient with each donor.
  cross_join(recipient) %>%
  # Calculate 8/8 match grade for each pair.
  mutate(
    match_8of8 = HLA_match_summary_HCT(
      GL_string_recip, GL_string_donor,
      direction = "bidirectional",
      match_grade = "Xof8"
    ),
    .after = donor
  ) %>%
  # Sort best matches first.
  arrange(desc(match_8of8))

donors %>% select(donor, match_8of8)
```

## Working with HLA allele names

### Truncation

`HLA_truncate()` reduces allele resolution to a specified number of fields:

```{r}
# Truncate a four-field allele to two fields.
HLA_truncate("HLA-A*02:01:01:01", fields = 2)

# Works on full GL strings too.
HLA_truncate("HLA-A*02:01:01:01+HLA-A*03:01:01:02^HLA-B*07:02:01:01+HLA-B*44:02:01:01",
  fields = 2
)
```

### Prefix management

`HLA_prefix_remove()` and `HLA_prefix_add()` manage the `HLA-` and locus
prefixes:

```{r}
# Remove all prefixes to get just the allele fields.
HLA_prefix_remove("HLA-A*02:01")

# Keep the locus designation but remove "HLA-".
HLA_prefix_remove("HLA-A*02:01", keep_locus = TRUE)

# Add the full prefix back.
HLA_prefix_add("02:01", "HLA-A*")

# "HLA-" is added by default.
HLA_prefix_add("A*02:01")
```

### Regex for GL string searching

`GLstring_regex()` creates regex patterns that accurately search within GL
strings, preventing partial matches across field boundaries:

```{r}
gl <- "HLA-A*02:01:01+HLA-A*68:01^HLA-B*07:01+HLA-B*15:01"

# A two-field search correctly matches the three-field allele.
pattern <- GLstring_regex("HLA-A*02:01")
stringr::str_detect(gl, pattern)

# But won't falsely match a longer allele number.
stringr::str_detect("HLA-A*02:149:01", GLstring_regex("HLA-A*02:14"))
```

## Column name repair

When working in the tidyverse, column names with dashes and asterisks are
inconvenient. `HLA_column_repair()` converts between WHO-standard (`HLA-A*`)
and tidyverse-friendly (`HLA_A`) formats:

```{r}
# GLstring_genes returns tidyverse-friendly names by default.
repaired <- GLstring_genes(single_patient, "GL_string")
names(repaired)

# Convert back to WHO format with asterisks.
who_names <- HLA_column_repair(repaired, format = "WHO", asterisk = TRUE)
names(who_names)
```

## Reading HML files

The `read_HML()` function extracts GL strings from HML (HLA Markup Language)
files, which are a standard format for reporting HLA typing results from
next-generation sequencing:

```{r}
# immunogenetr ships with two example HML files.
hml_path <- system.file("extdata", "HML_1.hml", package = "immunogenetr")
hml_result <- read_HML(hml_path)
hml_result
```

## Disclaimer
This library is intended for research use. Any application making use of this
package in a clinical setting will need to be independently validated according
to local regulations.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HLA_match_summary_HCT.R
\name{HLA_match_summary_HCT}
\alias{HLA_match_summary_HCT}
\title{HLA_match_summary_HCT}
\usage{
HLA_match_summary_HCT(
  GL_string_recip,
  GL_string_donor,
  direction = "bidirectional",
  match_grade,
  scope = "locus"
)
}
\arguments{
\item{GL_string_recip}{A GL string representing the recipient's HLA genotype,
and minimally containing the HLA-A, B, C and DRB1 loci (for Xof8 matching)
or the HLA-A, B, C, DRB1 and DQB1 loci (for Xof10 matching).}

\item{GL_string_donor}{A GL string representing the donor's HLA genotype,
and minimally containing the HLA-A, B, C and DRB1 loci (for Xof8 matching)
or the HLA-A, B, C, DRB1 and DQB1 loci (for Xof10 matching).}

\item{direction}{"GvH", "HvG" or "bidirectional". Default is "bidirectional".}

\item{match_grade}{"Xof8" for HLA-A, B, C and DRB1 matching or "Xof10" for
HLA-A, B, C, DRB1 and DQB1 matching.}

\item{scope}{"locus" or "genotype". Default is "locus". When "locus" (the default),
bidirectional matching takes the minimum match count at each locus independently
before summing. When "genotype", the GvH and HvG match summaries are calculated
separately across all loci, and the maximum of the two totals is returned.
Only affects results when direction is "bidirectional"; ignored otherwise.}
}
\value{
An integer value of the match grade summary.
}
\description{
Calculates the match summary for either the HLA-A, B, C and DRB1
loci (out-of-8 matching) or the HLA-A, B, C, DRB1 and DQB1 loci (out-of-10 matching),
as is commonly used for hematopoietic cell transplantation (HCT). Homozygous
mismatches are counted twice. Bidirectional matching is the default, but can
be overridden with the "direction" argument.
}
\examples{
# Example recipient and donor GL strings
file <- HLA_typing_1[, -1]
GL_string <- HLA_columns_to_GLstring(file, HLA_typing_columns = everything())

GL_string_recip <- GL_string[1]
GL_string_donor <- GL_string[2]

# Calculate mismatch numbers
HLA_match_summary_HCT(GL_string_recip, GL_string_donor,
  direction = "bidirectional", match_grade = "Xof8"
)

# Genotype-level bidirectional matching (max of GvH and HvG totals)
HLA_match_summary_HCT(GL_string_recip, GL_string_donor,
  direction = "bidirectional", match_grade = "Xof8", scope = "genotype"
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HLA_dictionary-data.R
\docType{data}
\name{HLA_dictionary}
\alias{HLA_dictionary}
\title{Data on HLA alleles in the 2008 HLA dictionary}
\format{
A tibble:
\describe{
  \item{IPD.Accession}{IPD-IMGT/HLA Accession number}
  \item{HLA.Allele}{HLA allele name}
  \item{WHO.Assigned.Type}{WHO-assigned serologic equivalent}
  ...
}
}
\source{
<https://doi.org/10.1111/j.1399-0039.2008.01183.x>
}
\usage{
data(HLA_dictionary)
}
\description{
Data on HLA alleles in the 2008 HLA dictionary
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GLstring_gene_copies_combine.R
\name{GLstring_gene_copies_combine}
\alias{GLstring_gene_copies_combine}
\title{GLstring_gene_copies_combine}
\usage{
GLstring_gene_copies_combine(.data, columns, sample_column = "sample")
}
\arguments{
\item{.data}{A data frame}

\item{columns}{The names of the columns in the data frame that contain typing
information to be combined}

\item{sample_column}{The name of the column that identifies samples in the
data frame. Default is "sample".}
}
\value{
A data frame with the specified columns combined into a single column
for each locus, in GL string format.
}
\description{
A function for combining two columns of typing from the same
locus into a single column in the appropriate GL string format.
}
\examples{
library(dplyr)
HLA_typing_1 \%>\%
  mutate(across(A1:B2, ~ HLA_prefix_add(.))) \%>\%
  GLstring_gene_copies_combine(c(A1:B2), sample_column = patient)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HLA_match_number.R
\name{HLA_match_number}
\alias{HLA_match_number}
\title{HLA_match_number}
\usage{
HLA_match_number(
  GL_string_recip,
  GL_string_donor,
  loci,
  direction = "bidirectional"
)
}
\arguments{
\item{GL_string_recip}{A GL string representing the recipient's HLA genotype.}

\item{GL_string_donor}{A GL string representing the donor's HLA genotype.}

\item{loci}{A character vector specifying the loci to be considered for
mismatch calculation. HLA-DRB3/4/5 (and their serologic equivalents DR51/52/53)
are considered once locus for this function, and should be called in this argument
as "HLA-DRB3/4/5" or "HLA-DR51/52/53", respectively.}

\item{direction}{A character string indicating the direction of match.
Options are "HvG" (host vs. graft), "GvH" (graft vs. host), "bidirectional"
(the minimum value of "HvG" and "GvH").}
}
\value{
An integer value or a character string:
- If `loci` includes only one locus, the function returns an integer
mismatch count for that locus.
- If `loci` includes multiple loci, the function returns a character
string in the format "Locus1=Count1, Locus2=Count2, ...".
}
\description{
Calculates the number of HLA matches as two minus the number of
mismatches from `HLA_mismatch_number`. Homozygous mismatches are counted twice.
Supports match calculations for host-vs-graft (HvG), graft-vs-host (GvH),
or bidirectional. Bidirectional matching is the default, but can be overridden
using the "direction" argument.
}
\examples{

file <- HLA_typing_1[, -1]
GL_string <- HLA_columns_to_GLstring(file, HLA_typing_columns = everything())
GL_string_recip <- GL_string[1]
GL_string_donor <- GL_string[2]

loci <- c("HLA-A", "HLA-B")

# Calculate mismatch numbers (Host vs. Graft)
HLA_match_number(GL_string_recip, GL_string_donor, loci, direction = "HvG")

# Calculate mismatch numbers (Graft vs. Host)
HLA_match_number(GL_string_recip, GL_string_donor, loci, direction = "GvH")

# Calculate mismatch numbers (Bidirectional)
HLA_match_number(GL_string_recip, GL_string_donor,
  loci,
  direction = "bidirectional"
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GLstring_genotype_ambiguity.R
\name{GLstring_genotype_ambiguity}
\alias{GLstring_genotype_ambiguity}
\title{GLstring_genotype_ambiguity}
\usage{
GLstring_genotype_ambiguity(data, columns, keep_ambiguities = FALSE)
}
\arguments{
\item{data}{A data frame}

\item{columns}{The names of the columns in the data frame that contain GL strings}

\item{keep_ambiguities}{A logical value indicating whether to retain the
remaining ambiguities in separate columns with "_genotype_ambiguity" appended
to the original column names. Default is FALSE.}
}
\value{
A data frame with the first genotype ambiguity retained in the
original columns. If \code{keep_ambiguities} is TRUE, the remaining
ambiguities are placed in separate columns.
}
\description{
This function processes GL strings in the specified columns of
a data frame to retain only the first genotype ambiguity, optionally
retaining the remaining ambiguities in a separate column with "_ambiguity"
appended. The function ensures that genes have been separated from the GL
strings prior to execution; otherwise, an error will be thrown if a "^" is
detected in the GL strings.
}
\examples{
HLA_type <- data.frame(
  sample = c("sample1", "sample2"),
  HLA_A = c("A*01:01+A*68:01|A*01:02+A*68:55|A*01:99+A*68:66", "A*02:01+A*03:01|A*02:02+A*03:03"),
  HLA_B = c("B*07:02+B*58:01|B*07:03+B*58:09", "B*08:01+B*15:01|B*08:02+B*15:17"),
  stringsAsFactors = FALSE
)

GLstring_genotype_ambiguity(HLA_type, columns = c("HLA_A", "HLA_B"), keep_ambiguities = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_HML.R
\name{read_HML}
\alias{read_HML}
\title{read_HML}
\usage{
read_HML(HML_file)
}
\arguments{
\item{HML_file}{The path to an HML file.}
}
\value{
A tibble with the sample name and the GL string.
}
\description{
Reads the GL strings of HML files and returns a tibble with
the full genotype for each sample.
}
\examples{
HML_1 <- system.file("extdata", "HML_1.hml", package = "immunogenetr")
HML_2 <- system.file("extdata", "hml_2.hml", package = "immunogenetr")

read_HML(HML_1)
read_HML(HML_2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HLA_mismatch_number.R
\name{HLA_mismatch_number}
\alias{HLA_mismatch_number}
\title{HLA_mismatch_number}
\usage{
HLA_mismatch_number(
  GL_string_recip,
  GL_string_donor,
  loci,
  direction,
  homozygous_count = 2
)
}
\arguments{
\item{GL_string_recip}{A GL string representing the recipient's HLA genotype.}

\item{GL_string_donor}{A GL string representing the donor's HLA genotype.}

\item{loci}{A character vector specifying the loci to be considered for
mismatch calculation. HLA-DRB3/4/5 (and their serologic equivalents DR51/52/53)
are considered once locus for this function, and should be called in this argument
as "HLA-DRB3/4/5" or "HLA-DR51/52/53", respectively.}

\item{direction}{A character string indicating the direction of mismatch.
Options are "HvG" (host vs. graft), "GvH" (graft vs. host), "bidirectional"
(the max value of "HvG" and "GvH"), or "SOT" (host vs. graft, as is used for
mismatching in solid organ transplantation).}

\item{homozygous_count}{An integer specifying how to count homozygous mismatches.
Defaults to 2, where homozygous mismatches are treated as two mismatches,
regardless if one or two alleles are supplied in the GL string (in cases
where one allele is supplied, it is duplicated by the function). If
specified as 1, homozygous mismatches are only counted once, regardless of
whether one or two alleles are supplied in the GL string (in cases where
two alleles are supplied, the second identical allele is deleted).}
}
\value{
An integer value or a character string:
- If `loci` includes only one locus, the function returns an integer
mismatch count for that locus.
- If `loci` includes multiple loci, the function returns a character
string in the format "Locus1=Count1, Locus2=Count2, ...".
}
\description{
Calculates the number of mismatched HLA alleles between a
recipient and a donor across specified loci. Supports mismatch calculations
for host-vs-graft (HvG), graft-vs-host (GvH), or bidirectional.
}
\examples{

file <- HLA_typing_1[, -1]
GL_string <- HLA_columns_to_GLstring(file, HLA_typing_columns = everything())

GL_string_recip <- GL_string[1]
GL_string_donor <- GL_string[2]

loci <- c("HLA-A", "HLA-DRB3/4/5", "HLA-DPB1")

# Calculate mismatch numbers (Host vs. Graft)
HLA_mismatch_number(GL_string_recip, GL_string_donor, loci, direction = "HvG")

# Calculate mismatch numbers (Graft vs. Host)
HLA_mismatch_number(GL_string_recip, GL_string_donor, loci, direction = "GvH")

# Calculate mismatch numbers (Bidirectional)
HLA_mismatch_number(GL_string_recip, GL_string_donor,
  loci,
  direction = "bidirectional"
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/immunogenetr-package.R
\docType{package}
\name{immunogenetr-package}
\alias{immunogenetr}
\alias{immunogenetr-package}
\title{immunogenetr: A Comprehensive Toolkit for Clinical HLA Informatics}
\description{
A comprehensive toolkit for clinical Human Leukocyte Antigen (HLA) informatics, built on 'tidyverse' \url{https://tidyverse.tidyverse.org/} principles and making use of genotype list string (GL string, Mack et al. (2023) \doi{10.1111/tan.15126}) for storing and computing HLA genotype data. Specific functionalities include: coercion of HLA data in tabular format to and from GL string; calculation of matching and mismatching in all directions, with multiple output formats; automatic formatting of HLA data for searching within a GL string; truncation of molecular HLA data to a specific number of fields; and reading HLA genotypes in HML files and extracting the GL string. This library is intended for research use. Any application making use of this package in a clinical setting will need to be independently validated according to local regulations.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/k96nb01/immunogenetr_package}
  \item \doi{10.1016/j.humimm.2025.111619}
  \item Report bugs at \url{https://github.com/k96nb01/immunogenetr_package/issues}
}

}
\author{
\strong{Maintainer}: Nicholas Brown \email{nicholas.brown@pennmedicine.upenn.edu} (\href{https://orcid.org/0000-0002-0046-2315}{ORCID})

Authors:
\itemize{
  \item Busra Coskun \email{bcoskun@sas.upenn.edu} (\href{https://orcid.org/0009-0008-6828-3453}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GLstring_genes_expanded.R
\name{GLstring_genes_expanded}
\alias{GLstring_genes_expanded}
\title{GLstring_genes_expanded}
\usage{
GLstring_genes_expanded(data, gl_string)
}
\arguments{
\item{data}{A data frame containing GL strings for HLA data.}

\item{gl_string}{The name of the column in the data frame that contains
GL strings.}
}
\value{
A data frame with expanded columns, where each row has a single
allele for a specific locus.
}
\description{
This function processes a specified column in a data frame
that contains GL strings. It separates the GL strings, identifies the HLA
loci, and transforms the data into a wider format with loci as column names.
It also creates multiple rows to separate each locus in the allele.
}
\examples{

file <- HLA_typing_1[, -1]
GL_string <- data.frame("GL_string" = HLA_columns_to_GLstring(
  file,
  HLA_typing_columns = everything()
))
GL_string <- GL_string[1, , drop = FALSE] # When considering first patient
result <- GLstring_genes_expanded(GL_string, "GL_string")
print(result)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HLA_typing_1-data.R
\docType{data}
\name{HLA_typing_1}
\alias{HLA_typing_1}
\title{Synthetic HLA typing data for 10 individuals for the HLA-A, B, C, DRB1, DRB3/4/5,
DQB1, DQA1, DPB1 and DPA1 loci.}
\format{
A tibble:
\describe{
  \item{patient}{Patients 1-10}
  \item{A1:DPB1_2}{HLA allele names}
    ...
}
}
\source{
Synthetic data from Nicholas K. Brown.
}
\usage{
data(HLA_typing_1)
}
\description{
Synthetic HLA typing data for 10 individuals for the HLA-A, B, C, DRB1, DRB3/4/5,
DQB1, DQA1, DPB1 and DPA1 loci.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HLA_column_repair.R
\name{HLA_column_repair}
\alias{HLA_column_repair}
\title{HLA_column_repair}
\usage{
HLA_column_repair(data, format = "tidyverse", asterisk = FALSE)
}
\arguments{
\item{data}{A data frame}

\item{format}{Either "tidyverse" or "WHO".}

\item{asterisk}{Logical value to return column with an asterisk.}
}
\value{
A data frame object with column names renamed in the specified format.
}
\description{
This function will change column names that have the official
 HLA nomenclature (e.g. "HLA-A*" or "HLA-A") to a format more easily selected
 in tidyverse functions (e.g. "HLA_A"). The dash and asterisk are special
 characters in R, and makes selecting columns by name difficult. This function
 will also allow for easily changing back to WHO-compliant nomenclature
 (e.g. "HLA-A*").
}
\examples{
HLA_type <- data.frame(
  "HLA-A*" = c("01:01", "02:01"),
  "HLA-B*" = c("07:02", "08:01"),
  "HLA-C*" = c("03:04", "04:01"),
  stringsAsFactors = FALSE
)

HLA_column_repair(HLA_type, format = "tidyverse")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GLstring_to_ambiguity_table.R
\name{GLstring_to_ambiguity_table}
\alias{GLstring_to_ambiguity_table}
\title{GLstring_to_ambiguity_table}
\usage{
GLstring_to_ambiguity_table(GL_string)
}
\arguments{
\item{GL_string}{A GL string that encodes HLA alleles and their potential
ambiguities}
}
\value{
A tibble that contains the expanded GL string with separate columns
for possible gene locations, loci, genotype ambiguities, genotypes, haplotypes,
and alleles, each with associated identifiers
}
\description{
A function that expands a GL string into a longer, more detailed
format (also known as an ambiguity table) by separating the string into its
components resulting from its hierarchical set of operators, including gene
locations, loci, genotypes, haplotypes, and alleles. The function processes
each level of the GL string and assigns identifiers for each hierarchical
component. The resulting table can be assembled back into a GL string using
the function `ambiguity_table_to_GLstring`. This function is an alias of
`GLstring_expand_longer`.
}
\examples{
file <- HLA_typing_1[, -1]
GL_string <- HLA_columns_to_GLstring(file, HLA_typing_columns = everything())
result <- GLstring_to_ambiguity_table(GL_string[1])
print(result)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HLA_truncate.R
\name{HLA_truncate}
\alias{HLA_truncate}
\title{HLA_truncate}
\usage{
HLA_truncate(
  data,
  fields = 2,
  keep_suffix = TRUE,
  keep_G_P_group = FALSE,
  remove_duplicates = FALSE
)
}
\arguments{
\item{data}{A string containing an HLA allele or a GL string.}

\item{fields}{An integer specifying the number of fields to retain in the
truncated values. Default is 2.}

\item{keep_suffix}{A logical value indicating whether to retain any
WHO-recognized suffixes. Default is TRUE.}

\item{keep_G_P_group}{A logical value indicating whether to retain any
G or P group designations. Default is FALSE.}

\item{remove_duplicates}{A logical value indicating whether to remove duplicated
values from a GL string after truncation. Default is FALSE.}
}
\value{
A string with the HLA typing truncated according to
the specified number of fields and optional suffix retention.
}
\description{
This function truncates HLA typing values in molecular nomenclature
(for example from 4 fields to 2 fields). The truncation is based on the number
of fields specified and optionally retains any WHO-recognized suffixes
(L, S, C, A, Q, or N) or G and P group designations (G or P). This function
will work on individual alleles (e.g. "HLA-A*02:01:01:01") or on all alleles
in a GL string (e.g. "HLA-A*02:01:01:01+HLA-A*68:01:01^HLA-DRB1*01:01:01+HLA-DRB1*03:01:01").

Note: depending on arguments used, this function can output HLA alleles that do not
exist in the IPD-IMGT/HLA database. For example, truncating the allele "DRB4*01:03:01:02N"
to 2 fields would result in "DRB4*01:03N," which does not exist in the IPD-IMGT/HLA database.
Users should take care in setting the parameters for this function.
}
\examples{

# The Haplotype_frequencies dataset contains a table with HLA typing spread across multiple columns:
print(Haplotype_frequencies)

# The `HLA_truncate` function can be used to truncate the typing results to 2 fields:
library(dplyr)
Haplotype_frequencies \%>\% mutate(
  across(
    "HLA-A":"HLA-DPB1",
    ~ HLA_truncate(
      .,
      fields = 2,
      keep_suffix = TRUE,
      keep_G_P_group = FALSE
    )
  )
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HLA_columns_to_GLstring.R
\name{HLA_columns_to_GLstring}
\alias{HLA_columns_to_GLstring}
\title{HLA_columns_to_GLstring}
\usage{
HLA_columns_to_GLstring(
  data,
  HLA_typing_columns,
  prefix_to_remove = "",
  suffix_to_remove = ""
)
}
\arguments{
\item{data}{A data frame with each row including an HLA typing result, with
individual columns containing a single allele.}

\item{HLA_typing_columns}{A list of columns containing the HLA alleles. Tidyselect is supported.}

\item{prefix_to_remove}{An optional string of characters to remove from the
locus names. The goal is to get the column names to the locus and a number. For example,
columns named "mDRB11Cd" and "mDRB12Cd" should use the `prefix_to_remove` value of "m".}

\item{suffix_to_remove}{An optional string of characters to remove from the
locus names. Using the example above, the `suffix_to_remove` value will be "Cd".}
}
\value{
A list of GL strings in the order of the original data frame.
}
\description{
A function to take HLA typing data spread across different columns,
as is often found in wild-caught data, and transform it to a GL string. If column names
have anything besides the locus name and a number (e.g. "mA1Cd" instead of just "A1"),
the function will have trouble determining the locus from the column name. The `prefix_to_remove`
and `suffix_to_remove` arguments can be used to clean up the column names. See the example for
how these arguments are used.
}
\examples{
# The HLA_typing_LIS dataset contains a table as might be found in a clinical laboratory
# information system:
print(HLA_typing_LIS)

# The `HLA_columns_to_GLString` function can be used to coerce typing spread across
# multiple columns into a GL string:
library(dplyr)
HLA_typing_LIS \%>\%
  mutate(
    GL_string = HLA_columns_to_GLstring(
      ., # Note that if this function is used inside a `mutate` call "." will have to be
      # used as the first argument to extract data from the working data frame.
      HLA_typing_columns = mA1Cd.recipient:mDPB12cd.recipient,
      prefix_to_remove = "m",
      suffix_to_remove = "Cd.recipient"
    ),
    .after = patient
  ) \%>\%
  select(patient, GL_string)

# Using the base pipe:
HLA_typing_LIS |>
  mutate(
    GL_string = HLA_columns_to_GLstring(
      HLA_typing_LIS, # If using the base pipe, the first argument will have to be
      # the working data frame.
      HLA_typing_columns = mA1Cd.recipient:mDPB12cd.recipient,
      prefix_to_remove = "m",
      suffix_to_remove = "Cd.recipient"
    ),
    .after = patient
  ) |>
  select(patient, GL_string)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mismatch_table_2016-data.R
\docType{data}
\name{mismatch_table_2016}
\alias{mismatch_table_2016}
\title{Consensus mismatch numbers for every possible allele combination
at a single locus, from the 2016 publication.}
\format{
A tibble:
\describe{
  \item{Patient:Donor}{allele combinations for patient and donor}
  \item{#GvH:#Max}{consensus mismatch number for each direction}
    ...
}
}
\source{
<https://www.nature.com/articles/bmt2010132>
}
\usage{
data(mismatch_table_2016)
}
\description{
Consensus mismatch numbers for every possible allele combination
at a single locus, from the 2016 publication.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HLA_prefix_add.R
\name{HLA_prefix_add}
\alias{HLA_prefix_add}
\title{HLA_prefix_add}
\usage{
HLA_prefix_add(data, prefix = "HLA-")
}
\arguments{
\item{data}{A string with a single HLA allele, a GL string of HLA alleles,
or a character vector containing either of the previous.}

\item{prefix}{A character string to be added as a prefix to the alleles.
Default is "HLA-".}
}
\value{
A vector with the specified prefix added to the values.
}
\description{
This function adds a specified prefix to the beginning of
each HLA type, and works on a single allele or all alleles in a GL string.
Useful for adding HLA or gene prefixes.
}
\examples{
# The HLA_typing_LIS dataset contains a table as might be found in a clinical
# laboratory information system:
print(HLA_typing_LIS)

# The `HLA_prefix_add` function can be used to add the correct HLA prefixes to the table:
library(dplyr)
HLA_typing_LIS \%>\% mutate(
  across(
    mA1Cd.recipient:mA2Cd.recipient,
    ~ HLA_prefix_add(., "HLA-A*")
  )
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GLstring_regex.R
\name{GLstring_regex}
\alias{GLstring_regex}
\title{GLstring_regex}
\usage{
GLstring_regex(data)
}
\arguments{
\item{data}{A string containing an HLA allele.}
}
\value{
A string with the HLA allele formatted as a regex pattern.
}
\description{
This function will format an HLA allele (e.g. "HLA-A*02:01") to
a regex pattern for searching within a GL string. Note that in order for this
function to work properly, the full HLA allele name, including prefixes, is
required. Allele values of "A*02:01" will need to be updated to "HLA-A*02:01",
and "A2" will need to be updated to "HLA-A2". The `HLA_prefix_add` function
is useful in these situations.
}
\examples{

# To understand how the function works we can see how it alters the allele "HLA-A*02:01":

GLstring_regex("HLA-A*02:01")

# The result is the same allele with extra formatting to escape special characters found
# in a GL string, as well as the ability to accurately search for an allele in a GL string.
# For example, we would not want the allele "HLA-A*02:14" to match to "HLA-A*02:149:01",
# which would happen if we simply escaped the special characters:

library(stringr)
str_view("HLA-A*02:149:01", str_escape("HLA-A*02:14"), match = NA)

# Using `GLstring_regex` prevents this:

str_view("HLA-A*02:149:01", GLstring_regex("HLA-A*02:14"), match = NA)

# Using a longer GL string with multiple alleles and loci:

GL_string <- "HLA-A*02:01:01+HLA-A*68:01^HLA-B*07:01+HLA-B*15:01"

# We can match any allele accurately:

str_view(GL_string, GLstring_regex("HLA-A*68:01"), match = NA)

# Note that alleles supplied with fewer fields than in the GL string will also match:

str_view(GL_string, GLstring_regex("HLA-A*02:01"), match = NA)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HLA_mismatch_base.R
\name{HLA_mismatch_base}
\alias{HLA_mismatch_base}
\title{HLA_mismatch_base}
\usage{
HLA_mismatch_base(
  GL_string_recip,
  GL_string_donor,
  loci,
  direction,
  homozygous_count = 2
)
}
\arguments{
\item{GL_string_recip}{A GL string representing the recipient's HLA genotype.}

\item{GL_string_donor}{A GL string representing the donor's HLA genotype.}

\item{loci}{A character vector specifying the loci to be considered for
mismatch calculation. HLA-DRB3/4/5 (and their serologic equivalents DR51/52/53)
are considered once locus for this function, and should be called in this argument
as "HLA-DRB3/4/5" or "HLA-DR51/52/53", respectively.}

\item{direction}{A character string indicating the direction of mismatch.
Options are "HvG" (host vs. graft) or "GvH" (graft vs. host).}

\item{homozygous_count}{An integer specifying how to handle homozygosity.
Defaults to 2, where homozygous alleles are treated as duplicated for
mismatch calculations. Can be specified to be 1, in which case homozygous
alleles are treated as single occurrences without duplication.}
}
\value{
A character vector, where each element is a string summarizing the
mismatches for the specified loci. The strings are formatted as
comma-separated locus mismatch entries if multiple loci are supplied, or
simple GL strings if a single locus is supplied.
}
\description{
A function to return a string of mismatches between recipient
and donor HLA genotypes represented as GL strings. The function finds
mismatches based on the direction of comparison specified in the inputs
and also handles homozygosity. As the name implies, this function is the base
for all other mismatch (and matching) functions. This function is not meant to be
called directly; it is better to use one of the derivative functions.
}
\examples{
file <- HLA_typing_1[, -1]
GL_string <- HLA_columns_to_GLstring(file, HLA_typing_columns = everything())

GL_string_recip <- GL_string[1]
GL_string_donor <- GL_string[2]

loci <- c("HLA-A", "HLA-DRB3/4/5", "HLA-DPB1")
mismatches <- HLA_mismatch_base(GL_string_recip, GL_string_donor, loci, direction = "HvG")
print(mismatches)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Haplotype_frequencies-data.R
\docType{data}
\name{Haplotype_frequencies}
\alias{Haplotype_frequencies}
\title{Ten HLA-A~C~B~DRB3/4/5~DRB1~DQA1~DQB1~DPA1~DPB1 haplotypes and their
frequencies in the global population.}
\format{
A tibble:
\describe{
  \item{HLA-A:HLA-DPB1}{HLA allele names}
  \item{Global_Fq}{haplotype frequencies}
  \item{Global_n}{haplotype observations}
  \item{Global_rank}{haplotype rank}
    ...
}
}
\source{
<https://doi.org/10.1016/j.humimm.2021.04.007>
}
\usage{
data(Haplotype_frequencies)
}
\description{
Ten HLA-A~C~B~DRB3/4/5~DRB1~DQA1~DQB1~DPA1~DPB1 haplotypes and their
frequencies in the global population.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HLA_mismatched_alleles.R
\name{HLA_mismatched_alleles}
\alias{HLA_mismatched_alleles}
\title{HLA_mismatched_alleles}
\usage{
HLA_mismatched_alleles(
  GL_string_recip,
  GL_string_donor,
  loci,
  direction,
  homozygous_count = 2
)
}
\arguments{
\item{GL_string_recip}{A GL strings representing the recipient's HLA genotypes.}

\item{GL_string_donor}{A GL strings representing the donor's HLA genotypes.}

\item{loci}{A character vector specifying the loci to be considered for
mismatch calculation. HLA-DRB3/4/5 (and their serologic equivalents DR51/52/53)
are considered once locus for this function, and should be called in this argument
as "HLA-DRB3/4/5" or "HLA-DR51/52/53", respectively.}

\item{direction}{A character string indicating the direction of mismatch.
Options are "HvG" (host vs. graft), "GvH" (graft vs. host), "bidirectional"
(the max value of "HvG" and "GvH"), or "SOT" (host vs. graft, as is used for
mismatching in solid organ transplantation).}

\item{homozygous_count}{An integer specifying how to handle homozygosity.
Defaults to 2, where homozygous alleles are treated as duplicated for
mismatch calculations. Can be specified as 1, in which case homozygous
alleles are treated as single occurrences without duplication (in other words,
homozygous mismatches are only "counted" once).}
}
\value{
A character vector, where each element is a string summarizing the
mismatches for the specified loci. The strings are formatted as
comma-separated locus mismatch entries if multiple loci were supplied, or as
simple GL strings if a single locus was supplied.
}
\description{
A function to return a string of mismatches between recipient
and donor HLA genotypes represented as GL strings. The function finds
mismatches based on the direction of comparison specified in the inputs
and also handles homozygosity.

`HLA_mismatch_alleles` and `HLA_mismatched_alleles` are synonyms.
}
\examples{
file <- HLA_typing_1[, -1]
GL_string <- HLA_columns_to_GLstring(file, HLA_typing_columns = everything())

GL_string_recip <- GL_string[1]
GL_string_donor <- GL_string[2]

loci <- c("HLA-A", "HLA-DRB3/4/5", "HLA-DPB1")
mismatches <- HLA_mismatched_alleles(GL_string_recip, GL_string_donor, loci, direction = "HvG")
print(mismatches)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ambiguity_table_to_GLstring.R
\name{ambiguity_table_to_GLstring}
\alias{ambiguity_table_to_GLstring}
\title{ambiguity_table_to_GLstring}
\usage{
ambiguity_table_to_GLstring(data, remove_duplicates = FALSE)
}
\arguments{
\item{data}{A data frame containing columns that represent possible gene
locations, loci, genotype ambiguities, genotypes, and haplotypes.}

\item{remove_duplicates}{A logical value indicating if the function will check
for duplicate entries at each step and remove them before assembling the final
GL string. Useful if the ambiguity table has been altered, for example by
truncating allele designations. Default is FALSE.}
}
\value{
A GL string representing the combined gene locations, loci, genotype
ambiguities, genotypes, and haplotypes.
}
\description{
A function that converts a data table of HLA allele ambiguities
(e.g. as created by `GLstring_expand_longer` or `GLstring_to_ambiguity_table`)
into a GL string format. The function processes the table by combining allele
ambiguities, haplotypes, gene copies, and loci into a structured GL string.
}
\examples{
# Example data frame input
data <- tibble::tribble(
  ~value, ~entry, ~possible_gene_location,
  ~locus, ~genotype_ambiguity, ~genotype, ~haplotype, ~allele,
  "HLA-A*01:01:01:01", 1, 1,
  1, 1, 1, 1, 1,
  "HLA-A*01:02", 1, 1,
  1, 1, 1, 1, 2,
  "HLA-A*01:03", 1, 1,
  1, 1, 1, 1, 3,
  "HLA-A*01:95", 1, 1,
  1, 1, 1, 1, 4,
  "HLA-A*24:02:01:01", 1, 1,
  1, 1, 2, 1, 1,
  "HLA-A*01:01:01:01", 1, 1,
  1, 2, 1, 1, 1,
  "HLA-A*01:03", 1, 1,
  1, 2, 1, 1, 2,
  "HLA-A*24:03:01:01", 1, 1,
  1, 2, 2, 1, 1,
  "HLA-B*07:01:01", 1, 1,
  2, 1, 1, 1, 1,
  "B*15:01:01", 1, 1,
  2, 1, 2, 1, 1,
  "B*15:02:01", 1, 1,
  2, 1, 2, 1, 2,
  "B*07:03", 1, 1,
  2, 2, 1, 1, 1,
  "B*15:99:01", 1, 1,
  2, 2, 2, 1, 1,
  "HLA-DRB1*03:01:02", 1, 1,
  3, 1, 1, 1, 1,
  "HLA-DRB5*01:01:01", 1, 1,
  3, 1, 1, 2, 1,
  "HLA-KIR2DL5A*0010101", 1, 1,
  3, 1, 2, 1, 1,
  "HLA-KIR2DL5A*0010201", 1, 1,
  3, 1, 3, 1, 1,
  "HLA-KIR2DL5B*0010201", 1, 2,
  1, 1, 1, 1, 1,
  "HLA-KIR2DL5B*0010301", 1, 2,
  1, 1, 2, 1, 1
)

ambiguity_table_to_GLstring(data)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HLA_prefix_remove.R
\name{HLA_prefix_remove}
\alias{HLA_prefix_remove}
\title{HLA_prefix_remove}
\usage{
HLA_prefix_remove(data, keep_locus = FALSE)
}
\arguments{
\item{data}{A string with a single HLA allele, a GL string of HLA alleles,
or a character vector containing either of the previous.}

\item{keep_locus}{A logical value indicating whether to retain any locus values.
The default value is FALSE.}
}
\value{
A vector modified to remove HLA and optionally locus prefixes.
}
\description{
This function removes HLA and optionally locus prefixes from a string of HLA typing:
"HLA-A2" changes to "A2" or "2". By default, HLA and locus prefixes are removed. This function
also works on each allele in a GL string.
}
\examples{
# The HLA_typing_1 dataset contains a table with HLA typing spread across multiple columns:
print(HLA_typing_1)

# The `HLA_prefix_remove` function can be used to get each column to have only the
# colon-separated fields:
library(dplyr)
HLA_typing_1 \%>\% mutate(
  across(
    A1:DPB1_2,
    ~ HLA_prefix_remove(.)
  )
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HLA_validate.R
\name{HLA_validate}
\alias{HLA_validate}
\title{HLA_validate}
\usage{
HLA_validate(data)
}
\arguments{
\item{data}{A string containing an HLA allele.}
}
\value{
A string with a valid HLA allele or NA if no valid allele was present.
}
\description{
Returns only HLA alleles in valid nomenclature, either serologic or molecular.
Simple numbers, such as "2" or "27" will be returned as-is. Suffixes that are not
WHO-recognized suffixes (L, S, C, A, Q, N) or G or P group designations will be removed.
For example "novel" at the end of the allele will be removed, while "n" at the end of the
allele will be retained. Other values, such as "blank" or "-" will be converted to NA values.
This function is helpful for cleaning up the typing of an entire table of HLA values.
}
\examples{
HLA_validate("HLA-A2")
HLA_validate("A*02:01:01:01N")
HLA_validate("A*02:01:01N")
HLA_validate("HLA-DRB1*02:03novel")
HLA_validate("HLA-DQB1*03:01v")
HLA_validate("HLA-DRB1*02:03P")
HLA_validate("HLA-DPB1*04:01:01G")
HLA_validate("2")
HLA_validate(2)
HLA_validate("B27")
HLA_validate("A*010101")
HLA_validate("-")
HLA_validate("blank")

# The HLA_typing_LIS dataset contains a table with HLA typing spread across multiple columns:
print(HLA_typing_LIS)

# Cleaning up the entire table. Note that blank values will be converted to "NA".
library(dplyr)
HLA_typing_LIS \%>\% mutate(
  across(
    mA1Cd.recipient:mDPB12cd.recipient,
    ~ HLA_validate(.)
  )
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HLA_mismatch_logical.R
\name{HLA_mismatch_logical}
\alias{HLA_mismatch_logical}
\title{HLA_mismatch_logical}
\usage{
HLA_mismatch_logical(GL_string_recip, GL_string_donor, loci, direction)
}
\arguments{
\item{GL_string_recip}{A GL strings representing the recipient's HLA genotypes.}

\item{GL_string_donor}{A GL strings representing the donor's HLA genotypes.}

\item{loci}{A character vector specifying the loci to be considered for
mismatch calculation. HLA-DRB3/4/5 (and their serologic equivalents DR51/52/53)
are considered once locus for this function, and should be called in this argument
as "HLA-DRB3/4/5" or "HLA-DR51/52/53", respectively.}

\item{direction}{A character string indicating the direction of mismatch.
Options are "HvG" (host vs. graft), "GvH" (graft vs. host), "bidirectional"
(if either "HvG" or "GvH" is TRUE), or "SOT" (host vs. graft, as is used for
mismatching in solid organ transplantation).}
}
\value{
A logical value (`TRUE` or `FALSE`):
- `TRUE` if there are mismatches between recipient and donor HLA alleles.
- `FALSE` if there are no mismatches.
}
\description{
Determines if there are any mismatches between recipient and
donor HLA alleles for the specified loci. Returns `TRUE` if mismatches are
present, and `FALSE` otherwise.
}
\examples{

file <- HLA_typing_1[, -1]
GL_string <- HLA_columns_to_GLstring(file, HLA_typing_columns = everything())

GL_string_recip <- GL_string[1]
GL_string_donor <- GL_string[2]

loci <- c("HLA-A", "HLA-DRB3/4/5", "HLA-DPB1")
mismatches <- HLA_mismatch_logical(GL_string_recip, GL_string_donor, loci, direction = "HvG")
print(mismatches)

}
